/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <client/sandboxer-service-glue.h>
#include <client/sandboxer-process-glue.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <glib.h>
#include <glib-object.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>


/*
 *
 *  Basic Operation
 *  can be called in one of two ways:
 *    sandboxer program arg[1] arg[2] arg[3] ... arg[n]
 *  
 *    program -> sandboxer
 *    program.bin arg[1] arg[2] arg[3] arg[3] ... arg[n]
 *
 *  in both cases, wait for and return
 *  exit value of program
 *
 *  if killed, then kill program as well
 */

DBusGConnection *connection;
DBusGProxy *process_proxy;
gchar *secret_cookie;
gboolean needkill;
GMainLoop *loop;

static const gchar* 
canonicalize_name_or_path(const char* arg);

static void
cleanup(const int unused);

static void
terminated_signal_handler (DBusGProxy *proxy, const int exit_status, 
			   gpointer user_data);

static gchar** 
setup_argv(gchar **argv);

static gchar*
find_program_in_path_or_as_is(gchar *arg);




static void
cleanup(const int unused)
{
  gboolean ret;
  GError *error= NULL;
  
  if (needkill) {
    ret= com_intel_sandboxer_Process_kill(process_proxy, secret_cookie, 
					  SIGTERM, &error);
    if (FALSE == ret) {
      g_error("Could not run Process.Kill method:%s", error->message);
    }
  }  
}


static void
terminated_signal_handler (DBusGProxy *proxy, const int exit_status, 
			   gpointer user_data)
{
  g_assert(proxy);
  g_assert(user_data);

  *((gint*)user_data)= exit_status;
  needkill= FALSE;
  g_main_loop_quit(loop);
}


static gchar*
find_program_in_path_or_as_is(gchar *arg)
{
  g_assert(arg);
  if (g_file_test(arg, G_FILE_TEST_IS_EXECUTABLE)) {
    return g_strdup(arg);
  }
  return g_find_program_in_path(arg);
}

static gchar** 
setup_argv(gchar **argv)
{
  gchar *argv0;
  GPtrArray *args;
  int start_from;
  int i;


  args= g_ptr_array_new();
  argv0= find_program_in_path_or_as_is(argv[0]);
  if (g_file_test(argv0, G_FILE_TEST_IS_SYMLINK)) {
    /* symlink case */
    char *realname;

    realname= g_strdup_printf("%s.bin", argv0);
    g_ptr_array_add(args, realname);
    start_from= 1;
  } else {
    /* non-symlink case */
    if (!argv[1]) {
      g_printerr("Usage: %s <program> [arg1, arg2, ... argn]\n");
      exit(EXIT_FAILURE);
    }
    g_ptr_array_add(args, find_program_in_path_or_as_is(argv[1]));
    start_from= 2;
  }
  for (i= start_from; argv[i]; i++) {
    g_ptr_array_add(args, argv[i]);
  }
  g_ptr_array_add(args, NULL); /* emulate strv */
  return (gchar**)(args->pdata);
}


int main(const int argc, const char **argv)
{
  DBusGProxy *sandboxer_proxy;
  GError *error = NULL;
  gboolean ret;
  gchar *process_path;
  gchar *process_unique_name;
  char **sandboxed_argv;
  gint exit_status= 256;
  
  g_type_init();

  if (!g_thread_supported ())
    g_thread_init (NULL);
  dbus_g_thread_init ();
  
  
#if 0 //just for testing
  g_assert(argv[0]);
  {
    char **ps= sandboxed_argv;
    guint i;
    
    for (i= 0; (int)(ps[i]); i++) {
      g_debug("argv[%d]= %s", i, ps[i]);
    }    
  }
#endif
  sandboxed_argv= setup_argv((gchar**)argv);
    
  connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
  if (NULL == connection) {
    g_error("could not get connection to DBUS Session: %s", error->message);
  }

  sandboxer_proxy = dbus_g_proxy_new_for_name (connection,
					       "com.intel.sandboxer.SandboxService",
					       "/com/intel/sandboxer/SandboxService",
					       "com.intel.sandboxer.SandboxService");

  if (NULL == sandboxer_proxy) {
    g_error("Could not get sandboxer_proxy for "
	    "com.intel.sandboxer.SandboxService");
  }
  
  ret= com_intel_sandboxer_SandboxService_get_process(sandboxer_proxy, 
						      sandboxed_argv[0], 
						      &process_path, 
						      &secret_cookie, 
						      &process_unique_name,
						      &error);
  if (FALSE == ret) {
    g_error("Could not run SandboxerService.GetProcess method: %s", 
	    error->message);
  }
  //g_debug("process unique name is: %s (%s)", process_unique_name, __FUNCTION__);


  process_proxy = dbus_g_proxy_new_for_name (connection,
					     process_unique_name,
					     process_path,
					     "com.intel.sandboxer.Process");
  if (NULL == process_proxy) {
    g_error("Could not get proxy for com.intel.sandboxer.Process (%s): %s", 
	    process_path);
  }
  

  needkill= TRUE;
  if (SIG_ERR == signal(SIGTERM, cleanup)) {
    g_error("could not set up signal handler: %s", g_strerror(errno));
  }
  if (SIG_ERR == signal(SIGINT, cleanup)) {
    g_error("could not set up signal handler: %s", g_strerror(errno));
  }

  dbus_g_proxy_add_signal(process_proxy, "terminated", 
			  G_TYPE_INT, G_TYPE_INVALID);
  dbus_g_proxy_connect_signal(process_proxy, "terminated", 
			      G_CALLBACK(terminated_signal_handler), 
			      &exit_status, NULL);
  
  
  ret= com_intel_sandboxer_Process_run (process_proxy, 
					secret_cookie, 
					(const char**)sandboxed_argv, 
					&error);
  if (FALSE == ret) {
    g_error("Could not run Process.Run method: %s", error->message);
  }
  
  /* wait for signal to hit */
  loop= g_main_loop_new(NULL, FALSE);
  g_main_loop_run(loop);
  
  return WEXITSTATUS(exit_status);
}
