/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <daemon/sandboxer-helper.h>
#include <daemon/base.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/types.h>
#include <sched.h>
#include <ftw.h>
#include <errno.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <stdio.h>
#include <daemon/fs_fixup.h>

//although the syscall is defined, the prototype isn't in headers
//so, defining as extern as does libvert.  util-linux uses #define
//with hard-coded syscall.
extern int pivot_root(const char *new_root, const char *put_old);


static void
sandboxer_base_init(SandboxerBase *self);

static void
sandboxer_base_class_init (SandboxerBaseClass *klass);

static void
sandboxer_base_finalize (GObject *object);

static int 
ftw_callback(const char *path, const struct stat *stat_buf, int flags, struct FTW *ftw_struct);

static gboolean
sandboxer_base_clean_jail_dir(SandboxerBase *self, GError **error);

static gboolean
make_and_mount_dirs(const gchar **dirs, GError **error);

static void
dump_dirs_in_jail(gchar *string, gint *i); //FIXME: just for debug


GQuark
sandboxer_base_error_quark (void);


G_DEFINE_ABSTRACT_TYPE(SandboxerBase, sandboxer_base, G_TYPE_OBJECT);




GQuark
sandboxer_base_error_quark (void)
{
  return g_quark_from_static_string ("sandboxer-base-error-quark");
}


static void
sandboxer_base_init(SandboxerBase *self)
{
  SandboxerBaseClass *klass;
  GRand* rand;
  guint random_number;
  gboolean collide;
  guint i;
  
  /* make the jail_dir */  
  klass= SANDBOXER_BASE_GET_CLASS(self); g_assert(klass);
  g_assert(klass->global_config);
  rand= g_rand_new(); g_assert(rand);

  do {
    random_number= g_rand_int(rand);
    self->jail_dir= g_strdup_printf("%s/sandbox-%u",
				    klass->global_config->jail_dir_path,
				    random_number);
    collide= g_file_test(self->jail_dir, G_FILE_TEST_EXISTS);
    if (collide) g_free(self->jail_dir);
  } while(collide);

  if (!sandboxer_base_make_jail_dir(self)) 
    g_error("couldn't make %s", self->jail_dir);
  g_rand_free(rand);

  /* make the base set of dirs */
  self->dirs_in_jail= g_ptr_array_sized_new(g_strv_length(klass->global_config->dirs_in_jail));
  for (i= 0;
       klass->global_config->dirs_in_jail[i];
       i++) {
    g_ptr_array_add(self->dirs_in_jail, 
		    g_strdup(klass->global_config->dirs_in_jail[i]));
  }
}

static void
sandboxer_base_class_init (SandboxerBaseClass *klass)
{
  GObjectClass *g_klass = G_OBJECT_CLASS (klass);


  g_debug("base_class_init");
  klass->config_store= sandbox_config_store_new(NULL);   g_assert(klass->config_store);
  klass->global_config= sandbox_global_config_new(NULL); g_assert(klass->global_config);

  klass->clean_jail_dir= sandboxer_base_clean_jail_dir;
  klass->make_and_mount_dirs= make_and_mount_dirs;

  g_klass->finalize= sandboxer_base_finalize;
}


static void
sandboxer_base_finalize (GObject *object)
{
  SandboxerBase *self;
  SandboxerBaseClass *klass;
  
  self= (SandboxerBase*)object; g_assert(self);
  klass= SANDBOXER_BASE_GET_CLASS(self); g_assert(klass);
  g_free(self->jail_dir); self->jail_dir= NULL;
  g_free(self->dbus_path); self->dbus_path= NULL;
  if (G_OBJECT_CLASS (sandboxer_base_parent_class)->finalize)
    G_OBJECT_CLASS (sandboxer_base_parent_class)->finalize (object);
  here();
}

gboolean 
sandboxer_base_instance_caller(SandboxerBase *self, 
			       const SandboxerPluginHookStage stage,
			       GError **error)
{
  SandboxerBaseClass *klass;
  guint i;
  
  g_assert(self);
  
  klass= SANDBOXER_BASE_GET_CLASS(self);
  g_assert(klass);

  for (i= 0; i < plugins->len; i++) {
    SandboxerPlugin *plugin;
    
    plugin= g_ptr_array_index(plugins, i);
    if (!klass->plugin_instance_hook_caller(self, plugin->desc, stage, error)) {
      g_message("stage %d of plugin %s returned error", stage, plugin->desc->plugin_name);
      return FALSE;
    }
  }
  return TRUE;
}

gboolean
sandboxer_base_class_caller(SandboxerBaseClass *klass,
			    GError **error)
{
  guint i;

  g_assert(klass);
  
  for (i= 0; i < plugins->len; i++) {
    SandboxerPlugin *plugin;
    
    plugin= g_ptr_array_index(plugins, i);
    if (!klass->plugin_class_hook_caller(klass, plugin->desc, error)) {
      g_debug("class hook returned error");
      return FALSE;
    }
    here();
  }
  return TRUE;

}


static gboolean
make_and_mount_dirs(const gchar **dirs, GError **error)
{
  gchar **sp;
  gchar *dir;
  
  for (sp= (char**)dirs; *(sp); sp++) {    
    dir= g_strdup_printf("old_root/%s", *sp);

    if (g_mkdir_with_parents(*(sp), 
			     S_IRWXU | S_IXGRP | S_IRGRP | S_IXOTH | S_IROTH))  { 
      g_set_error(error, SANDBOXER_BASE_ERROR, SANDBOXER_BASE_ERROR_MOUNT,
		  "could not make basedir %s: %s", *sp, g_strerror(errno));
      return FALSE;
    }
    if(mount(dir, *(sp), NULL, MS_BIND | MS_REC | MS_RDONLY, NULL)) {
      g_set_error(error, SANDBOXER_BASE_ERROR, SANDBOXER_BASE_ERROR_MOUNT,
		  "could not mount --bind %s %s: %s", dir, *(sp), g_strerror(errno));
      return FALSE;
    }
    g_free(dir); 
  }      
  return TRUE;
}

gboolean
sandboxer_base_make_jail_dir(SandboxerBase *self)
{
  SandboxerBaseClass *klass;
  char *old_root;
  gboolean ret= FALSE;

  g_assert(self);
  klass= SANDBOXER_BASE_GET_CLASS(self); g_assert(klass);

  old_root= g_strdup_printf("%s/old_root", self->jail_dir); 
  
  do {
    if (g_mkdir_with_parents(self->jail_dir, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)) break; 
    if (g_mkdir(old_root, S_IXUSR | S_IWUSR)) break; 
    push_cred(0, 0);
    if (chown(self->jail_dir, 0, 0)) { pop_cred(); break; }
    pop_cred();
    g_free(old_root);
    ret= TRUE; g_debug("sandboxer_base_make_jail_dir: OK");
  }
  while(0);
  return ret;  
}

gboolean
sandboxer_base_unshare_through_umount(SandboxerBase *self, 
				      const gchar **dirs,
				      GError **error)
{
  SandboxerBaseClass *klass;
  gchar *old_root;
  GError *merror= NULL;

  g_assert(self);
  klass= SANDBOXER_BASE_GET_CLASS(self); g_assert(klass);
  g_assert(self->jail_dir); g_assert(dirs);
  push_cred(0, 0);
  do {    
    if (!sandboxer_base_instance_caller(self,
					PRE_UNSHARE,
					&merror)) break;
    
    if (unshare(CLONE_NEWNS)) {
      g_set_error(&merror, SANDBOXER_BASE_ERROR,
		  SANDBOXER_BASE_ERROR_FAILED,
		  "failed to unshare: %s", g_strerror(errno));
      break;
    }
    
    /* bind mount the jail_dir to make it pivot_root able */
    if (mount(self->jail_dir, self->jail_dir, NULL, MS_BIND, NULL)) {
      g_set_error(&merror, SANDBOXER_BASE_ERROR,
		  SANDBOXER_BASE_ERROR_MOUNT,
		  "could not mount --bind %s %s: %s", 
		  self->jail_dir, self->jail_dir,
		  g_strerror(errno));
      break;
    }
    if (!sandboxer_base_instance_caller(self,
					PRE_PIVOT_ROOT,
					&merror)) break;
        
    old_root= g_strdup_printf("%s/old_root", self->jail_dir);
    if (pivot_root(self->jail_dir, old_root)) {
      g_set_error(&merror, SANDBOXER_BASE_ERROR,
		  SANDBOXER_BASE_ERROR_FAILED,
		  "failed to pivot_root: %s", g_strerror(errno));
      break;
    }
    if (chdir("/")) {
      g_set_error(&merror, SANDBOXER_BASE_ERROR,
		  SANDBOXER_BASE_ERROR_FAILED,
		  "chdir(/): %s", g_strerror(errno));
      break;
    }
    
    if(!klass->make_and_mount_dirs(dirs, &merror)) break;
    

    if (!sandboxer_base_instance_caller(self,
					PRE_UMOUNT,
					&merror)) break;
    
    if (umount2("/old_root", MNT_DETACH)) {
      g_set_error(&merror, SANDBOXER_BASE_ERROR,
		  SANDBOXER_BASE_ERROR_FAILED,
		  "umount(/old_root): %s", g_strerror(errno));
      break;
    } 
  } while(0);
  pop_cred();
  if (merror) { 
    g_propagate_error(error, merror);
    g_error_free(merror);
    return FALSE;
  }
  return TRUE;
}


static int 
ftw_callback(const char *path, const struct stat *stat_buf, int flags, struct FTW *ftw_struct)
{
  return remove(path);
}


static gboolean
sandboxer_base_clean_jail_dir(SandboxerBase *self, GError **error)
{
  g_assert(self);
  
  if (!(self->jail_dir)) return TRUE;
    
  push_cred(0, 0);
  nftw(self->jail_dir, ftw_callback, 
       100, FTW_DEPTH | FTW_PHYS);  
  pop_cred();

  return TRUE;
}

static void
dump_dirs_in_jail(gchar *string, gint *i)
{
  g_assert(i);
  
  g_debug("dirs[%d]= %s", *i, string);
  *i= *i + 1;
}

void
sandboxer_base_dump_dirs(SandboxerBase *self)
{
  gint i= 0;

  g_assert(self);

  g_ptr_array_foreach(self->dirs_in_jail,
		      (GFunc)dump_dirs_in_jail,
		      &i);  
}
