/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#ifndef    _BASE_H_
#define    _BASE_H_

#include <config.h>
#include <glib-object.h>
#include <dbus/dbus-glib.h>
#include <daemon/config_store.h>
#include <daemon/plugin_impl.h>


#define SANDBOXER_TYPE_BASE            (sandboxer_base_get_type ())
#define SANDBOXER_BASE(object)         (G_TYPE_CHECK_INSTANCE_CAST ((object), SANDBOXER_TYPE_BASE, SandboxerBase))
#define SANDBOXER_BASE_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), SANDBOXER_TYPE_BASE, SandboxerBaseClass))
#define SANDBOXER_IS_BASE(object)      (G_TYPE_CHECK_INSTANCE_TYPE ((object), SANDBOXER_TYPE_BASE))
#define SANDBOXER_IS_BASE_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), SANDBOXER_TYPE_BASE))
#define SANDBOXER_BASE_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), SANDBOXER_TYPE_BASE, SandboxerBaseClass))

#define SANDBOXER_BASE_ERROR sandboxer_base_error_quark ()

G_BEGIN_DECLS

/**
 * GError types for #SandboxerBase
 */
typedef enum {
  SANDBOXER_BASE_ERROR_MOUNT, /** there was an error in bind mounting a directory */
  SANDBOXER_BASE_ERROR_FAILED /** generic error */
} SandboxerBaseError;


typedef struct _SandboxerBase SandboxerBase;
typedef struct _SandboxerBaseClass SandboxerBaseClass;

/**
 * Base instance-object for #SandboxerProcess and #SandboxerSandbox
 */
struct _SandboxerBase {
  /** the GObject that this object is derived from */
  GObject parent; 
  
  /** the dbus_path of this object */
  gchar *dbus_path; 

  /** the directory where the sandbox is constructed */
  gchar *jail_dir; 

  /** the directories that are to be mounted in the sandbox */
  GPtrArray *dirs_in_jail; 
};

/**
 * Base class of #SandboxerProcess and #SandboxerSandbox
 */
struct _SandboxerBaseClass {
  /** the GObjectClass that this class is derived from */
  GObjectClass parent; 

   /** the DBusGConnection used */
  DBusGConnection *connection;
  
  /**  
   * the #SandboxConfigStore to use for retrieving .package files
   */
  SandboxConfigStore *config_store; 

  /** the #SandboxGlobalConfig to use */
  SandboxGlobalConfig *global_config; 

  /** function to call to clean up the jail upon process exit */
  gboolean (*clean_jail_dir)(SandboxerBase *self, GError **error); 
  
  /** 
   * function to call to make the mount points and bind mount the
   * originals in old_root in the sandbox
   */
  gboolean (*make_and_mount_dirs)(const gchar **dirs, GError **error); 

  /**
   * function that sandboxer_base_instance_caller() should call when
   * iterating through #SandboxerPluginDescriptor%s
   */
  gboolean (*plugin_instance_hook_caller)(SandboxerBase *self,
					  SandboxerPluginDescriptor *desc,
					  const SandboxerPluginHookStage stage,
					  GError **error); 

  /**
   * function that sandboxer_base_class_caller() should call when
   * iterating through #SandboxerPluginDescriptor%s
   */
  gboolean (*plugin_class_hook_caller)(SandboxerBaseClass *klass,
				       SandboxerPluginDescriptor *desc,
				       GError **error); 
				       
};

GType sandboxer_base_get_type(void);

/**
 *  calls plugin hooks as defined in _SandboxPluginDescriptor in
 *  either _SandboxerPluginDescriptor.process_hooks or
 *  _SandboxerPluginDescriptor.sandbox_hooks.  This iterates through
 *  #plugins and calls _SandboxerBaseClass.plugin_instance_hook_caller
 *  on each plugin
 *  
 *  \param[in]  self    instance of the #SandboxerBase
 *  \param[in]  stage   #SandboxerPluginHookStage to call
 *  \param[out] error   return location for a GError or NULL
 *  \returns TRUE for success, FALSE for failure
 */
gboolean 
sandboxer_base_instance_caller(SandboxerBase *self, 
			       const SandboxerPluginHookStage stage,
			       GError **error);

/**
 * calls class hooks as defined in _SandboxerPluginDescriptor in
 * either _SandboxerPluginDescriptor.sandbox_class_init or
 * _SandboxerPluginDescriptor.process_class_init.  This iterates
 * through #plugins and calls
 * _SandboxerBaseClass.plugin_class_hook_caller
 *
 *  \param[in]  klass   klass pointer
 *  \param[out] error   return location for a GError or NULL
 *  \returns TRUE for success, FALSE for failure
 */
gboolean
sandboxer_base_class_caller(SandboxerBaseClass *klass,
			    GError **error);


/**
 * perform the unshare through umount stages of sandbox creation.
 *
 * \param[in] self    a #SandboxerBase
 *
 * \param[in] dirs   a NULL terminated list of strings to pass call
 * #SandboxerBaseClass.make_and_mount_dirs with
 *
 * \param[out] error a location for a GError
 *
 * \returns TRUE for success; FALSE for failure
 */
gboolean
sandboxer_base_unshare_through_umount(SandboxerBase *self, 
				      const gchar **dirs, GError **error);

/**
 * creates the jail dir and
 * jail_dir/old_root.  The jail dir created is the one in self
 *
 * \param[in] self   a #SandboxerBase object
 * 
 * \returns TRUE for success; FALSE for failure
 */
gboolean
sandboxer_base_make_jail_dir(SandboxerBase *self);
				      

//FIXME: just for debug
void
sandboxer_base_dump_dirs(SandboxerBase *self);


G_END_DECLS


#endif  /* _BASE_H_ */
