/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <daemon/config_store.h>
#include <sys/types.h>
#include <pwd.h>
#include <glib.h>
#include <glib/gstdio.h>

#include <daemon/sandboxer-helper.h> //for test only

#define GLOBAL_CONFIG_GROUP                  "Sandboxerd"

typedef struct _GlobalConfigTuple GlobalConfigTuple;
struct _GlobalConfigTuple {
  const gchar *key_file_key;
  const gchar *env_key;
  const gchar *default_value;
};

const GlobalConfigTuple enabled_plugins_tuple= {
  "EnabledPlugins",
  "SANDBOXER_ENABLED_PLUGINS",
  DEFAULT_ENABLED_PLUGINS
};

const GlobalConfigTuple plugins_path_tuple= {
  "PluginsPath",
  "SANDBOXER_PLUGINS_PATH",
  DEFAULT_PLUGINS_PATH
};

const GlobalConfigTuple sandbox_user_tuple= {
  "SandboxerUser",
  "SANDBOXER_USER",
  DEFAULT_SANDBOXER_USER
};

const GlobalConfigTuple package_file_tuple= {
  "PackageFilePath",
  "SANDBOXER_PACKAGE_FILE_PATH",
  DEFAULT_PACKAGE_FILE_PATH
};

const GlobalConfigTuple jail_dir_path_tuple= {
  "JailDirPath",
  "SANDBOXER_JAIL_DIR_PATH",
  DEFAULT_JAIL_DIR_PATH
};

const GlobalConfigTuple pid_file_path_tuple= {
  "PidFilePath",
  "SANDBOXER_PID_FILE_PATH",
  DEFAULT_PID_FILE_PATH
};

const GlobalConfigTuple lib_exec_path_tuple= {
  "LibExecPath",
  "SANDBOXER_LIB_EXEC_PATH",
  DEFAULT_LIB_EXEC_PATH
};

const GlobalConfigTuple dirs_in_jail_tuple= {
  "DirsInJail",
  "SANDBOXER_DIRS_IN_JAIL",
  DEFAULT_DIRS_IN_JAIL
};

const GlobalConfigTuple log_level_tuple = {
  "LogLevel",
  "SANDBOXER_LOG_LEVEL",
  DEFAULT_LOG_LEVEL
};


static int 
compare_sandbox_config_by_package_name(const SandboxConfig* sconfig,
				       const gchar* package_name);

static guint
g_key_file_count_groups(GKeyFile *key_file);

#if 0
static gboolean
env_string_to_boolean(const gchar *string);

static gboolean
get_boolean_config_value(GKeyFile *keyfile, const GlobalConfigTuple *tuple);
#endif

static gchar* 
get_string_config_value(GKeyFile *keyfile, const GlobalConfigTuple *tuple);

static gchar**
get_string_list_config_value(GKeyFile *keyfile, const GlobalConfigTuple *tuple);

static GLogLevelFlags 
get_log_level_config_value(GKeyFile *keyfile, const GlobalConfigTuple *tuple);


static gboolean
add_file(SandboxConfigStore *self,
	 const gchar *path, GError **error);

static gboolean
user_list_to_uid_list(const gchar **user_list, const gsize num_users, 
		      gint **p_uids, GError **error);

static gchar**
uid_list_to_user_list(const gint *uid_list, const gsize num_uids);

static GLogLevelFlags
string_to_g_log_level_flags(const gchar *input);

static void
ptr_array_destroy(gpointer ptr_array);




GQuark
sandboxer_config_store_error_quark (void)
{
  return g_quark_from_static_string ("sandboxer-config-store-error-quark");
}


static guint
g_key_file_count_groups(GKeyFile *key_file)
{
  gchar **groups;
  gsize size;

  g_assert(key_file);
  groups= g_key_file_get_groups(key_file, &size);
  g_strfreev(groups);
  
  return (guint)(size);  
}

SandboxConfig *
sandbox_config_new()
{
  return g_new0(SandboxConfig, 1);
}

void
sandbox_config_free(SandboxConfig *self)
{
  if (self) {
    gsize i;

    g_free(self->package_name); self->package_name= NULL;
    g_free(self->sandbox_name); self->sandbox_name= NULL;
    g_free(self->uids); self->uids= NULL;
    for (i= 0; i < self->num_exec_paths; i++)
      g_free(self->exec_paths[i]);
    g_free(self->package_file_path);

    g_free(self);
  }
}

SandboxConfig* 
sandbox_config_new_from_file(const gchar *path, GError **error)
{
  GKeyFile *keyfile;  
  GError *merror= NULL;
  SandboxConfig *sconfig;
  gchar **user_list;

  do {
    sconfig= sandbox_config_new(); g_assert(sconfig);
    sconfig->package_file_path= g_strdup(path);

    keyfile= g_key_file_new();
    if (!g_key_file_load_from_file(keyfile, path, 
				   G_KEY_FILE_NONE, &merror)) break;

    sconfig->package_name= g_key_file_get_string(keyfile, 
						 PACKAGE_FILE_SANDBOX_GROUP,
						 PACKAGE_FILE_PACKAGE_NAME_KEY,
						 &merror);

    if (merror) break;
    user_list= g_key_file_get_string_list(keyfile,
					  PACKAGE_FILE_SANDBOX_GROUP,
					  PACKAGE_FILE_USER_LIST_KEY,
					  &(sconfig->num_uids),
					  &merror);
    if (merror) break;
    if (!user_list_to_uid_list((const gchar**)user_list, 
			       sconfig->num_uids, &(sconfig->uids), &merror)) break;
    g_strfreev(user_list);
    
    sconfig->sandbox_name= g_key_file_get_string(keyfile, 
						 PACKAGE_FILE_SANDBOX_GROUP,
						 PACKAGE_FILE_SANDBOX_NAME_KEY,
						 &merror);
    if (merror) break;
    sconfig->exec_paths= g_key_file_get_string_list(keyfile,
						    PACKAGE_FILE_SANDBOX_GROUP,
						    PACKAGE_FILE_EXEC_PATHS_KEY,
						    &(sconfig->num_exec_paths),
						    &merror);

    if (merror) break;
    
    if (g_key_file_count_groups(keyfile) > 1)
      sconfig->extra_config_groups= TRUE;
  } while(0);
  g_key_file_free(keyfile);  
  if (merror){
    g_propagate_error(error, merror);
    sandbox_config_free(sconfig);      	    
    return NULL;
  }
  return sconfig;
}

static gboolean
user_list_to_uid_list(const gchar **user_list, const gsize num_users, 
		      gint **p_uids, GError **error)
{
  struct passwd *ppasswd;
  gsize i;
  
  g_assert(user_list);
  g_assert(p_uids);

  *p_uids= g_new(gint, num_users);

  for (i= 0; i < num_users; i++) {
    ppasswd= getpwnam(user_list[i]);
    if (!ppasswd) {
      g_set_error(error, 
		  SANDBOXER_CONFIG_STORE_ERROR,
		  SANDBOXER_CONFIG_STORE_ERROR_FAILED,
		  "%s isn't in passwd file", user_list[i]);
      return FALSE;
    }
    (*p_uids)[i]= ppasswd->pw_uid;
  }
  return TRUE;  
}

static gchar**
uid_list_to_user_list(const gint *uid_list, const gsize num_uids) 
{
  struct passwd *ppasswd;
  gchar **ret;
  gsize i;

  g_assert(uid_list);
  ret= g_new(gchar*, num_uids + 1);
  for (i= 0; i < num_uids; i++) {
    ppasswd= getpwuid(uid_list[i]);
    g_assert(ppasswd);
    ret[i]= g_strdup(ppasswd->pw_name);   
  }
  ret[num_uids]= NULL;
  return ret;
}

gboolean
sandbox_config_to_file(const SandboxConfig *self, 
		       const gchar *comment,
		       GError **error)
{
  GKeyFile *keyfile;
  gchar *data;
  gsize data_len;
  gsize bytes_written;
  GIOChannel *io;
  gboolean ret;
  gchar **user_list;
  
  g_assert(self);
  g_assert(self->package_file_path);
  keyfile= g_key_file_new();
  if (comment) {
    g_key_file_set_comment(keyfile,
			   NULL,
			   NULL,
			   comment,
			   NULL);
  }
  g_key_file_set_string(keyfile,
			PACKAGE_FILE_SANDBOX_GROUP,
			PACKAGE_FILE_SANDBOX_NAME_KEY,
			self->sandbox_name);
  g_key_file_set_string(keyfile,
			PACKAGE_FILE_SANDBOX_GROUP,
			PACKAGE_FILE_PACKAGE_NAME_KEY,
			self->package_name);
  user_list= uid_list_to_user_list(self->uids, self->num_uids);
  g_key_file_set_string_list(keyfile,
			     PACKAGE_FILE_SANDBOX_GROUP,
			     PACKAGE_FILE_USER_LIST_KEY,
			     (const gchar**)user_list,
			     self->num_uids);
  g_strfreev(user_list);
  g_key_file_set_string_list(keyfile,
			     PACKAGE_FILE_SANDBOX_GROUP,
			     PACKAGE_FILE_EXEC_PATHS_KEY,
			     (const gchar**)(self->exec_paths),
			     self->num_exec_paths);
  ret= FALSE;
  do {
    data= g_key_file_to_data(keyfile,
			     &data_len,
			     error);
    if (NULL == data) break;
    io= g_io_channel_new_file(self->package_file_path,
			      "w",
			      error);
    if (NULL == io) break;
    if (G_IO_STATUS_NORMAL != g_io_channel_write_chars(io,
						       data,
						       data_len,
						       &bytes_written,
						       error)) break;  
    
    if (G_IO_STATUS_NORMAL != g_io_channel_shutdown(io,
						    TRUE,
						    error)) break;
    ret= TRUE;
  } while(0);
  return ret;
}		   

static gboolean
add_file(SandboxConfigStore *self,
	 const gchar *path, GError **error)
{
  SandboxConfig *sconfig;

  g_assert(self);
  g_assert(path);
  
  sconfig= sandbox_config_new_from_file(path, error);

  if (sconfig)    
    {
      gsize i;
      GPtrArray *sandbox_configs;
      
      
      for (i= 0; i < sconfig->num_exec_paths; i++)
	g_hash_table_insert(self->path_hash, sconfig->exec_paths[i], sconfig);
      
      /*
       * note that for the sandbox hash, we must duplicate the sandbox name
       * string.  Because any package can be removed, we cannot rely on a 
       * the sconfig->sandbox_name pointer not belonging to the removed package
       */

      sandbox_configs= g_hash_table_lookup(self->sandbox_sandbox_configs,
					   sconfig->sandbox_name);
      if (sandbox_configs) {	  
	g_ptr_array_add(sandbox_configs, sconfig);
      } else {	  
	sandbox_configs= g_ptr_array_new(); g_assert(sandbox_configs);
	g_ptr_array_add(sandbox_configs, sconfig);
	g_hash_table_insert(self->sandbox_sandbox_configs,
			    g_strdup(sconfig->sandbox_name),
			    sandbox_configs);
	
      }

      self->sandbox_configs= g_slist_prepend(self->sandbox_configs, 
					    (gpointer)sconfig);

      g_debug("adding: %s", sconfig->package_name);
      
      
      return TRUE;
    }
  return FALSE;
}

gboolean
sandbox_config_store_add_file(SandboxConfigStore *self,
			      const gchar *path, GError **error)
{
  SandboxGlobalConfig *global_config;
  gchar *filename;
  gboolean ret;
  GError *merror= NULL;


  g_assert(self);
  g_assert(path);

  global_config= sandbox_global_config_new(NULL);
  g_assert(global_config);

  ret= FALSE;
  do {
    SandboxConfig *sconfig;

    sconfig= sandbox_config_new_from_file(path, &merror); 
    if (NULL == sconfig) break;
    
    sconfig->extra_config_groups= FALSE; /* extra groups will not be copied over */
    
    //TODO: basic sanity check of sconfig
    filename= g_strdup_printf("%s/%s.package", 
			      global_config->package_file_directory,
			      sconfig->package_name); 
    g_debug("filename=\"%s\"", filename);
    g_free(sconfig->package_file_path);
    sconfig->package_file_path= filename;
    if (g_file_test(filename, G_FILE_TEST_EXISTS)) break; 
    if (!sandbox_config_to_file(sconfig, "imported", &merror)) break;        
    if (!add_file(self, filename, &merror)) break;

    ret= TRUE;
  } while(0);
  if (!ret) { 
    g_set_error(error,
		SANDBOXER_CONFIG_STORE_ERROR,
		SANDBOXER_CONFIG_STORE_ERROR_FAILED,
		"error loading %s: %s", 
		path, merror ? merror->message : "UNKNOWN");
    if (merror) { 
      g_error_free(merror);    
    }
  }
  return ret;
}

static void
ptr_array_destroy(gpointer ptr_array)
{
  g_ptr_array_free(ptr_array, FALSE);
}


SandboxConfigStore* 
sandbox_config_store_new(const gchar *path)
{
  static SandboxConfigStore *self= NULL;
  GDir *dir;
  gchar *dent;

  if (self) {
    return self;
  }
  g_assert(path);
  self= g_new0(SandboxConfigStore, 1);
  g_assert(self);

  self->sandbox_sandbox_configs= g_hash_table_new_full(g_str_hash,
						       g_str_equal,
						       g_free,
						       ptr_array_destroy);
  self->path_hash= g_hash_table_new(g_str_hash,
				   g_str_equal);

  dir= g_dir_open(path, 0, NULL);
  g_return_val_if_fail(dir, NULL);
  while ((dent = (gchar*)g_dir_read_name(dir))) {
    
    if (g_str_has_suffix(dent, PACKAGE_FILE_SUFFIX))  {
      gchar *package_file_path;
      GError *error= NULL;

      package_file_path= g_strdup_printf("%s/%s", path, dent); 
      if (!add_file(self, package_file_path, &error)) {
	g_debug("couldn't add .package file \"%s\" to database: %s", package_file_path,
		error ? error->message : (gchar*)error);
	g_error_free(error);
      }
      g_free(package_file_path);
    } else {
      g_message("%s is not a %s file", dent, PACKAGE_FILE_SUFFIX);
    }
  }
  g_dir_close(dir);
  
  return self;
}


SandboxConfig* 
sandbox_config_store_lookup_path(SandboxConfigStore *self,
			   const gchar *path)
{
  g_assert(self);
  g_assert(path);

  return (SandboxConfig*)g_hash_table_lookup(self->path_hash, (gpointer)path);
}

guint
sandbox_config_store_lookup_num_sandboxes(SandboxConfigStore *self,
					  const gchar *sandbox_name)
{
  GPtrArray *sandbox_configs;

  g_assert(self);
  g_assert(sandbox_name);

  sandbox_configs= (GPtrArray*)g_hash_table_lookup(self->sandbox_sandbox_configs,
						   (gconstpointer)sandbox_name);
  if (NULL == sandbox_configs)
    return 0;
  return sandbox_configs->len;
}
				     
static int 
compare_sandbox_config_by_package_name(const SandboxConfig* sconfig,
				       const gchar* package_name)
{

  g_assert(package_name);
  g_assert(sconfig);
  return  g_strcmp0(package_name, sconfig->package_name);
}

gboolean
sandbox_config_store_remove_package(SandboxConfigStore *self,
			      const gchar *package_name, GError **error)
{
  GSList* item;

  g_assert(self);
  g_assert(package_name);
  
  item= g_slist_find_custom(self->sandbox_configs, 
			    package_name,
			    (GCompareFunc)compare_sandbox_config_by_package_name);
  
  if (item) {
    gsize i;
    SandboxConfig *sconfig;
    GPtrArray *sandbox_configs;

    g_assert(item->data);
    sconfig= item->data;

    sandbox_configs= g_hash_table_lookup(self->sandbox_sandbox_configs,
				 sconfig->sandbox_name);
    g_assert(sandbox_configs);

    if (1 == sandbox_configs->len) {
      g_hash_table_remove(self->sandbox_sandbox_configs,
			  sconfig->sandbox_name);      
    } else {
      if (!g_ptr_array_remove_fast(sandbox_configs, sconfig)) {
	g_error("sandbox %s is not part of sandbox %s?!", sconfig->sandbox_name,
		sconfig->sandbox_name);
      }
    }

    
    for (i= 0; i < sconfig->num_exec_paths; i++) {
      gboolean ret;

      ret= g_hash_table_remove(self->path_hash, sconfig->exec_paths[i]);
      g_assert(ret);
    }
    self->sandbox_configs= g_slist_remove_link(self->sandbox_configs, item);
    
    if (g_unlink(sconfig->package_file_path))
      g_warning("unable to unlink %s", sconfig->package_file_path);

    sandbox_config_free(sconfig);
  }
  else {
    g_set_error(error, SANDBOXER_CONFIG_STORE_ERROR,
		SANDBOXER_CONFIG_STORE_ERROR_FAILED,
		"%s is not a known package", package_name);
    
    return FALSE;
  }
  return TRUE;
}

#if 0 //This was used earlier but no longer used
static gboolean
env_string_to_boolean(const gchar *string)
{
  const char *trues[]=  { "TRUE", "T", "YES", "Y", "1", NULL};
  const char *falses[]= {"FALSE", "F",  "NO", "N", "0", NULL};
  guint i;
  
  g_assert(string);
  for (i= 0; trues[i]; i++) {
    if (0 == g_strcasecmp(string, trues[i]))
      return TRUE;
  }
  for (i= 0; falses[i]; i++) {
    if (0 == g_strcasecmp(string, falses[i]))
      return FALSE;
  }
  g_error("%s is neither true nor false", string);
}

static gboolean
get_boolean_config_value(GKeyFile *keyfile, const GlobalConfigTuple *tuple)
{
  gchar *env_value;
  GError *error= NULL;
  gboolean ret;
  
  env_value = (gchar*)g_getenv(tuple->env_key);
  if (env_value)
    return env_string_to_boolean(env_value);
  
  ret= g_key_file_get_boolean(keyfile,
			      GLOBAL_CONFIG_GROUP,
			      tuple->key_file_key,
			      &error);
  if (!error)
    return ret;
  return env_string_to_boolean(tuple->default_value);
}
#endif

static gchar* 
get_string_config_value(GKeyFile *keyfile, const GlobalConfigTuple *tuple)
{
  gchar *env_value;
  gchar *config_file_value;

  g_assert(tuple);
    
  env_value = (gchar*)g_getenv(tuple->env_key);
  if (env_value) 
    return g_strdup(env_value); 
  config_file_value = g_key_file_get_string(keyfile,
					    GLOBAL_CONFIG_GROUP,
					    tuple->key_file_key,
					    NULL);
  if (config_file_value)
    return config_file_value;
  return (gchar*)tuple->default_value;  
}

static gchar**
get_string_list_config_value(GKeyFile *keyfile, const GlobalConfigTuple *tuple)
{
  gchar *env_value;
  gchar **config_file_value;

  g_assert(tuple);
    
  env_value = (gchar*)g_getenv(tuple->env_key);
  if (env_value) {
    return g_strsplit(env_value, ":", 0);
  }
  config_file_value= g_key_file_get_string_list(keyfile,
						GLOBAL_CONFIG_GROUP,
						tuple->key_file_key,
						NULL,
						NULL);
  if (config_file_value)
    return config_file_value;
  return g_strsplit(tuple->default_value, ";", 0);
}

//returns a mask of log levels to eat
static GLogLevelFlags
string_to_g_log_level_flags(const gchar *input)
{

  g_assert(input);
  if (0 == g_strcasecmp("DEBUG", input)) return 0;
  if (0 == g_strcasecmp("NORMAL", input)) 
    return G_LOG_LEVEL_DEBUG;
  if (0 == g_strcasecmp("QUIET", input)) 
    return G_LOG_LEVEL_DEBUG | G_LOG_LEVEL_INFO | G_LOG_LEVEL_MESSAGE;
  return 0;
}

static GLogLevelFlags 
get_log_level_config_value(GKeyFile *keyfile, const GlobalConfigTuple *tuple)
{
  gchar *env_value;
  gchar *config_file_value;

  g_assert(tuple);
    
  env_value = (gchar*)g_getenv(tuple->env_key);
  if (env_value) 
    return string_to_g_log_level_flags(env_value);
  config_file_value = g_key_file_get_string(keyfile,
					    GLOBAL_CONFIG_GROUP,
					    tuple->key_file_key,
					    NULL);
  if (config_file_value)
    return string_to_g_log_level_flags(config_file_value);
  return string_to_g_log_level_flags(tuple->default_value);  
}

SandboxGlobalConfig* 
sandbox_global_config_new(const gchar *path)
{
  static SandboxGlobalConfig *self= NULL;
  GKeyFile *keyfile;
  GError *error= NULL;
  struct passwd *ppasswd;
  gchar *sandboxer_user;

  if (self)
    return self;
  
  g_assert(path);
  self= g_new0(SandboxGlobalConfig, 1); g_assert(self);
  keyfile= g_key_file_new(); g_assert(keyfile);
  g_key_file_load_from_file(keyfile, path, G_KEY_FILE_NONE, &error);
  g_return_val_if_fail(NULL == error, NULL);    
  
  self->global_config_file_path= g_strdup(path);
  
  self->enabled_plugins= get_string_list_config_value(keyfile, &enabled_plugins_tuple);
  
  self->plugins_path= get_string_config_value(keyfile, &plugins_path_tuple);
  g_return_val_if_fail(self->plugins_path, NULL);

  self->package_file_directory= get_string_config_value(keyfile, &package_file_tuple);
  g_return_val_if_fail(self->package_file_directory, NULL);  

  sandboxer_user= get_string_config_value(keyfile, &sandbox_user_tuple);
  g_assert(sandboxer_user);  
  ppasswd= getpwnam(sandboxer_user);
  g_return_val_if_fail(ppasswd, NULL);
  self->sandbox_user_uid= ppasswd->pw_uid;
  self->sandbox_user_gid= ppasswd->pw_gid;

  here();
  g_debug("sandboxer_uid is %d", ppasswd->pw_uid);
  g_debug("sandboxer_gid is %d", ppasswd->pw_gid);
  

  self->pid_file_path= get_string_config_value(keyfile, &pid_file_path_tuple);
  g_return_val_if_fail(self->pid_file_path, NULL);

  self->jail_dir_path= get_string_config_value(keyfile, &jail_dir_path_tuple);
  g_return_val_if_fail(self->jail_dir_path, NULL);

  self->lib_exec_path= get_string_config_value(keyfile, &lib_exec_path_tuple);
  g_return_val_if_fail(self->lib_exec_path, NULL);

  self->dirs_in_jail= get_string_list_config_value(keyfile, &dirs_in_jail_tuple);
  g_return_val_if_fail(self->dirs_in_jail, NULL);

  self->log_level= get_log_level_config_value(keyfile, &log_level_tuple);  

  g_key_file_free(keyfile);
  
  return self;
}

#if 0 //save for later
gboolean
sandbox_global_config_to_file(SandboxGlobalConfig *self,
			      const gchar *comment,
			      const gchar **extra_groups,
			      const gchar **extra_keys,
			      const gchar **extra_values,
			      GError **error)
{
  const gchar *default_comment= "this is an programatically generated sandboxerd.conf file";
  GKeyFile *keyfile;
  gboolean ret;
  GError *merror= NULL;
  
  g_assert(self); g_assert(self->global_config_file_path);
  
  keyfile= g_key_file_new(); g_assert(keyfile);
  do {
    gchar *keyfile_data;
    gsize keyfile_length;
    struct passwd *sandboxer_passwd;
    guint i;

    ret= FALSE;
    
    if (!g_key_file_set_comment(keyfile, NULL, NULL, 
				comment ? comment : default_comment, 
				&merror)) break;
    g_key_file_set_string_list(keyfile, 
			       GLOBAL_CONFIG_GROUP,
			       enabled_plugins_tuple.key_file_key,
			       (const gchar **)(self->enabled_plugins),
			       g_strv_length(self->enabled_plugins));
    g_key_file_set_string(keyfile,
			  GLOBAL_CONFIG_GROUP,
			  plugins_path_tuple.key_file_key,
			  self->plugins_path);

    sandboxer_passwd= getpwuid(self->sandbox_user_uid);
    if (!sandboxer_passwd) break;

    g_key_file_set_string(keyfile,
			  GLOBAL_CONFIG_GROUP,
			  sandbox_user_tuple.key_file_key,
			  sandboxer_passwd->pw_name); 
    here();
    g_key_file_set_string(keyfile,
			  GLOBAL_CONFIG_GROUP,
			  package_file_tuple.key_file_key,
			  self->package_file_directory);
    here();
    g_key_file_set_string(keyfile,
			  GLOBAL_CONFIG_GROUP,
			  jail_dir_path_tuple.key_file_key,
			  self->jail_dir_path);
    here();
    g_key_file_set_string(keyfile,
			  GLOBAL_CONFIG_GROUP,
			  pid_file_path_tuple.key_file_key,
			  self->pid_file_path);
    here();
    g_key_file_set_string(keyfile,
			  GLOBAL_CONFIG_GROUP,
			  lib_exec_path_tuple.key_file_key,
			  self->lib_exec_path);
    here();
    g_key_file_set_string(keyfile,
			  GLOBAL_CONFIG_GROUP,
			  pid_file_path_tuple.key_file_key,
			  self->pid_file_path);
    here();
    g_key_file_set_string_list(keyfile,
			       GLOBAL_CONFIG_GROUP,
			       dirs_in_jail_tuple.key_file_key,
			       (const gchar**)(self->dirs_in_jail),
			       g_strv_length(self->dirs_in_jail));    
    here();
    {//log level
      gchar *value;
      
      switch(self->log_level) {
      case G_LOG_LEVEL_DEBUG:
	value= "NORMAL";
	break;
	
      case G_LOG_LEVEL_DEBUG | G_LOG_LEVEL_INFO | G_LOG_LEVEL_MESSAGE:
	value= "QUIET";
	
      default:
	value= "DEBUG";
	break;
      }
      here();
      g_key_file_set_string(keyfile,
			    GLOBAL_CONFIG_GROUP,
			    log_level_tuple.key_file_key,
			    value);
      here();
    }
    
    if (extra_groups) {
      for (i= 0; extra_groups[i]; i++) {
	g_debug("i= %d", i);
	g_key_file_set_value(keyfile,
			     extra_groups[i],
			     extra_keys[i],
			     extra_values[i]);
      }
    }
    here();
    keyfile_data= g_key_file_to_data(keyfile, &keyfile_length, &merror);
    if (!keyfile_data) break;
    here();
    if (!g_file_set_contents(self->global_config_file_path, keyfile_data, keyfile_length, &merror)) break;    
  } while(0);
  if (!ret) {
    if (merror) {
      g_propagate_error(error, merror);
      g_error_free(merror);
    } else {
      g_set_error(error, 
		  SANDBOXER_CONFIG_STORE_ERROR, 
		  SANDBOXER_CONFIG_STORE_ERROR_FAILED, 
		  "unknown error");
    }    
  }
  g_key_file_free(keyfile);
  here();
  return ret;
}
#endif

SandboxConfig**
sandbox_config_store_lookup_sandbox(SandboxConfigStore *self,
				    const gchar *sandbox_name,
				    guint *length) //FIXME: untested
{
  GPtrArray *array;

  g_assert(self); g_assert(sandbox_name);
  array= g_hash_table_lookup(self->sandbox_sandbox_configs,
			     sandbox_name);
  if (array) {
    if (length) *length= array->len;
    return (SandboxConfig**)(array->pdata);
  }
  if (length) *length= 0;
  return NULL;
}
