/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#ifndef __CONFIG_STORE_H__
#define __CONFIG_STORE_H__

#include <config.h>
#include <glib.h>
#include <sys/types.h>

#ifndef DEFAULT_PLUGINS_PATH
#define DEFAULT_PLUGINS_PATH "/usr/lib/sandboxer/plugins"
#endif

#ifndef DEFAULT_ENABLED_PLUGINS
#define DEFAULT_ENABLED_PLUGINS "*"
#endif

#ifndef DEFAULT_SANDBOXER_USER
#define DEFAULT_SANDBOXER_USER "sandboxer"
#endif

#ifndef DEFAULT_PACKAGE_FILE_PATH
#define DEFAULT_PACKAGE_FILE_PATH "/usr/share/sandboxer/packages"
#endif

#ifndef DEFAULT_JAIL_DIR_PATH
#define DEFAULT_JAIL_DIR_PATH  "/var/run/sandboxer"
#endif

#ifndef DEFAULT_PID_FILE_PATH
#define DEFAULT_PID_FILE_PATH "/var/run/sandboxer/sandboxerd.pid"
#endif

#ifndef DEFAULT_LIB_EXEC_PATH
#define DEFAULT_LIB_EXEC_PATH "/usr/libexec/sandboxer/"
#endif

#ifndef DEFAULT_DIRS_IN_JAIL
#define DEFAULT_DIRS_IN_JAIL "/bin;/usr;/lib;/opt;/tmp;/var;/etc;/dev;/proc;/sys"
#endif

#ifndef DEFAULT_LOG_LEVEL
#define DEFAULT_LOG_LEVEL "NORMAL"
#endif

#define PACKAGE_FILE_SUFFIX           ".package"


#define PACKAGE_FILE_SANDBOX_GROUP    "Sandbox"
#define PACKAGE_FILE_SANDBOX_NAME_KEY "SandboxName"
#define PACKAGE_FILE_PACKAGE_NAME_KEY "PackageName"
#define PACKAGE_FILE_USER_LIST_KEY    "Users"
#define PACKAGE_FILE_EXEC_PATHS_KEY   "ExecPaths"

#define SANDBOXER_CONFIG_STORE_ERROR sandboxer_config_store_error_quark ()

G_BEGIN_DECLS

typedef struct _SandboxConfig SandboxConfig;
typedef struct _SandboxConfigStore SandboxConfigStore;
typedef struct _SandboxGlobalConfig SandboxGlobalConfig;

typedef enum {
  SANDBOXER_CONFIG_STORE_ERROR_FAILED
}SandboxerConfigStoreError;




/**
 * a representation of one package's sandbox config
 *
 */
struct _SandboxConfig {
  /** the path to the config file representing this package */
  gchar *package_file_path;  

  /** the name of this package */
  gchar *package_name; 

  /** the name of the sandbox the executables in this .package file belong in */
  gchar *sandbox_name; 

  /** an array of uids that this package may run in */
  gint *uids; 
  
  /** the number of uids in uids */
  gsize num_uids; 
  
  /** an array of exectable paths that this package provides */
  gchar **exec_paths; 
  
  /** the number of strings in exec_paths */
  gsize num_exec_paths; 

  /** TRUE if the config_file for this package has additional config
      information */
  gboolean extra_config_groups; 
};

/**
 * make a new #SandboxConfig
 *
 * \returns  #SandboxConfig
 */
SandboxConfig *
sandbox_config_new();

/**
 *  free a #SandboxConfig
 *
 * \param[in] self  a #SandboxConfig
 */
void
sandbox_config_free(SandboxConfig *self);

/**
 * create a new #SandboxConfig from a .package file
 *
 * \param[in]  path    the pathanem to a .package file to read
 * \param[out] error    a location to return an error or NULL
 *
 * \returns a newly allocated #SandboxConfig. Free with sandbox_config_free()
 */
SandboxConfig*
sandbox_config_new_from_file(const gchar *path, GError **error);

/**
 * write a #SandboxConfig to a .package file Note that even if
 * extra_groups is TRUE, the extra_gropus will not be written
 *
 * \param[in]  self      a #SandboxConfig
 * \param[in]  comment   the comment to put at the begining of the .package file
 * \param[out] error     a location to return an error
 *
 * \returns TRUE for success, FALSE for failure
 */
gboolean
sandbox_config_to_file(const SandboxConfig *self, 
		       const gchar *comment,
		       GError **error);


struct _SandboxConfigStore {
  GSList *sandbox_configs;
  GHashTable *sandbox_sandbox_configs;
  GHashTable *path_hash;
};

/**
 *  Get the instance of the #SandboxConfigStore
 *
 *  \param[in] path   the path to the .package file database 
 * 
 *  \note This call acts a singleton.  No matter how many times you call it
 *        after the first time, you get the pointer to the same object.  This 
 *        happens regardless of the input parameter.
 *
 *  \returns a pointer to the #SandboxConfigStore that is created
 */
SandboxConfigStore* 
sandbox_config_store_new(const gchar *path);

/**
 *  Get a #SandboxConfig for the path of an executable.  This path is part of the
 *  ExecPaths key given in a .package file
 *
 *  \param[in] self  the #SandboxConfigStore
 *  \param[in] path  the path to lookup
 * 
 *  \returns a #SandboxConfig for the path or NULL if there is none
 */
SandboxConfig* 
sandbox_config_store_lookup_path(SandboxConfigStore *self,
				 const gchar *path);

/**
 *  Lookup the number of .package files that share the same sandbox name
 *
 *  \param[in] self           the #SandboxConfigStore
 *  \param[in] sandbox_name   the name of the sandbox (SandboxName in the .package file)
 * 
 *  \returns the count of the number of .package files that share sandbox_name
 */
guint
sandbox_config_store_lookup_num_sandboxes(SandboxConfigStore *self,
					  const gchar *sandbox_name);
/**
 * gives a list of SandboxConfigs that run in the sandbox name given
 * in sandbox_name
 *
 * \param[in]  self           a #SandboxConfigStore
 * \param[in]  sandbox_name   the name of the sandbox to look up as given in a .package file
 * \param[out] length         a location to return the length of the list or NULL
 * \returns an array of pointers to the SandboxConfigs.  These pointers are private to 
 *          the #SandboxConfigStore and should not be freed
 */
SandboxConfig**
sandbox_config_store_lookup_sandbox(SandboxConfigStore *self,
				    const gchar *sandbox_name,
				    guint *length);
				     
/**
 * add a .package file to the #SandboxConfigStore.  If the .package
 * file could not be added, error is set with a
 * #SandboxerConfigStoreError
 *
 * \param[in]  self   a #SandboxConfigStore
 * \param[in]  path   the path to a .package file
 * \param[out] error  return location for a GError or NULL
 *
 * \returns TRUE on success, FALSE on failure
 */
gboolean
sandbox_config_store_add_file(SandboxConfigStore *self,
			      const gchar *path, GError **error);


/**
 * Remove a package from a #SandboxConfigStore 
 * as well as the on-disk database
 *
 * \param[in]  self            a #SandboxConfigStore
 * \param[in]  package_name    the name of the package to remove
 * \param[out] error           a location to return an error of NULL
 *
 * \returns TRUE for success; FALSE for failure
 */
gboolean
sandbox_config_store_remove_package(SandboxConfigStore *self,
				    const gchar *package_name, GError **error);

/**
 * a representation of the global sandboxer config
 *
 * \note no matter how you call it after the first time, you get the first
 *       global config
 */
struct _SandboxGlobalConfig {
  /** path to the sandboxerd.conf that this represents */
  gchar *global_config_file_path;
  
  /** uid sandboxerd should run as */
  uid_t sandbox_user_uid;

  /** gid that sandboxerd should run as */
  gid_t sandbox_user_gid;

  /** NULL terminated list of plugins that should be loaded */
  gchar **enabled_plugins;

  /** directory to search for plugins in */
  gchar *plugins_path;

  /** directory to look for .package files in */
  gchar *package_file_directory;

  /** directory to construct jails in */
  gchar *jail_dir_path;
  
  /** full path of pid file for sandboxerd */
  gchar *pid_file_path;

  /** directory to look for stubd in */
  gchar *lib_exec_path;

  /** NULL terminated list of directories to put in sandboxes */
  gchar **dirs_in_jail;

  /** logging level (see GLogLevelFlags) */
  GLogLevelFlags log_level;

  /** TRUE if there are addtional groups beyond the Sandboxerd group.
      Useful for plugins */
  gboolean extra_config_groups;
};

/**
 * Acts as a singleton factory for the global sandboxer config.  In
 * the first calling, path is used to specify the location of
 * .package files.  In subsequent invocations, the pointer value is
 * ignored and the single instance is returned.
 *
 * \param[in] path   path to a directory of .package files or NULL if one has
 * already been defined.
 *
 * \returns   a #SandboxGlobalConfig on success or NULL on error
 */
SandboxGlobalConfig* 
sandbox_global_config_new(const gchar *path);

#if 0
/**
 * write a #SandboxGlobalConfig to a file.  Used for testing
 *
 * \param[in]  comment         comment to put at top of file
 *
 * \param[in]  extra_groups    a NULL terminated array of strings containing any
 *                             extra groups to write in the config file
 *
 * \param[in]  extra_keys      a NULL terminated array of strings containing any
 *                             extra keys to write in the config file
 *
 * \param[in]  extra_values    a NULL terminated array of strings containing the
 *                             values of the extra keys
 *
 * \param[out] error           location of a GError or NULL
 *
 * \note the number of elements in extra_groups, extra_values, and
 * extra_keys must be the same
 *
 * \returns TRUE for success; FALSE for failure
 *
 */
gboolean
sandbox_global_config_to_file(SandboxGlobalConfig *self,
			      const gchar *comment,
			      const gchar **extra_groups,
			      const gchar **extra_keys,
			      const gchar **extra_values,
			      GError **error);

#endif

G_END_DECLS

#endif 
