/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

/** 
 * \file plugin.h
 * This file should be included in all plugins and provides the
 * required datastructures and functions to create a plugin.
 *
 * \mainpage Sandboxer Plugin Documentation 
 *
 * \section intro_section Introduction
 * This documenation is designed for plugin writers.  The plugin
 * system provides several hooks to interject plugin code into the
 * workings of sandboxer.  Before describing them in detail, let's
 * examine the underlying process of sandboxing and the hook points.
 *
 * \section theory_of_operation Theory Of Operation
 * For the purpose of illustration, we will give a simplified example
 * of the sandboxing process.  The code below is analagous to the
 * actual code, but not the same code.
 *
 * \code
 * if (fork()) {
 *   chdir("/jail");
 *   unshare(CLONE_NEWNS);
 *   mount("/jail", "/jail", NULL, MS_BIND, NULL);
 *   pivot_root("/jail", "/jail/old_root");
 *   chdir("/");
 *   mount("/old_root/bin", "bin", NULL, MS_BIND, NULL);
 *   mount("/old_root/usr", "usr", NULL, MS_BIND, NULL);
 *   mount("/old_root/lib", "lib", NULL, MS_BIND, NULL);
 *   umount2("/old_root", MNT_DETACH);
 *   setgid(gid); 
 *   setuid(uid);
 *   exec(application);
 * }
 * \endcode
 * 
 * The above segment typifies the case where there is one executable
 * per sandbox.  The actual code is in sandboxer_pocess_run().  In the
 * case where there are multiple executables per sandbox, then the
 * flow is a little more complicated.
 *
 * \code
 * if (fork()) {
 *   chdir("/jail");
 *   unshare(CLONE_NEWNS);
 *   mount("/jail", "/jail", NULL, MS_BIND, NULL);
 *   pivot_root("/jail", "/jail/old_root");
 *   chdir("/");
 *   mount("/old_root/bin", "bin", NULL, MS_BIND, NULL);
 *   mount("/old_root/usr", "usr", NULL, MS_BIND, NULL);
 *   mount("/old_root/lib", "lib", NULL, MS_BIND, NULL);
 *   umount2("/old_root", MNT_DETACH);
 *   while (1) {
 *     application= wait_for_request();
 *     if (fork()) {
 *       setgid(gid); 
 *       setuid(uid);
 *       exec(application);
 *     }
 *   }
 * }
 * \endcode
 *
 * the above segment is simplified greatly, but gives the overall
 * semantics.  In actuality, sandboxer_service_get_process() will
 * vfork/exec stubd with g_spawn_async.  Stubd will instantiate a
 * #SandboxerSandbox.
 *
 * There are several hook points that are provided during the process
 * that a plugin can hook into.  There are class lifecycle hooks, such
 * as #_SandboxerPluginDescriptor.process_class_init_hook and
 * #_SandboxerPluginDescriptor.sandbox_class_init_hook.
 *
 * Additionally, there are two arrays of #SandboxerProcessHookFuncs
 * and #SandboxerSandboxHookFuncs.  These are called in the various
 * "phases" of sandboxing.  Since both #SandboxerProcessClass and
 * #SandboxerSandboxClass inherit from #SandboxerBaseClass, the phases
 * of PRE_UNSHARE through PRE_UMOUNT and covered in
 * sandboxer_base_unshare_through_umount().  
 *
 * The structure holding the hook function pointers is
 * #_SandboxerPluginDescriptor.  Each plugin should define a
 * #_SandboxerPluginDescriptor, which should be passed to
 * sandboxer_register_plugin() the a plugin's g_module_init or
 * g_module_check_init function.  All the #SandboxerPluginDescriptors
 * are held in a global array and the hook functions are iterated
 * through for each plugin for a particular stage.  That is, a plugin
 * has no control over the order in the list of plugins that its hook
 * gets called.  
 *
 * \section example_section A Simple Plugin Example
 * In our example, we will create a simple example plugin.
 * \code
 * #include <plugin.h>
 * 
 * static gboolean
 * process_class_init(_SandboxerProcessClass *klass, GError **self)
 * {
 *   //insert your plugin code here 
 * }
 *
 * static gboolean
 * process_instance_init(_SandboxerProcess *self, GError **self) 
 * {
 *  //insert your plugin code here
 * }
 *
 * static gboolean
 * process_pre_fork(_SandboxerProcess *self, GError **self) 
 * {
 *  //insert your plugin code here
 * }
 *
 * static gboolean
 * process_pre_unshare(_SandboxerProcess *self, GError **self) 
 * {
 *  //insert your plugin code here
 * }
 *
 * static gboolean
 * process_pre_pivot_root(_SandboxerProcess *self, GError **self) 
 * {
 *  //insert your plugin code here
 * }
 *
 * static gboolean
 * process_pre_umount(_SandboxerProcess *self, GError **self) 
 * {
 *  //insert your plugin code here
 * }
 *
 * static gboolean
 * process_pre_setuid(_SandboxerProcess *self, GError **self) 
 * {
 *  //insert your plugin code here
 * }
 *
 * static gboolean
 * process_pre_exec(_SandboxerProcess *self, GError **self) 
 * {
 *  //insert your plugin code here
 * }
 *
 * static gboolean
 * process_finalize(_SandboxerProcess *self, GError **self) 
 * {
 *  //insert your plugin code here
 * }
 *
 * static gboolean
 * sandbox_instance_init(_SandboxerSandbox *self, GError **self) 
 * {
 *  //insert your plugin code here 
 * }
 *
 * static gboolean
 * sandbox_pre_fork(_SandboxerSandbox *self, GError **self) 
 * {
 *  //insert your plugin code here 
 * }
 *
 * static gboolean
 * sandbox_pre_unshare(_SandboxerSandbox *self, GError **self) 
 * {
 *  //insert your plugin code here 
 * }
 *
 * static gboolean
 * sandbox_pre_pivot_root(_SandboxerSandbox *self, GError **self) 
 * {
 *  //insert your plugin code here 
 * }
 *
 * static gboolean
 * sandbox_pre_umount(_SandboxerSandbox *self, GError **self) 
 * {
 *  //insert your plugin code here 
 * }
 *
 * static gboolean
 * sandbox_finalize(_SandboxerSandbox *self, GError **self) 
 * {
 *  //insert your plugin code here 
 * }
 * 
 * static gboolean 
 * sandbox_class_init(_SandboxerSandboxClass *klass, GError **error) 
 * { 
 *   //insert your plugin code here 
 * }
 *
 * static void
 * close_hook(SandboxerPlugin *plugin) 
 * {
 *  //insert your plugin code here 
 * }
 *
 * static const SandboxerPluginDescriptor plugin_descriptor= {
 *   .plugin_name= "example",
 *   .major_version= PLUGIN_MAJOR_VERSION,
 *   .minor_version= PLUGIN_MINOR_VERSION,
 *   .process_class_init_hook= process_class_init,
 *   .process_hooks= {
 *     process_instance_init,
 *     process_pre_fork,
 *     process_pre_unshare,
 *     process_pre_pivot_root,
 *     process_pre_umount,
 *     process_pre_setuid,
 *     process_pre_exec,
 *     process_finalize},
 *   .sandbox_class_init_hook= sandbox_class_init,
 *   .sandbox_hooks= {
 *     sandbox_instance_init,
 *     sandbox_pre_fork,
 *     sandbox_pre_unshare,
 *     sandbox_pre_pivot_root,
 *     sandbox_pre_umount,
 *     NULL,  //PRE_SETUID never called
 *     NULL,  //PRE_EXEC never called
 *     sandbox_finalize
 *     },
 *   .close_hook= close_hook,
 *   .padding= PADDING_INIT
 *  };
 *   
 *  const gchar* g_module_check_init(GModule *module)
 *  {
 *    if (sandboxer_register_plugin(&plugin_descriptor, module)) {
 *      g_debug("plugin loaded"); 
 *    } else {
 *     g_error("plugin failed to load");
 *    }
 *    return NULL;
 *  }
 * \endcode
 *
 * \section further_reading Further Reading
 * It is highly recommended that one look at some of the plugins in
 * the plugins directory of the source distribution to get a feel for
 * how to use a plugin.  The following data structures and files are
 * likely to be of help:
 *  \see _SandboxGlobalConfig
 *  \see _SandboxConfig
 *  \see _SandboxerPluginDescriptor
 *
 */

#ifndef __PLUGIN_H__
#define __PLUGIN_H__

#include <config.h>
#include <glib.h>
#include <gmodule.h>

#define PADDING_SIZE 4
#define PADDING_INIT {NULL}
#define PLUGIN_MAJOR_VERSION 0
#define PLUGIN_MINOR_VERSION 1

G_BEGIN_DECLS

/* forward declarations  */
struct _SandboxerProcess;
struct _SandboxerProcessClass;
struct _SandboxerSandboxClass;
struct _SandboxerSandbox;



extern GPtrArray *plugins;

/**  
 * Used as array indexes into
 * #SandboxerPluginDescriptor.process_hooks and
 * #SandboxerPluginDescriptor.sandbox_hooks
 * 
 * \see sandboxer_process_run()
 * \see sandboxer_sandbox_new()
 */

typedef enum _SandboxerPluginHookStage {
  /** 
   * Called in the instance init of #SandboxerProcess 
   * Runs as SandboxerUser in sandboxerd.conf
   */
  INSTANCE_INIT, 

  /** 
   * Runs in the sandboxerd process in the non-shared sandbox case,
   * but in the stubd process in the shared-sandbox case.  Runs as
   * SandboxerUser in sandboxerd.conf
   * 
   * \see sandboxer_process_run()
   */
  PRE_FORK, 
  
  /** 
   * Called just before calling unshare(CLONE_NEWNS). Runs as root 
   * 
   * \see sandboxer_base_unshare_through_umount()
   **/
  PRE_UNSHARE, 

  /** 
   * called just before pivot_root.  Runs as root 
   *
   * \see sandboxer_base_unshare_through_umount()
   */
  PRE_PIVOT_ROOT,
 
  /** 
   * Called just before umounting old_root.  Runs as root 
   *
   * \see sandboxer_base_unshare_through_umount()
   */
  PRE_UMOUNT, 
  
  /** 
   * Called just before dropping privileges.  Runs as root
   *
   * \see sandboxer_process_child_setup_function()
   */
  PRE_SETUID,

  /** 
   * Called just before exiting (and thus exec'ing).  Runs as user
   * defined in package file. 
   *
   * \see sandboxer_process_child_setup_function()
   */
  PRE_EXEC, 
  
  /** 
   * Called in the instance finalize.  Runs as SandboxerUser 
   * 
   * \see sandboxer_process_finalize()
   * \see sandboxer_sandbox_finalize()
   */
  INSTANCE_FINALIZE,
  
  /** The number of hooks that there are (used for indexing) */
  HOOK_COUNT 
} SandboxerPluginHookStage;

  
typedef struct _SandboxerPluginDescriptor SandboxerPluginDescriptor;
typedef struct _SandboxerPlugin SandboxerPlugin;

typedef gboolean (*SandboxerProcessHookFunc) (struct _SandboxerProcess *self,
					      GError **error);
typedef gboolean (*SandboxerSandboxHookFunc) (struct _SandboxerSandbox *self,
					      GError **error);

typedef gboolean (*SandboxerProcessClassHookFunc) (struct _SandboxerProcessClass *klass, GError **error);
typedef gboolean (*SandboxerSandboxClassHookFunc) (struct _SandboxerSandboxClass *klass, GError **error);

typedef void     (*SandboxerPluginCloseFunc) (SandboxerPlugin *plugin);

/** 
 * Holds the pointers that define what a plugin does.  This structure
 * should be passed to sandboxer_register_plugin()
 *
 **/
struct _SandboxerPluginDescriptor {
  
  /** name of this plugin */
  const gchar *plugin_name;   


  /** 
   *  minor version of sandboxer that this plugin is compiled against
   *  \see PLUGIN_MINOR_VERSION 
   */
  const gint minor_version; 

  /** 
   *  major version of sandboxer that this plugin is compiled
   *  against 
   *  \see PLUGIN_MAJOR_VERSION
   */
  const gint major_version; 

  /** a hook function to be called during sandboxer_process_class_init() */
  SandboxerProcessClassHookFunc       process_class_init_hook; 

  /** 
   * an array of hook functions called during phases of the
   * #SandboxerProcess. 
   *
   * \see #SandboxerPluginHookStage
   */
  SandboxerProcessHookFunc            process_hooks[HOOK_COUNT];

  /** a hook function to be called during the sandbox_class_init() */
  SandboxerSandboxClassHookFunc       sandbox_class_init_hook; 

  /** 
   * an array of hook functions called during the phase of
   * #SandboxerSandbox 
   * 
   * \note In this case, only INSTANCE_INIT, PRE_FORK, PRE_UNSHARE,
   * PRE_PIVOT_ROOT, PRE_UMOUNT, and INSTANCE_FINALIZE will be
   * called.  PRE_SETUID and PRE_EXEC are called in the
   * #SandboxerProcess context.
   * 
   * \see #SandboxerPluginHookStage
   */
  SandboxerSandboxHookFunc            sandbox_hooks[HOOK_COUNT]; 

  /** a hook function to be called upon exit */
  SandboxerPluginCloseFunc            close_hook; 

  /** padding reserved for future versions */
  gpointer padding[PADDING_SIZE];
};

/**
 * Registers a plugin to be used.  Implementers should call this 
 * function in their g_module_init or g_module_check_init function
 * 
 * \param[in] desc    a #SandboxerPluginDescriptor that describes this plugin
 * \param[in] module   the GModule passed in by g_module_init
 *
 * \returns TRUE if the plugin was initialized succesfully; FALSE otherwise
 */
gboolean
sandboxer_register_plugin(const SandboxerPluginDescriptor *desc, GModule *module); 


G_END_DECLS

#endif /* __PLUGIN_H__ */
