/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <daemon/plugin_impl.h>
#include <daemon/config_store.h>
#include <daemon/sandboxer-helper.h>
#include <sys/types.h>
#include <unistd.h>

GPtrArray *plugins;

void
plugin_system_init(void)
{
  plugins= g_ptr_array_new();
}

static gboolean
is_plugin_enabled(SandboxGlobalConfig *global_config, const gchar *plugin_name)
{
  guint i;
  gchar **enabled_plugins;

  g_assert(global_config); g_assert(plugin_name);
  enabled_plugins= global_config->enabled_plugins;

  for (i= 0; enabled_plugins[i]; i++) {
    if (g_str_equal(enabled_plugins[i], plugin_name)) {
      g_debug("plugin: \"%s\" enabled due to match in config file", plugin_name);
      return TRUE;
    }
    if (g_str_equal(enabled_plugins[i], "*")) {
      g_debug("plugin: \"%s\" enabled due to wildcard in config file", plugin_name);
      return TRUE;
    }
  }
  g_debug("plugin: \"%s\" disabled due to no match in config file", plugin_name);
  return FALSE;
}

void
load_plugins(void) 
{
  GDir *plugin_dir;
  const gchar *dent;
  GError *error= NULL;
  SandboxGlobalConfig *global_config;

  global_config= sandbox_global_config_new(NULL); 
  g_assert(global_config);
  
  g_debug("plugin_dir= %s", global_config->plugins_path);
  plugin_dir=  g_dir_open (global_config->plugins_path,
			   0,
			   &error);
  g_assert(plugin_dir);
  g_assert(NULL == error);
  while ((dent= g_dir_read_name(plugin_dir))) {
    if (g_str_has_suffix(dent, ".so")) { 
      GModule* plugin;
      gchar *plugin_path;
      
      g_debug("%s is a plugin", dent);
      if (!is_plugin_enabled(global_config, dent))
	continue;
      plugin_path= g_strdup_printf("%s/%s", global_config->plugins_path, dent);
      g_assert(plugin_path);
      plugin= g_module_open(plugin_path, 0);
      if (plugin) {
	g_debug("%s loads", dent);
      } else {
	g_debug("%s does not load: %s", dent, g_module_error());
      }
      g_free(plugin_path);
    }
  }
}

gboolean
sandboxer_register_plugin(const SandboxerPluginDescriptor *desc, 
			  GModule *module)
{
  SandboxerPlugin *plugin;

  g_assert(desc); g_assert(module);
  if (PLUGIN_MAJOR_VERSION == desc->major_version)  {
    plugin= g_new(SandboxerPlugin, 1); g_assert(plugin);
    plugin->desc= (SandboxerPluginDescriptor*)desc; 
    plugin->module= module;
    g_ptr_array_add(plugins, (gpointer)plugin);
  } else {
    g_debug("not loading %s plugin compiled with major version %d",
	    desc->plugin_name,
	    desc->major_version);
    return FALSE;
  }
  return TRUE;
}

