/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <daemon/sandboxer-helper.h>
#include <daemon/process.h>
#include <daemon/plugin.h>
#include <daemon/plugin_impl.h>
#include <daemon/config_store.h>
#include <daemon/prctl_fixup.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/prctl.h>
#include <ftw.h>
#include <pwd.h>
#include <grp.h>
#include <sched.h>
#include <string.h>
#include <errno.h>
#include <gmodule.h>
#include <uuid/uuid.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-bindings.h>

#include <daemon/fs_fixup.h>

G_DEFINE_TYPE(SandboxerProcess, sandboxer_process, SANDBOXER_TYPE_BASE)

#include <daemon/sandboxer-process-glue.h>
#include <daemon/process-marshal.h>

/* prototypes */

static void
sandboxer_process_init (SandboxerProcess *self);

static void
sandboxer_process_class_init (SandboxerProcessClass *klass);

static void
sandboxer_process_finalize(GObject *object);

static void
sandboxer_process_set_property (GObject *obj, guint prop_id,
				const GValue *value,
				GParamSpec *pspec);

static void 
sandboxer_process_get_property (GObject * object, guint prop_id,
				GValue * value,
				GParamSpec * pspec);

static void
process_child_wait_func(GPid pid, gint status, SandboxerProcess *self);

static gboolean 
instance_plugin_hook_caller(SandboxerBase *self,
			    SandboxerPluginDescriptor *desc,
			    const SandboxerPluginHookStage stage,
			    GError **error);
static gboolean
klass_plugin_hook_caller(SandboxerBaseClass *klass,
			 SandboxerPluginDescriptor *desc,
			 GError **error);

     
/* types */

enum {
  PROP_0,
  PROP_PATH,
  PROP_DBUS_PATH,
  PROP_SANDBOX,
  PROP_PID
};

enum {
  NOTRUN,
  RUNNING,
  RAN
};


/* implementations */

GQuark
sandboxer_process_error_quark (void)
{
  return g_quark_from_static_string ("sandboxer-process-error-quark");
}

//FIXME: move away from rand
static void
sandboxer_process_init (SandboxerProcess *self)
{
  SandboxerProcessClass *klass=  SANDBOXER_PROCESS_GET_CLASS(self);
  GError *error = NULL;
  uuid_t uuid;
  GRand* rand;
  guint random_number;
  static guint counter=0;

  g_assert(self);

  rand= g_rand_new();
  uuid_generate(uuid);
  uuid_unparse_upper(uuid, self->secret_cookie);
  g_debug("secret cookie: %s", self->secret_cookie);

  counter++;
  random_number= g_rand_int(rand);
  self->parent.dbus_path= g_strdup_printf("%s_%u", 
					  SANDBOXER_DBUS_PROCESS_PATH, 
					  random_number);
  g_rand_free(rand);

  self->status = NOTRUN; 
  self->env= g_ptr_array_new();

  

  /* plugins */
  if (!sandboxer_base_instance_caller(SANDBOXER_BASE(self),
				      INSTANCE_INIT,
				      &error)) {
    g_error("%s", error->message);
  }  

  g_debug("process: registering dbus path \"%s\"", self->parent.dbus_path);
  dbus_g_connection_register_g_object(klass->parent.connection, 
				      self->parent.dbus_path, 
				      G_OBJECT(self));
  
}

static void
sandboxer_process_class_init (SandboxerProcessClass *klass)
{
  const gchar* signalNames[E_SIGNAL_COUNT] = {
    SIGNAL_RUN,
    SIGNAL_TERMINATED
  };
  GObjectClass *g_klass = G_OBJECT_CLASS (klass);
  GError *error= NULL;
  GParamSpec *pspec;
  
  g_assert(klass);
  g_assert(could_become_root());

  g_klass->set_property = sandboxer_process_set_property;
  g_klass->get_property = sandboxer_process_get_property;

  pspec= g_param_spec_string( "process-path",
			      "path",
			      "path of binary to execute",
			      "/bin/false",
			      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
  g_object_class_install_property (g_klass, PROP_PATH, pspec);  

  pspec= g_param_spec_string( "process-dbus_path",
			      "dbus_path",
			      "dbus path of process object",
			      NULL,
			      G_PARAM_READABLE);
  g_object_class_install_property (g_klass, PROP_DBUS_PATH, pspec);  

  pspec= g_param_spec_int( "process-pid",
			   "pid",
			   "pid of currently running property",
			   0, /* actually an invalid pid */
			   G_MAXINT32,
			   0, 
			   G_PARAM_READABLE);
  g_object_class_install_property (g_klass, PROP_PID, pspec);  

  pspec= g_param_spec_pointer ("process-sandbox",
			       "sandbox",
			       "pointer to sandbox process is in, if any",
			       G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
  g_object_class_install_property(g_klass, PROP_SANDBOX, pspec);


  klass->parent.connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
  g_assert(klass->parent.connection);

  klass->signals[E_SIGNAL_RUN] =
    g_signal_new(signalNames[E_SIGNAL_RUN],
		 G_OBJECT_CLASS_TYPE(klass),
		 G_SIGNAL_RUN_LAST,
		 0,
		 NULL,
		 NULL,
		 g_cclosure_marshal_VOID__STRING,
		 G_TYPE_NONE,
		 1,
		 G_TYPE_STRING);
  klass->signals[E_SIGNAL_TERMINATED] =
    g_signal_new(signalNames[E_SIGNAL_TERMINATED],
		 G_OBJECT_CLASS_TYPE(klass),
		 G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
		 0,
		 NULL,
		 NULL,
		 g_cclosure_marshal_VOID__INT,
		 G_TYPE_NONE,
		 1,
		 G_TYPE_INT);

  g_klass->finalize= sandboxer_process_finalize;

  /* plugins */
  klass->parent.plugin_instance_hook_caller= instance_plugin_hook_caller;
  klass->parent.plugin_class_hook_caller= klass_plugin_hook_caller;
  here();
  if (!sandboxer_base_class_caller(SANDBOXER_BASE_CLASS(klass), &error)) {
    g_error(error->message);
  }
  
  dbus_g_object_type_install_info (SANDBOXER_TYPE_PROCESS, 
				   &dbus_glib_sandboxer_process_object_info);
}

static void
sandboxer_process_finalize(GObject *object)
{
  SandboxerProcess *self= SANDBOXER_PROCESS(object); 
  SandboxerProcessClass *klass= SANDBOXER_PROCESS_GET_CLASS(self); 
  GError *error = NULL;

  g_assert(self);  g_assert(klass);

  if (RUNNING == self->status) {
    push_cred(0, 0);
    kill(self->pid, SIGTERM);
    g_usleep(G_USEC_PER_SEC / 100);
    kill(self->pid, SIGKILL);
    self->status= RAN;
    pop_cred();
  }

  if (NULL == self->sandbox) {
    /* because process holds a pointer to the sandbox that it is in, 
       base can't tell if the process is in a shared sandbox or not,
       thus we nuke it here */
    SANDBOXER_BASE_CLASS(sandboxer_process_parent_class)->clean_jail_dir(SANDBOXER_BASE(self),
									 NULL);
  }
  
  g_ptr_array_free(self->env, TRUE); 
  
  /* plugins */
  if (!sandboxer_base_instance_caller(SANDBOXER_BASE(self),
				      INSTANCE_FINALIZE,
				      &error)) {
    g_error(error->message);
  }
  
  g_free(self->path);
  bzero(self->secret_cookie, sizeof(self->secret_cookie));

  if (G_OBJECT_CLASS (sandboxer_process_parent_class)->finalize)
    G_OBJECT_CLASS (sandboxer_process_parent_class)->finalize(object); 
}

static void
sandboxer_process_set_property (GObject *obj, guint prop_id,
				const GValue *value,
				GParamSpec *pspec)
{
  SandboxerProcess *self= (SandboxerProcess*)obj;


  g_assert(obj);
  g_assert(value);
  g_assert(pspec);


  switch (prop_id)
    {
    case PROP_PATH:
      self->path= g_value_dup_string(value);                                
      break;

    case PROP_SANDBOX:
      self->sandbox= g_value_get_pointer(value);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, prop_id, pspec);
      break;
    }
}


static void 
sandboxer_process_get_property (GObject * object, guint prop_id,
				GValue * value,
				GParamSpec * pspec)
{
  SandboxerProcess *self = (SandboxerProcess*)object;

  g_assert(object);  g_assert(value);  g_assert(pspec);

  switch (prop_id)
    {
    case PROP_PATH:
      g_value_set_string(value, self->path);
      break;

    case PROP_DBUS_PATH:      
      g_value_set_string(value, self->parent.dbus_path);
      break;

    case PROP_PID:
      g_value_set_int(value, self->pid);
      break;

    case PROP_SANDBOX:
      g_value_set_pointer(value, self->sandbox);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
      break;      
    }
}

#if 0 //just for debug
static void
dump_perms(const char *path)
{
  struct stat statbuf;

  lstat(path, &statbuf);
  g_debug("%s perm= %o", path, statbuf.st_mode & (S_IRWXU | S_IRWXG | S_IRWXO));
}
#endif

/*
 *
 * Do the mechanics of sandboxing here
 *
 */
static void
process_child_setup_function(SandboxerProcess *self)
{
  SandboxerProcessClass   *klass;
  struct passwd           *ppasswd;
  gboolean                 err;
  GError                  *error= NULL;
  gchar                   *old_root= NULL;

  g_assert(self);
  klass= SANDBOXER_PROCESS_GET_CLASS(self); g_assert(klass);

  ppasswd= getpwuid(self->uid);
  if (!ppasswd) g_error("could not find passwd entry for: %d", self->uid);

  if (NULL == self->sandbox) {
    g_ptr_array_add(self->parent.dirs_in_jail,
		    g_strdup(ppasswd->pw_dir));
    g_ptr_array_add(self->parent.dirs_in_jail, NULL);
    sandboxer_base_dump_dirs(SANDBOXER_BASE(self));

    if (!sandboxer_base_unshare_through_umount(SANDBOXER_BASE(self),
					       (const gchar**)(self->parent.dirs_in_jail->pdata),
					       &error)) { 
      g_error("could not set up jail correctly: %s", error->message);
    }
    
    here();
  } 
  do { /* from here onwards is used in studb'd cases as well as normal ones */
    err= TRUE;
    i_am_here();
    push_cred(0, 0);

    if (!sandboxer_base_instance_caller(SANDBOXER_BASE(self),
					PRE_SETUID,
					&error)) break;
    here();
    if(initgroups(ppasswd->pw_name, self->gid)) break;
    if (setresgid(self->gid, self->gid, self->gid)) break; 
    if (setresuid(self->uid, self->uid, self->uid)) break;     
    
    here();
    if (!sandboxer_base_instance_caller(SANDBOXER_BASE(self),
					PRE_EXEC,
					&error)) break;
    here();
    
    err= FALSE;
  }
  while (0);
  g_free(old_root);
  if (err) {
    here();
    /* we really have no choice, but to abort this
     * process, since there is no way to return an error
     * properly
     */
    g_error("could not setup jail properly: %s", error ? error->message : g_strerror(errno));
  }
  /* reset range timer slack to the default */
  prctl(PR_SET_TIMERSLACK, 0, 0, 0, 0);

  g_debug("set up jail correctly");   
}

SandboxerProcess* 
sandboxer_process_new(const gchar *path, gchar **secret_cookie, SandboxerSandbox *sandbox)
{
  SandboxerProcess *self;

  g_assert(path);
  g_assert(secret_cookie);

  self= (SandboxerProcess*)g_object_new(SANDBOXER_TYPE_PROCESS, 
					"process-path", path, 
					"process-sandbox", sandbox, 
					NULL);
  g_assert(self);
  *secret_cookie= g_strdup(self->secret_cookie);

  return self;
}

void 
sandboxer_process_add_to_env(SandboxerProcess *self, gchar *key, gchar *value)
{
  gchar *keyval;
  

  g_assert(self);
  g_assert(key);
  g_assert(value);

  keyval= g_strdup_printf("%s=%s", key, value);
  g_ptr_array_add(self->env, keyval);
}

void
sandboxer_process_add_to_env_pair(SandboxerProcess *self, gchar *string)
{

  g_assert(self);
  g_assert(string);
  g_ptr_array_add(self->env, g_strdup(string));
}

gboolean
setup_env(SandboxerProcess *self)
{
  struct passwd *ppasswd;
  /*
   *  env variables to set up:
   *   PATH
   *   HOME
   *   USER
   */

  g_assert(self);
  
  ppasswd= getpwuid(self->uid);
  if (!ppasswd) return FALSE;

  sandboxer_process_add_to_env(self, "HOME", ppasswd->pw_dir);
  sandboxer_process_add_to_env(self, "USER", ppasswd->pw_name);
  sandboxer_process_add_to_env(self, "PATH", "/usr/local/bin:/usr/bin:/bin"); //FIXME: hard coding
  
  return TRUE;
}


gboolean
sandboxer_process_run (SandboxerProcess *self, 
		       const gchar *secret_cookie,
		       char **argv, 
		       GError **error)
{
  SandboxerProcessClass *klass;
  SandboxGlobalConfig *global_config;
  SandboxConfigStore *config_store;
  gboolean ret;
  gboolean err= TRUE;
  GSpawnFlags spawn_flags;
  gchar *msg= NULL;
  
  here();
  g_assert(self); g_assert(argv);  g_assert(argv[0]);
  klass= SANDBOXER_PROCESS_GET_CLASS(self); g_assert(klass);
  global_config= klass->parent.global_config; g_assert(global_config);
  config_store= klass->parent.config_store; g_assert(config_store);

  if (!g_str_equal(secret_cookie, self->secret_cookie)) {
    g_set_error(error,
		SANDBOXER_PROCESS_ERROR,
		SANDBOXER_PROCESS_ERROR_COOKIE,
		"incorrect cookie");
    return FALSE;
  }

  g_debug("dbus_path: %s", self->parent.dbus_path);

  g_debug("running %s", self->path);
  g_debug("argv is %s", g_strjoinv(", ", argv));
  do {
    //should do in instance init?
    SandboxConfig *sconfig;
    struct passwd *ppasswd;    

    sconfig= sandbox_config_store_lookup_path(config_store,
					      self->path);
    if (NULL == sconfig) { 
      msg= g_strdup_printf("file \"%s\" not listed in .package file", self->path); 
      break; 
    }
    self->uid= sconfig->uids[0];
    if (0 == self->uid) {
      /* refuse to make a root sandbox */
      msg= g_strdup_printf("refusing to make UID 0 sandbox");
      break;  
    } 
    ppasswd= getpwuid(self->uid);
    g_assert(ppasswd);
    self->gid= ppasswd->pw_gid;
    push_cred(self->uid, self->gid);
    if (!g_file_test(self->path, G_FILE_TEST_IS_EXECUTABLE)) {
      msg= g_strdup_printf("file \"%s\" is not executable", self->path);
      i_am_here();
      pop_cred();
      break;
    }
    pop_cred();
    err= FALSE; 
  }
  while(0);
  if (err) {
    g_set_error(error, 
		SANDBOXER_PROCESS_ERROR,
		SANDBOXER_PROCESS_ERROR_SETUP,
		"Error in setting up sandbox environment: %s", msg ? msg : "(null)");
    g_free(msg);
    return FALSE;
  }
  /* create env */
  if (!setup_env(self)) {
    g_set_error(error, 
		SANDBOXER_PROCESS_ERROR,
		SANDBOXER_PROCESS_ERROR_SETUP,
		"Error setting up sandbox environment variables");
    return FALSE; 
  }

  if (!sandboxer_base_instance_caller(SANDBOXER_BASE(self),
				      PRE_FORK,
				      error)) {
    g_prefix_error(error, "pre_fork: ");
    return FALSE;
  }

  /* close the env */
  g_ptr_array_add(self->env, NULL);

  spawn_flags= G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD;
  if (!((global_config->log_level) & G_LOG_LEVEL_DEBUG)) {
    spawn_flags |= G_SPAWN_STDOUT_TO_DEV_NULL | G_SPAWN_STDERR_TO_DEV_NULL;
  }

  ret= g_spawn_async("/",
		     argv,
		     (gchar**)(self->env->pdata),
		     spawn_flags,
		     (GSpawnChildSetupFunc)process_child_setup_function,
		     self,
		     &(self->pid),
		     error);
  
  if (ret) {
    self->status= RUNNING;
    g_signal_emit(self, 
		  SANDBOXER_PROCESS_GET_CLASS(self)->signals[E_SIGNAL_RUN],
		  0,
		  self->parent.dbus_path);
    g_child_watch_add(self->pid, (GChildWatchFunc)process_child_wait_func, 
		      (gpointer)self);
  } 
  here();
  return ret;
}


static void
process_child_wait_func(GPid pid, gint status, SandboxerProcess* self)
{
  SandboxerProcessClass   *klass;

  g_assert(self);
  klass= SANDBOXER_PROCESS_GET_CLASS(self);
  g_assert(klass);
  
  g_assert(self->pid == pid);
  

  self->status= RAN;
  self->exit_status = status;

  g_debug("%s (%d) exited w/ status %d", self->path, self->pid, WEXITSTATUS(status)); 

  g_signal_emit(self, /* instance */ 
		klass->signals[E_SIGNAL_TERMINATED], /* signal id */
		0, /* detail */
		status); 
}

gboolean
sandboxer_process_kill (SandboxerProcess *self, 
			const gchar *secret_cookie,
			const gint sig, 
			GError **error)
{
  g_assert(self);

  if (!g_str_equal(secret_cookie, self->secret_cookie)) {
    g_set_error(error,
		SANDBOXER_PROCESS_ERROR,
		SANDBOXER_PROCESS_ERROR_COOKIE,
		"incorrect cookie");
    return FALSE;
  }


  if (RUNNING != self->status) {
    g_set_error(error,
		SANDBOXER_PROCESS_ERROR,
		SANDBOXER_PROCESS_ERROR_KILL,
		"Cannot kill process that is not running");
    return FALSE;
  }
  push_cred(0, 0);
  if (kill(self->pid, sig)) {
    g_set_error(error,
		SANDBOXER_PROCESS_ERROR,
		SANDBOXER_PROCESS_ERROR_KILL,
		"kill: %s", g_strerror(errno));
    pop_cred();
    return FALSE;
  }
  pop_cred();
  return TRUE;
}


//FIXME: unit tests
void
sandboxer_process_list_env(SandboxerProcess *self, gchar ***envp)
{
  guint i;

  g_assert(self);
  g_assert(envp);
  
  *envp= (gchar**)(g_new0(gpointer, self->env->len + 1)); /* ugly cast due to ?bug? in glib */
  for (i= 0; i < self->env->len; i++) {
    (*envp)[i]= g_strdup((gchar*)(g_ptr_array_index(self->env, i)));
  }
}

gchar*
sandboxer_process_get_secret_cookie(SandboxerProcess *self)
{

  g_assert(self);
  g_assert(self->secret_cookie);
  return g_strdup(self->secret_cookie);
}

static gboolean 
instance_plugin_hook_caller(SandboxerBase *self,
			    SandboxerPluginDescriptor *desc,
			    const SandboxerPluginHookStage stage,
			    GError **error)
{
  g_assert(self);
  g_assert(desc);
  
  if (desc->process_hooks[stage]) {
    return (desc->process_hooks[stage])((SandboxerProcess*)self, error);
  }
  return TRUE;
}

static gboolean
klass_plugin_hook_caller(SandboxerBaseClass *klass,
			 SandboxerPluginDescriptor *desc,
			 GError **error)
{
  g_assert(klass);
  g_assert(desc);
  
  if (desc->process_class_init_hook) {
    return (desc->process_class_init_hook)((SandboxerProcessClass*)klass, error);
  }
  return TRUE;  
}
