/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#ifndef _PROCESS_H_
#define _PROCESS_H_

#include <config.h>
#include <daemon/base.h>
#include <daemon/sandbox.h>

#include <glib.h>
#include <glib-object.h>
#include <gmodule.h>


#define SANDBOXER_DBUS_PROCESS_PATH  "/com/intel/sandboxer/Process"
#define SANDBOXER_DBUS_PROCESS       "com.intel.sandboxer.Process"

#define SIGNAL_TERMINATED            "terminated"
#define SIGNAL_RUN                   "run"

#define SANDBOXER_TYPE_PROCESS            (sandboxer_process_get_type ())
#define SANDBOXER_PROCESS(object)         (G_TYPE_CHECK_INSTANCE_CAST ((object), SANDBOXER_TYPE_PROCESS, SandboxerProcess))
#define SANDBOXER_PROCESS_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), SANDBOXER_TYPE_PROCESS, SandboxerProcessClass))
#define SANDBOXER_IS_PROCESS(object)      (G_TYPE_CHECK_INSTANCE_TYPE ((object), SANDBOXER_TYPE_PROCESS))
#define SANDBOXER_IS_PROCESS_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), SANDBOXER_TYPE_PROCESS))
#define SANDBOXER_PROCESS_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), SANDBOXER_TYPE_PROCESS, SandboxerProcessClass))


#define SANDBOXER_PROCESS_ERROR sandboxer_process_error_quark ()


G_BEGIN_DECLS


typedef enum {
  /** signal emmitted when the process in a #SandboxerProcess is run */
  E_SIGNAL_RUN, 

  /** signal emmitted when the process in a #SandboxerProcess terminates */
  E_SIGNAL_TERMINATED, 
  
  /** signal count (used for iterating) */
  E_SIGNAL_COUNT,
} ProcessSignalNumber;


/**
 *  GError type for #SandboxerProcess
 */
typedef enum {
  /** the caller used the incorrect secret cookie */
  SANDBOXER_PROCESS_ERROR_COOKIE, 

  /** an error occurred in the setup of a process */
  SANDBOXER_PROCESS_ERROR_SETUP,  

  /** kill reported an error */
  SANDBOXER_PROCESS_ERROR_KILL,   

  /** Generic error */
  SANDBOXER_PROCESS_ERROR_FAILED  
} SandboxerProcessError;


typedef struct _SandboxerProcess SandboxerProcess;
typedef struct _SandboxerProcessClass SandboxerProcessClass;

/**
 * The public representation of a SandboxerProcess instance object
 */
struct _SandboxerProcess {
  /** the #SandboxerBase parent instance */
  SandboxerBase parent; 
 
  /* properties */
  gchar *path;                /* can be set at construction */
  SandboxerSandbox *sandbox;  /* can be set at construction */
  GPid pid;                   /* readable                   */

  /* emitted via signal */
  gint exit_status;    
  
  /* entirely private */
  gint status;     //NOTRUN, RUNNING, RAN
  uid_t uid;       //uid to run as 
  gid_t gid;       
  GPtrArray *env;
  gchar secret_cookie[37]; /* defined length in uuid_unparse(3) */
};

/**
 *  The class struct for #SandboxerProcess.  Derived from #SandboxerBase.
 */
struct _SandboxerProcessClass {
  SandboxerBaseClass parent; /** #SandboxerBaseClass parent that this class is derived from */

  guint signals[E_SIGNAL_COUNT]; /** the signals that this class emits */
};

GType sandboxer_process_get_type (void);

/**
 * create a new #SandboxerProcess object
 *
 * \param[in]  path            path to the executable for this #SandboxerProcess
 * \param[out] secret_cookie   return location for a string that will be used as access control for other methods.
 * \param[in]  sandbox         a pointer to a #SandboxerSandbox that this process belongs to, if any.
 *
 * \return the newly created #SandboxerProcess
 */
SandboxerProcess* 
sandboxer_process_new(const gchar *path, gchar **secret_cookie, SandboxerSandbox *sandbox);

/**
 * run the #SandboxerProcess.  If the process
 * could not be run, then error is set with a #SandboxerProcessError
 *
 * \param[in]  self             a #SandboxerProcess
 * \param[in]  secret_cookie    the secret cookie for this sandboxer process as given by sandboxer_process_new().
 * \param[in]  argv             the argument vector to use for this process
 * \param[out] error            return location for a GError or NULL
 *
 * \returns TRUE for success; FALSE for failure.  
 */
gboolean
sandboxer_process_run (SandboxerProcess *self,
		       const gchar *secret_cookie,
		       gchar **argv, 
		       GError **error);

/**
 * kill the #SandboxerProcess
 *
 * \param[in]  self            a #SandboxerProcess
 * \param[in]  secret_cookie   the secret cookie for this #SandboxerProcess.  If there is an error, error is set to a #SandboxerProcessError
 * \param[in]  sig             the signal number to send (as with kill(2))
 * \param[out] error           return location for a GError or NULL
 *
 * \returns  TRUE for success; FALSE for failure
 * \sa sandboxer_process_new()
 */
gboolean
sandboxer_process_kill (SandboxerProcess *self,
			const gchar *secret_cookie,
			const gint sig, 
			GError **error);

/**
 * Add an environment variable to be in this #SandboxerProcess environment.  Note that it does
 * not replace the variable if it already exists.
 *
 * \param[in] self   a #SandboxerProcess
 * \param[in] key    the environment variable name
 * \param[in] value  the environment variable value
 */
void 
sandboxer_process_add_to_env(SandboxerProcess *self, gchar *key, gchar *value);

/**
 * add a string of the form KEY=VALUE
 *
 * \param[in] self    a #SandboxerProcess
 * \param[in] string  string of the form KEY=VALUE
 *
 * \note it does not replace the variable if it already exists.
 */
void
sandboxer_process_add_to_env_pair(SandboxerProcess *self, gchar *string);

/**
 * gives back a gchar **env suitable for exec and g_spawn_async()
 * the memory returned should be freed by the user with g_freev()
 *
 * \param[in]   self    a #SandboxerProcess
 * \param[out] envp    a pointer to the env pointer to be set
 */
void
sandboxer_process_list_env(SandboxerProcess *self, gchar ***envp);

/**
 *  return the secret cookie for this particular #SandboxerProcess
 *
 *  \param[in]  self   a #SandboxerProcess
 *
 *  \returns  a string containing the secret cookie.  This string can be freed with g_free()
 */
gchar*
sandboxer_process_get_secret_cookie(SandboxerProcess *self);

/**
 * give back an error quark for sandboxer process
 */
GQuark
sandboxer_process_error_quark (void);


G_END_DECLS

#endif /* _PROCESS_H_ */
