/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <daemon/process_container.h>
#include <daemon/sandboxer-helper.h>

struct list_all_proc_context
{
  gchar **dbus_paths;
  gchar **secret_cookies;
  gchar **unique_names;
  guint len;
  gchar *unique_name;
};


static void
list_all_proc_for_each(SandboxerProcess *proc, struct list_all_proc_context *context);

static void     
process_run_signal_callback(gpointer instance, const gchar *path, SandboxerProcessContainer *self);

static void
process_terminated_signal_callback(gpointer instance, const gint status, SandboxerProcessContainer *self);




SandboxerProcessContainer*
sandboxer_process_container_new()
{
  SandboxerProcessContainer *self= g_new0(SandboxerProcessContainer, 1);

  self->pid_process= g_hash_table_new_full(g_int_hash, 
					   g_str_equal,
					   g_free,
					   g_object_unref);
  g_assert(self->pid_process);
  return self;
}

void
sandboxer_process_container_add_process(SandboxerProcessContainer *self,
					SandboxerProcess *process)
{
  g_assert(self); g_assert(process);
    
  self->processes= g_slist_prepend(self->processes, process);
  g_signal_connect(process, "run", G_CALLBACK(process_run_signal_callback), self);
  g_signal_connect(process, "terminated", G_CALLBACK(process_terminated_signal_callback), self);
}

void
sandboxer_process_container_list_all(SandboxerProcessContainer *self,
				     gchar *unique_name,
				     gchar ***paths,
				     gchar ***secret_cookies,
				     gchar ***unique_names)
{
  struct list_all_proc_context context;
  
  g_assert(unique_name);
  g_assert(self); g_assert(paths); g_assert(secret_cookies); g_assert(unique_names);
  
  context.dbus_paths=     g_new0(gchar*, g_slist_length(self->processes) + 1);  
  context.secret_cookies= g_new0(gchar*, g_slist_length(self->processes) + 1);  
  context.unique_names=   g_new0(gchar*, g_slist_length(self->processes) + 1);  
  context.len= 0;
  context.unique_name= unique_name;

  g_slist_foreach(self->processes, (GFunc)list_all_proc_for_each, &context);
  g_debug("conext.dbus_paths length: %u (%s)", g_strv_length(context.dbus_paths), __FUNCTION__);
  g_debug("conext.secret_cookies length: %u (%s)", g_strv_length(context.secret_cookies), __FUNCTION__);
  g_debug("conext.unique_names length: %u (%s)", g_strv_length(context.unique_names), __FUNCTION__);
 
  *paths= context.dbus_paths;
  *secret_cookies= context.secret_cookies;
  *unique_names= context.unique_names;  
}

void
sandboxer_process_container_find_process_by_pid(SandboxerProcessContainer *self,
						const gint pid,
						gchar **process,
						gchar **secret_cookie)
{
  SandboxerProcess *proc;
  const gchar* prop= "process-dbus_path";

  g_assert(process);
  g_assert(secret_cookie);
  
  proc= (SandboxerProcess*)g_hash_table_lookup(self->pid_process, &pid);
  if (proc) {
    g_object_get(proc, prop, process, NULL);
    *secret_cookie= sandboxer_process_get_secret_cookie(proc);
  } else {
    *process= NULL;
    *secret_cookie= NULL;
  }
}

static void
list_all_proc_for_each(SandboxerProcess *proc, struct list_all_proc_context *context)
{ 
  gchar *dbus_path;
  gchar *secret_cookie;

  g_assert(proc);
  g_assert(context);
  
  g_object_get(proc, "process-dbus_path", &dbus_path, NULL);
  g_assert(dbus_path);
  secret_cookie= sandboxer_process_get_secret_cookie(proc);
  g_assert(secret_cookie);

  context->dbus_paths[context->len]= dbus_path;
  context->secret_cookies[context->len]= secret_cookie;
  context->unique_names[context->len]= g_strdup(context->unique_name);
  context->len++;
}

static void     
process_run_signal_callback(gpointer instance, const gchar *path, SandboxerProcessContainer *self)
{
  gpointer *key_p;

  g_assert(instance); g_assert(path); g_assert(self);
  

  key_p= g_malloc(sizeof(gint)); g_assert(key_p);
  
  g_object_get(G_OBJECT(instance), "process-pid", key_p, NULL);
  g_hash_table_insert(self->pid_process, key_p, instance);
}

static void
process_terminated_signal_callback(gpointer instance, const gint status, SandboxerProcessContainer *self)
{
  gint pid= -1;

  g_assert(self);
  
  self->processes= g_slist_remove_all(self->processes, instance);
  g_object_get(G_OBJECT(instance), "process-pid", &pid, NULL);  
  g_assert(pid >= 1);
  if(!g_hash_table_remove (self->pid_process, &pid)) {
    g_error("couldn't remove from hashtable!");
  }
  if ((0 ==  g_hash_table_size(self->pid_process)) && 
      self->empty_func) {
    self->empty_func(self, self->empty_func_user_data);
  }
}

GSList *
sandboxer_process_container_give_process_list(SandboxerProcessContainer *self)
{
  g_assert(self);

  return self->processes;
}

void
sandboxer_process_container_set_empty_func(SandboxerProcessContainer *self, 
					   ProcessContainerEmptyFunc func, gpointer user_data)
{
  g_assert(self);

  self->empty_func= func;
  self->empty_func_user_data= user_data;
}
