/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#ifndef   __PROCESS_CONTAINER_H_
#define   __PROCESS_CONTAINER_H_

#include <config.h>
#include <daemon/process.h>
#include <glib.h>

G_BEGIN_DECLS

typedef struct _SandboxerProcessContainer SandboxerProcessContainer;

typedef void (*ProcessContainerEmptyFunc)(SandboxerProcessContainer *self, gpointer user_data);

/**
 * holds #SandboxerProcess's and allows them to be searched by PID
 */
struct _SandboxerProcessContainer {
  GHashTable *pid_process; /** a GHashTable where the keys are PID's
			    *  of the #SandboxerProcess the values are
			    *  pointers to the #SandboxerProcess held
			    */
  GSList *processes; /** a GSList containing pointers to #SandboxerProcess's */
  ProcessContainerEmptyFunc empty_func; /** a #ProcessContainerEmptyFunc function pointer that is
					 *  called when all the processeses have exited 
					 */
  gpointer empty_func_user_data; /** a gpointer that will be passed to empty_func when it is called */
};

/**
 * get a new #SandboxerProcessContainer
 *
 * \returns  a #SandboxerProcessContainer
 */
SandboxerProcessContainer*
sandboxer_process_container_new(void);

/**
 * add a #SandboxerProcess to a #SandboxerProcessContainer
 *
 * \param[in]  self     a #SandboxerProcessContainer
 * \param[in]  process  a #SandboxerProcess
 */
void
sandboxer_process_container_add_process(SandboxerProcessContainer *self, 
					SandboxerProcess *process);


/**
 * list all of the #SandboxerProcess's in this #SandboxerProcessContainer
 *
 * \param[in]  self         a #SandboxerProcessContainer 
 * 
 * \param[in]  unique_name  the unique DBUS connection name to replicate in
 * unique_names
 *
 * \param[out] paths         return location for a NULL terminated string array
 * containing the DBUS paths of the #SandboxerProcess.  The
 * array should be freed with g_strfreev().
 *
 * \param[out] secret_cookies  return location for a NULL terminated string
 * array containing the secret_cookies of the #SandboxerProcess.
 * The array should be freed with g_strfreev().
 *
 * \param[out] unique_names return location for a NULL terminated string array
 * containing replicated copies of the unique_name.  The array should
 * be free with g_strfreev().
 */
//FIXME: could paths -> process_dbus_paths or some such
void
sandboxer_process_container_list_all(SandboxerProcessContainer *self,
				     gchar *unique_name,
				     gchar ***paths,
				     gchar ***secret_cookies,
				     gchar ***unique_names);

/**
 * find a #SandboxerProcess in a #SandboxerProcessContainer by its
 * PID.
 *
 * \param[in] self  a #SandboxerProcessContainer
 * 
 * \param[in] pid a PID of a running #SandboxerProcess
 *
 * \param[out] process  return location for a string that contains the dbus path
 * to the #SandboxerProcess.  This string can be freed with g_free
 *
 * \param[out] secret_cookie return location for a string that contains the
 * secret cookie to be used with this #SandboxerProcess.  This string
 * can be freed with gfree.
 *
 */
//FIXME: process -> process_dbus_path
void
sandboxer_process_container_find_process_by_pid(SandboxerProcessContainer *self,
						const gint pid,
						gchar **process,
						gchar **secret_cookie);
/**
 * sandboxer_process_container_give_process_list: return a pointer to
 * the internal GSList that holds all #SandboxerProcess's in
 * the #SandboxerProcessContainer
 *
 * \param[in] self: a #SandboxerProcessContainer
 *
 * \returns a GSList.  The GSList cannot be freed and is internal to
 * the #SandboxerProcessContainer
 */
GSList *
sandboxer_process_container_give_process_list(SandboxerProcessContainer *self);

/**
 * set a function to be called when all processes in this container
 * exit.  Note that if the function is already set, then it is
 * replaced with the new value.
 *
 * \param[in]  self  a #SandboxerProcessContainer
 * \param[in]  func  a #ProcessContainerEmptyFunc to call
 * \param[in]  user_data pointer to be passed as user_data in calls to func
 */
void
sandboxer_process_container_set_empty_func(SandboxerProcessContainer *self,
					   ProcessContainerEmptyFunc func, gpointer user_data);

G_END_DECLS

#endif /* __PROCESS_CONTAINER_H_ */
