/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <daemon/sandbox.h>
#include <daemon/sandboxer-helper.h>
#include <daemon/process_container.h>
#include <daemon/plugin_impl.h>

#include <sys/socket.h>
#include <sys/un.h>
#include <sys/prctl.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <pwd.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <dbus/dbus-glib-bindings.h>

#ifndef UNIX_PATH_MAX
#define UNIX_PATH_MAX 108 //as per LSB
#endif

#define THRESHOLD 3

G_DEFINE_TYPE(SandboxerSandbox, sandboxer_sandbox, SANDBOXER_TYPE_BASE);

#include <daemon/sandboxer-sandbox-server-glue.h>
#include <daemon/sandbox-marshal.h>

#define SANDBOXER_SANDBOX_GET_PRIVATE(o)  \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), SANDBOXER_TYPE_SANDBOX, SandboxerSandboxPrivate))


typedef struct _SandboxerSandboxPrivate SandboxerSandboxPrivate;

struct _SandboxerSandboxPrivate {
  SandboxerProcessContainer *container;
  guint times;
  gboolean enable_timer;
  gchar *sandbox_name;
};

enum {
  PROP_0, 
  PROP_SANDBOX_NAME
};
  

static void
sandboxer_sandbox_init (SandboxerSandbox *self);

static void
sandboxer_sandbox_class_init (SandboxerSandboxClass *klass);

static void
sandboxer_sandbox_finalize(GObject *object);

static void
sandboxer_sandbox_set_property (GObject *object, guint prop_id,
				const GValue *value,
				GParamSpec *pspec);

static void 
sandboxer_sandbox_get_property (GObject *object, guint prop_id,
				GValue *value,
				GParamSpec *pspec);

static GObject *
sandboxer_sandbox_constructor(GType type,
			      guint n_construct_properties,
			      GObjectConstructParam *construct_properties);

static void
no_running_proc_hook(SandboxerProcessContainer *proc_cont, SandboxerSandbox *self);

static gboolean
timeout_source_func(SandboxerSandbox *self);


static void
stubd_send_connection_name(SandboxerSandboxClass *klass);

static gboolean 
instance_plugin_hook_caller(SandboxerBase *self,
			    SandboxerPluginDescriptor *desc,
			    const SandboxerPluginHookStage stage,
			    GError **error);
static gboolean
klass_plugin_hook_caller(SandboxerBaseClass *klass,
			 SandboxerPluginDescriptor *desc,
			 GError **error);

static void
add_to_dirs_list_if_unique(GPtrArray *dirs, const gchar *dir);





static gboolean
timeout_source_func(SandboxerSandbox *self)
{
  SandboxerSandboxPrivate *priv;

  g_assert(self);
  priv= SANDBOXER_SANDBOX_GET_PRIVATE(self); g_assert(priv);
  if (!priv->enable_timer) return FALSE;
  g_debug("priv->times= %u", priv->times);
  if (priv->times > THRESHOLD) {
    g_debug("timed out");
    exit(EXIT_SUCCESS);
  } else {
    priv->times++;
  }
  return TRUE;
}

static void
no_running_proc_hook(SandboxerProcessContainer *proc_cont, SandboxerSandbox *self)
{
  SandboxerSandboxPrivate *priv;
  
  g_assert(proc_cont); g_assert(self);
  priv= SANDBOXER_SANDBOX_GET_PRIVATE(self); g_assert(priv);
  g_debug("all procs are done");
  priv->times= 0;
  priv->enable_timer= TRUE;
}

GQuark
sandboxer_process_error_quark (void);




GQuark
sandboxer_sandbox_error_quark (void)
{
  return g_quark_from_static_string ("sandboxer-sandbox-error-quark");
}

gchar *
sandboxer_sandbox_generate_dbus_path(const GPid pid)
{ 
  return g_strdup_printf("%s/%d", SANDBOXER_DBUS_SANDBOX_PATH, 
			pid);
}


static void
sandboxer_sandbox_class_init(SandboxerSandboxClass *klass)
{
  GObjectClass *g_klass= G_OBJECT_CLASS (klass);
  GParamSpec *pspec;
  GError *error= NULL;

  g_type_class_add_private (klass, sizeof (SandboxerSandboxPrivate));
    
  g_klass->constructor = sandboxer_sandbox_constructor;
  g_klass->finalize= sandboxer_sandbox_finalize;
  g_klass->set_property= sandboxer_sandbox_set_property;
  g_klass->get_property= sandboxer_sandbox_get_property;

  pspec= g_param_spec_string("sandbox-sandbox_name",
			     "sandbox_name",
			     "name of this sandbox as in the .package file",
			     "",
			     G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
  g_object_class_install_property (g_klass, PROP_SANDBOX_NAME, pspec);  


  klass->connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
  g_assert(klass->connection);

  klass->parent.plugin_instance_hook_caller= instance_plugin_hook_caller;
  klass->parent.plugin_class_hook_caller= klass_plugin_hook_caller;
  here();
  if (!sandboxer_base_class_caller(SANDBOXER_BASE_CLASS(klass), &error)) {
    g_error(error->message);
  }

  dbus_g_object_type_install_info(SANDBOXER_TYPE_SANDBOX, 
				  &dbus_glib_sandboxer_sandbox_object_info);
}

static void
sandboxer_sandbox_init(SandboxerSandbox *self)
{
  SandboxerSandboxPrivate *priv;

  g_assert(self);
  priv= SANDBOXER_SANDBOX_GET_PRIVATE(self); g_assert(priv);
  priv->container= sandboxer_process_container_new();   g_assert(priv->container);
  sandboxer_process_container_set_empty_func(priv->container, 
					     (ProcessContainerEmptyFunc)no_running_proc_hook, 
					     self);
  g_timeout_add_seconds(10, 
			(GSourceFunc) timeout_source_func,
			(gpointer)self);
}

static GObject *
sandboxer_sandbox_constructor(GType type,
			      guint n_construct_properties,
			      GObjectConstructParam *construct_properties)
{
  GObject *obj;
  GObjectClass *parent_class;
  SandboxerSandboxClass *klass;
  SandboxerSandbox *self;
  gchar *dbus_path;
  
  /* Invoke parent constructor. */
  klass = SANDBOXER_SANDBOX_CLASS (g_type_class_peek (SANDBOXER_TYPE_SANDBOX));
  parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
  obj = parent_class->constructor (type,
				   n_construct_properties,
				   construct_properties);
  
  self= SANDBOXER_SANDBOX(obj);
  dbus_path= sandboxer_sandbox_generate_dbus_path(getpid());
		
  g_debug("registering dbus_path=\"%s\"", dbus_path);
  dbus_g_connection_register_g_object(klass->connection,
				      dbus_path,
				      G_OBJECT(self));
  stubd_send_connection_name(klass);
  g_free(dbus_path);
  return obj;
}

gboolean
sandboxer_sandbox_get_process(SandboxerSandbox *self, const gchar *path,
			      gchar **process, gchar **secret_cookie, gchar **unique_name,
			      GError **error)
{
  SandboxerSandboxPrivate *priv;
  SandboxerSandboxClass *klass;
  SandboxerProcess *proc;


  g_assert(self);
  priv= SANDBOXER_SANDBOX_GET_PRIVATE(self); g_assert(priv);
  klass= SANDBOXER_SANDBOX_GET_CLASS(self); g_assert(klass);
  proc= sandboxer_process_new(path, secret_cookie, self); 
  g_assert(process);
  sandboxer_process_container_add_process(priv->container, proc);
  g_object_get(G_OBJECT(proc), "process-dbus_path", process, NULL);
  *unique_name= g_strdup((gchar*)get_unique_connection_name(klass->connection));
  priv->enable_timer= 0;
  priv->times= 0;
  return TRUE;
}

static void
add_to_dirs_list_if_unique(GPtrArray *dirs, const gchar *dir)
{
  guint i;
  gboolean unique;

  g_assert(dirs); g_assert(dir);
  unique= TRUE;
  for (i= 0; i < dirs->len; i++) {
    if (g_str_equal(dir, (gchar*)g_ptr_array_index(dirs, i)))
      unique= FALSE;
  }
  if (unique) g_ptr_array_add(dirs, g_strdup(dir));
}

SandboxerSandbox*
sandboxer_sandbox_new(const gchar *sandbox_name)
{
  SandboxerSandbox *self;
  SandboxerSandboxClass *klass;
  GError *error= NULL;
  
  here();
  g_assert(sandbox_name); 
  g_debug("creating sandbox with name \"%s\"", sandbox_name);
  self= (SandboxerSandbox*)g_object_new(SANDBOXER_TYPE_SANDBOX, 
					"sandbox-sandbox_name", sandbox_name, 
					NULL); 
  g_assert(SANDBOXER_IS_SANDBOX(self));  
  klass= SANDBOXER_SANDBOX_GET_CLASS(self); g_assert(klass);

  {
    guint num_sandboxes;
    guint i;
    SandboxConfig **sandboxes;
    
    sandboxes= sandbox_config_store_lookup_sandbox(klass->parent.config_store,
						   sandbox_name,
						   &num_sandboxes);
    g_assert(sandboxes); g_assert(num_sandboxes);
    for (i= 0; i < num_sandboxes; i++) {
      struct passwd *ppasswd;

      ppasswd= getpwuid(sandboxes[i]->uids[0]);
      g_return_val_if_fail(ppasswd, NULL);
      add_to_dirs_list_if_unique(self->parent.dirs_in_jail, ppasswd->pw_dir);
    }
    g_ptr_array_add(self->parent.dirs_in_jail, NULL);
  }
  
  //NOTE:
  //  when you ensure the location of .package files must be present or plugins
  //  may not work right.
  if (!sandboxer_base_unshare_through_umount(SANDBOXER_BASE(self), 
					     (const gchar**)(self->parent.dirs_in_jail->pdata),
					     &error)) {
    g_error("could not set up jail correctly: %s", error->message);
  }
  return self;
}

gboolean 
sandboxer_sandbox_list_all_processes(SandboxerSandbox *self, 
				     gchar ***paths, gchar ***secret_cookies,
				     gchar ***unique_names, GError **error)
{
  SandboxerSandboxPrivate *priv;
  SandboxerSandboxClass   *klass;

  g_assert(self);
  priv= SANDBOXER_SANDBOX_GET_PRIVATE(self); g_assert(priv);
  klass= SANDBOXER_SANDBOX_GET_CLASS(self); g_assert(klass);

  sandboxer_process_container_list_all(priv->container,
				       klass->unique_name,
				       paths,
				       secret_cookies,
				       unique_names);

  return TRUE;
}

gboolean 
sandboxer_sandbox_find_process_by_pid(SandboxerSandbox *self, 
				      const gint pid, gboolean *found,
				      gchar **process, gchar **secret_cookie, gchar **unique_name,
				      GError **error)
{
  SandboxerSandboxPrivate *priv;
  SandboxerSandboxClass   *klass;

  g_assert(self); g_assert(process); g_assert(secret_cookie); g_assert(unique_name);
  priv= SANDBOXER_SANDBOX_GET_PRIVATE(self); g_assert(priv);
  klass= SANDBOXER_SANDBOX_GET_CLASS(self); g_assert(klass);

  *process= NULL;
  sandboxer_process_container_find_process_by_pid(priv->container,
						  pid,
						  process,
						  secret_cookie);
  if (*process) {
    *found= TRUE;
    *unique_name= g_strdup(klass->unique_name);
  } else {
    *found= FALSE; 
  }
  
  return TRUE;
}

static void
stubd_send_connection_name(SandboxerSandboxClass *klass)
{
  int sfd;
  struct sockaddr_un serv_addr;
  GIOChannel *io;
  gchar *msg;
  gchar *env_value;
  gsize bytes_written;
  GError *merror= NULL;
  
  env_value= (gchar*)g_getenv("STUBD_SOCKET_PATH");
  if (env_value) {
    serv_addr.sun_family= AF_UNIX;
    serv_addr.sun_path[UNIX_PATH_MAX -1]= 0;
    strncpy(serv_addr.sun_path, (char*)env_value, UNIX_PATH_MAX - 2); 
    
    sfd = socket(AF_UNIX, SOCK_STREAM, 0);
    if (-1 == sfd) {
      g_error("socket: %s", g_strerror(errno));
    }
    if (-1 == connect(sfd, (const struct sockaddr*)(&serv_addr),
		      sizeof(struct sockaddr_un))) {
      g_error("connect: %s", g_strerror(errno));
    }
    io= g_io_channel_unix_new(sfd); g_assert(io);
    msg= g_strdup_printf("%d=%s\n", getpid(), get_unique_connection_name(klass->connection));
    if (G_IO_STATUS_NORMAL != g_io_channel_write_chars(io,
						       msg,
						       -1,
						       &bytes_written,
						       &merror))
      {
	g_error("g_io_channel_write_chars: %s", merror ? merror->message : "(NULL)");
      }
    g_free(msg);
    g_io_channel_shutdown(io, TRUE, &merror);
    if (merror) g_error("merror: %s", merror->message);
    //shutdown socket (?)
  }
}

static gboolean 
instance_plugin_hook_caller(SandboxerBase *self,
			    SandboxerPluginDescriptor *desc,
			    const SandboxerPluginHookStage stage,
			    GError **error)
{
  g_assert(self);
  g_assert(desc);

  if (stage > PRE_UMOUNT) {
        g_set_error(error,
		    SANDBOXER_SANDBOX_ERROR,
		    SANDBOXER_SANDBOX_ERROR_FAILED,
		    "invalid hook stage: %d", stage);
	return FALSE;
  }
  if (desc->sandbox_hooks[stage]) {    
    return (desc->sandbox_hooks[stage])((SandboxerSandbox*)self, error);
  }
  return TRUE;
}

static gboolean
klass_plugin_hook_caller(SandboxerBaseClass *klass,
			 SandboxerPluginDescriptor *desc,
			 GError **error)
{
  g_assert(klass);
  g_assert(desc);
  
  if (desc->sandbox_class_init_hook) {
    return (desc->sandbox_class_init_hook)((SandboxerSandboxClass*)klass, error);
  }
  return TRUE;  
}


static void
sandboxer_sandbox_finalize(GObject *object)
{
  SandboxerSandbox *self= SANDBOXER_SANDBOX(object); 
  GError *error= NULL;

  SANDBOXER_BASE_CLASS(sandboxer_sandbox_parent_class)->clean_jail_dir(SANDBOXER_BASE(self), 
								       NULL);
  
  

  /* plugins */
  if (!sandboxer_base_instance_caller(SANDBOXER_BASE(self),
				      INSTANCE_FINALIZE,
				      &error)) {
    g_error(error->message);
    g_error_free(error);
  }
  
  g_free(self->sandbox_dir);

  if (G_OBJECT_CLASS (sandboxer_sandbox_parent_class)->finalize)
    G_OBJECT_CLASS (sandboxer_sandbox_parent_class)->finalize(object); 
}

static void
sandboxer_sandbox_set_property (GObject *object, guint prop_id,
				const GValue *value,
				GParamSpec *pspec)
{
  SandboxerSandbox *self;
  SandboxerSandboxPrivate *priv;

  g_assert(object); g_assert(value); g_assert(pspec);
  self= SANDBOXER_SANDBOX(object); g_assert(self);
  priv= SANDBOXER_SANDBOX_GET_PRIVATE(self); g_assert(self);
  
  switch(prop_id) 
    {
    case PROP_SANDBOX_NAME:
      priv->sandbox_name= g_value_dup_string(value);
      break;
      
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    }
}

static void 
sandboxer_sandbox_get_property (GObject *object, guint prop_id,
				GValue *value,
				GParamSpec *pspec)
{
  SandboxerSandbox *self;
  SandboxerSandboxPrivate *priv;

  g_assert(object); g_assert(value); g_assert(pspec);
  self= SANDBOXER_SANDBOX(object); g_assert(self);
  priv= SANDBOXER_SANDBOX_GET_PRIVATE(self); g_assert(self);
  switch(prop_id) 
    {
    case PROP_SANDBOX_NAME:
      g_value_set_string(value, priv->sandbox_name);
      break;
      
    default: 
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    }
}
