/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <daemon/config_store.h>
#include <daemon/sandboxer-service.h>
#include <daemon/sandboxer-helper.h>
#include <daemon/sandboxer-service-glue.h>
#include <daemon/sandboxer-sandbox-client-glue.h>
#include <daemon/process.h>
#include <daemon/process_container.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <errno.h> 
#include <glib.h>
#include <glib/gstdio.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-bindings.h>

#ifndef UNIX_PATH_MAX
#define UNIX_PATH_MAX 108 //as per LSB
#endif


G_DEFINE_TYPE(SandboxerSandboxService, sandboxer_sandbox_service, G_TYPE_OBJECT);

#define SANDBOXER_SERVICE_GET_PRIVATE(o)  \
   (G_TYPE_INSTANCE_GET_PRIVATE ((o), SANDBOXER_TYPE_SANDBOX_SERVICE, SandboxerServicePrivate))


/*  types   */
     
typedef struct _SandboxerServicePrivate SandboxerServicePrivate;
 
struct _SandboxerServicePrivate
{
  SandboxerProcessContainer *container;

  /* for stub'd processes */
  GHashTable *sandbox_proxy;
  GHashTable *pid_sandbox;
};

struct list_all_tuple {
  gchar **dbus_paths;
  gchar **secret_cookies;
  gchar **unique_names;
  guint len;
};

struct hash_for_each_context {
  GPtrArray *tuples;
  GError **error; 
};





/* prototypes */
static void
sandboxer_sandbox_service_finalize (GObject *object);

static void
sandboxer_sandbox_service_class_init (SandboxerSandboxServiceClass *klass);

static void
sandboxer_sandbox_service_init (SandboxerSandboxService *object);

static gboolean 
get_stubd_bus_name(SandboxerSandboxService *self,
		   const GPid pid,
		   gchar **stubd_unique_name,
		   GError **error);

static void
setup_stubd_socket(SandboxerSandboxServiceClass *klass);

static void
list_all_hash_foreach(gchar *sandbox_name, DBusGProxy *sandbox_proxy, 
		      struct hash_for_each_context *context);
static void
get_tot_len(struct list_all_tuple *tuple, guint *tot);

static void
copy_tuples_foreach(struct list_all_tuple *src_tuple, 
		    struct list_all_tuple *dest_tuple);

static void 
stubd_reap_func(GPid pid,
		gint status,
		SandboxerServicePrivate *priv);

static gboolean
get_process_stubd(SandboxerSandboxService *self, const gchar* path, 
		  gchar **process, gchar **secret_cookie, gchar **unique_name, 
		  SandboxConfig *conf,
		  GError **error);



/* implementations */
GQuark
sandboxer_sandbox_service_error_quark (void)
{
  return g_quark_from_static_string ("sandboxer-sandbox-service-error-quark");
}

//FIXME: this never really happens, but we need to kill stuff on shutdown
static void
sandboxer_sandbox_service_finalize (GObject *object)
{
  SandboxerServicePrivate *priv;
  
  g_assert(object);
  priv= SANDBOXER_SERVICE_GET_PRIVATE(SANDBOXER_SANDBOX_SERVICE(object));

  g_debug("sandbox_service_finalize");
  G_OBJECT_CLASS (sandboxer_sandbox_service_parent_class)->finalize (object);
}


static void
sandboxer_sandbox_service_class_init (SandboxerSandboxServiceClass *klass)
{
  guint result;
  GObjectClass *object_class;
  GError *error = NULL;

  g_assert(klass);

  g_type_class_add_private(klass, sizeof (SandboxerServicePrivate));

  klass->connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
  g_assert(klass->connection);
  klass->unique_name= (gchar*)get_unique_connection_name(klass->connection);
  g_debug("unique name= %s (%s)", klass->unique_name, __FUNCTION__);
  klass->proxy =  dbus_g_proxy_new_for_name (klass->connection,
					     DBUS_SERVICE_DBUS,
					     DBUS_PATH_DBUS,
					     DBUS_INTERFACE_DBUS);
  
  g_assert(klass->proxy);
  org_freedesktop_DBus_request_name (klass->proxy,
				     SANDBOXER_DBUS_SANDBOX_SERVICE, 
				     DBUS_NAME_FLAG_DO_NOT_QUEUE,
				     &result, &error);  
  if (error) {
    g_error(error->message);
  }
  if (DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER != result) {
    g_error("Could not become owner of %s.  Quitting", SANDBOXER_DBUS_SANDBOX_SERVICE);
  }
    

  dbus_g_object_type_install_info(SANDBOXER_TYPE_SANDBOX_SERVICE, 
				  &dbus_glib_sandboxer_service_object_info);

  klass->config_store= sandbox_config_store_new(NULL);   g_assert(klass->config_store);

  object_class = G_OBJECT_CLASS (klass);
  object_class->finalize = sandboxer_sandbox_service_finalize;

  setup_stubd_socket(klass);
}

static void
sandboxer_sandbox_service_init (SandboxerSandboxService *object)
{
  SandboxerServicePrivate      *priv;
  SandboxerSandboxServiceClass *klass;

  g_assert(object);
  klass= SANDBOXER_SANDBOX_SERVICE_GET_CLASS(object); g_assert(klass);
  priv=  SANDBOXER_SERVICE_GET_PRIVATE(object); g_assert(priv);

  priv->container= sandboxer_process_container_new();
  g_assert(priv->container);

  priv->sandbox_proxy= g_hash_table_new_full(g_str_hash,
					     g_direct_equal,
					     g_free, 
					     g_object_unref);
  priv->pid_sandbox= g_hash_table_new_full(g_int_hash,
					   g_str_equal,
					   g_free,
					   NULL);

  dbus_g_connection_register_g_object (klass->connection, 
				       SANDBOXER_DBUS_SANDBOX_SERVICE_PATH, 
				       G_OBJECT(object));
}


SandboxerSandboxService *
sandboxer_sandbox_service_new (void)
{
  return g_object_new (SANDBOXER_TYPE_SANDBOX_SERVICE, NULL);
}

//FIXME: clean out procs that are gotten but not ran every so often
gboolean 
sandboxer_service_get_process (SandboxerSandboxService *obj, const gchar* path, 
			       gchar **process, gchar **secret_cookie, gchar **unique_name, 
			       GError **error)
{
  SandboxerServicePrivate *priv;
  SandboxerSandboxServiceClass *klass;
  SandboxerProcess *proc;
  SandboxConfig *conf; 
  gint num_sandboxes;  //FIXME: name

 
  g_assert(obj);
  g_assert(path);
  g_assert(process); g_assert(secret_cookie); g_assert(unique_name);

  priv= SANDBOXER_SERVICE_GET_PRIVATE(obj); g_assert(priv);
  klass= SANDBOXER_SANDBOX_SERVICE_GET_CLASS(obj); g_assert(klass);
  
  conf= sandbox_config_store_lookup_path(klass->config_store, path);
  if (!conf) {
    *process= NULL;
    *secret_cookie= NULL;
    *unique_name= NULL;
    g_set_error(error,
		SANDBOXER_SANDBOX_SERVICE_ERROR,
		SANDBOXER_SANDBOX_SERVICE_ERROR_GET_PROCESS,
		"%s is not associated with a .package file", path);
    return FALSE;
  }
   
  num_sandboxes= sandbox_config_store_lookup_num_sandboxes(klass->config_store, 
							   conf->sandbox_name);
  g_debug("adding new process. path = \"%s\"", path);
  g_debug("%s has %d processes that can go in it", conf->sandbox_name, num_sandboxes);
  if (1 == num_sandboxes) {
    g_debug("klass->unique_name= %s (%s)", klass->unique_name, __FUNCTION__);
    *unique_name= g_strdup(klass->unique_name);

    proc= sandboxer_process_new(path, secret_cookie, NULL);
    g_assert(proc);
    
    sandboxer_process_container_add_process(priv->container, proc);
    g_object_get(G_OBJECT(proc), "process-dbus_path", process, NULL);
  } else {
    if (!get_process_stubd(obj, path, process, secret_cookie, unique_name,
			   conf, error))
      return FALSE;
  }
  return TRUE;  
}

static void
list_all_hash_foreach(gchar *sandbox_name, DBusGProxy *sandbox_proxy, 
		      struct hash_for_each_context *context)
{
  struct list_all_tuple *tuple;
  
  g_assert(sandbox_name); g_assert(sandbox_proxy); g_assert(context);
  
  //once we've had one error, we don't compound the problem
  if (*(context->error) == NULL) {
    
    tuple= g_new(struct list_all_tuple, 1); //should be slice?
    com_intel_sandboxer_Sandbox_list_all_processes (sandbox_proxy, 
						    &(tuple->dbus_paths),
						    &(tuple->secret_cookies),
						    &(tuple->unique_names),
						    context->error);
    tuple->len= g_strv_length(tuple->dbus_paths);
    g_ptr_array_add(context->tuples, (gpointer)tuple);
  }
}

static void
get_tot_len(struct list_all_tuple *tuple, guint *tot)
{
  g_assert(tuple); g_assert(tuple);

  *tot+= tuple->len;
}

static void
copy_tuples_foreach(struct list_all_tuple *src_tuple, 
		    struct list_all_tuple *dest_tuple)
{
  guint i;
    
  g_assert(src_tuple); g_assert(dest_tuple);
  for(i= 0; i < src_tuple->len; i++) {
    dest_tuple->dbus_paths[i+dest_tuple->len]= g_strdup(src_tuple->dbus_paths[i]);
    dest_tuple->secret_cookies[i+dest_tuple->len]= g_strdup(src_tuple->secret_cookies[i]); 
    dest_tuple->unique_names[i+dest_tuple->len]= g_strdup(src_tuple->unique_names[i]);
  }
  g_strfreev(src_tuple->dbus_paths);
  g_strfreev(src_tuple->secret_cookies);
  g_strfreev(src_tuple->unique_names);
  dest_tuple->len+= src_tuple->len;
}

gboolean 
sandboxer_service_list_all_processes(SandboxerSandboxService *obj, 
				     gchar ***processes, gchar ***secret_cookies,
				     gchar ***unique_names, GError **error)				     
{
  SandboxerServicePrivate        *priv;
  SandboxerSandboxServiceClass   *klass;
  struct hash_for_each_context context;
  struct list_all_tuple *pc_tuple;
  struct list_all_tuple final_tuple;
  guint tot_len= 0;
  GError *merror= NULL;

  g_assert(processes); g_assert(secret_cookies); g_assert(unique_names);
  g_assert(obj);
  priv= SANDBOXER_SERVICE_GET_PRIVATE(obj); g_assert(priv); 
  klass= SANDBOXER_SANDBOX_SERVICE_GET_CLASS(obj); g_assert(klass);
  context.tuples= g_ptr_array_sized_new(1 + g_hash_table_size(priv->sandbox_proxy));
  context.error= &merror;
  g_assert(context.tuples);
  
  pc_tuple= g_new0(struct list_all_tuple, 1);
  sandboxer_process_container_list_all(priv->container,
				       klass->unique_name,
				       &(pc_tuple->dbus_paths),
				       &(pc_tuple->secret_cookies),
				       &(pc_tuple->unique_names));
  pc_tuple->len= g_strv_length(pc_tuple->dbus_paths);
  
  g_ptr_array_add(context.tuples, (gpointer)pc_tuple);
  g_hash_table_foreach(priv->sandbox_proxy,
		       (GHFunc)list_all_hash_foreach,
		       &context);
  if (merror) {
    g_propagate_error(error, merror);
    return FALSE;
  }
  /* get all the tuples and weld them together into strv's */
  g_ptr_array_foreach(context.tuples, (GFunc)get_tot_len, (gpointer)(&tot_len));
  final_tuple.dbus_paths= g_new0(gchar*, tot_len + 1);
  final_tuple.secret_cookies= g_new0(gchar*, tot_len + 1);
  final_tuple.unique_names= g_new0(gchar*, tot_len + 1);
  final_tuple.len= 0;
  g_ptr_array_foreach(context.tuples, (GFunc)copy_tuples_foreach, (gpointer)(&final_tuple));
  *processes= final_tuple.dbus_paths;
  *secret_cookies= final_tuple.secret_cookies;
  *unique_names= final_tuple.unique_names;
  g_ptr_array_free(context.tuples, TRUE);
  
  return TRUE;
}

gboolean 
sandboxer_service_find_process_by_pid(SandboxerSandboxService *obj, 
				      const gint pid, 
				      gchar **process, gchar **secret_cookie, gchar **unique_name,
				      GError **error)
{
  SandboxerServicePrivate        *priv;
  SandboxerSandboxServiceClass   *klass;

  g_assert(obj); g_assert(process); g_assert(secret_cookie); g_assert(unique_name);
  priv= SANDBOXER_SERVICE_GET_PRIVATE(obj); g_assert(priv);
  klass= SANDBOXER_SANDBOX_SERVICE_GET_CLASS(obj); g_assert(klass);

  sandboxer_process_container_find_process_by_pid(priv->container,
						  pid,
						  process,
						  secret_cookie);
  if (*process) {
    *unique_name= g_strdup(klass->unique_name);
    return TRUE;
  } else {  //look in the stubds
    GHashTableIter iter;
    gpointer key, value;
    gboolean found;
    
    g_hash_table_iter_init (&iter, priv->sandbox_proxy);
    while (g_hash_table_iter_next (&iter, &key, &value)) {
      if (!com_intel_sandboxer_Sandbox_find_process_by_pid((DBusGProxy*)value,
							    pid, 
							    &found,
							    process,
							    secret_cookie,
							    unique_name,
							    error))
	return FALSE;

      if (found) 
	return TRUE;	
    }

  } 
  g_set_error(error,
	      SANDBOXER_SANDBOX_SERVICE_ERROR,
	      SANDBOXER_SANDBOX_SERVICE_ERROR_LIST_PROCESSES,
	      "pid is not a sandboxed process");
  return FALSE;    
}

#ifdef UNITTEST
/*
 *  This method is really just for unit tests
 *   it gives up a pointer to the private list
 *   of processes
 *
 */
GSList *
sandboxer_service_give_process_list(SandboxerSandboxService *obj)
{
  SandboxerServicePrivate *priv;
  g_assert(obj);
  priv= SANDBOXER_SERVICE_GET_PRIVATE(obj);
  g_assert(priv);

  return sandboxer_process_container_give_process_list(priv->container);
}
#endif

gboolean 
sandboxer_service_add_package_file(SandboxerSandboxService *self, 
				   const gchar *path, GError **error)
{
  SandboxConfigStore *config_store;
  GError *merror= NULL;

  g_assert(self);
  
  if (NULL == path) {
    g_set_error(error,
		SANDBOXER_SANDBOX_SERVICE_ERROR,
		SANDBOXER_SANDBOX_SERVICE_ERROR_ADD_PACKAGE_FILE,
		"path is empty");
    return FALSE;
  }
  config_store= sandbox_config_store_new(NULL); g_assert(config_store);
  if (!sandbox_config_store_add_file(config_store, path, &merror)) {
    g_set_error(error,
		SANDBOXER_SANDBOX_SERVICE_ERROR,
		SANDBOXER_SANDBOX_SERVICE_ERROR_ADD_PACKAGE_FILE,
		merror->message);
    g_error_free(merror);
    return FALSE;
  }
  return TRUE;
}

gboolean
sandboxer_service_remove_package(SandboxerSandboxService *self,
				 const gchar *package_name, GError **error)
{
  SandboxConfigStore *config_store;
  GError *merror= NULL;

  g_assert(self);
  if (NULL == package_name) {
    g_set_error(error,
		SANDBOXER_SANDBOX_SERVICE_ERROR,
		SANDBOXER_SANDBOX_SERVICE_ERROR_REMOVE_PACKAGE,
		"package_name is empty");
    return FALSE;
  }
  config_store= sandbox_config_store_new(NULL); g_assert(config_store);
  if (!sandbox_config_store_remove_package(config_store, package_name, &merror)) {
    g_set_error(error,
		SANDBOXER_SANDBOX_SERVICE_ERROR,
		SANDBOXER_SANDBOX_SERVICE_ERROR_REMOVE_PACKAGE,
		merror->message);
    g_error_free(merror);
    return FALSE;
  }
  return TRUE;
}

static gboolean 
get_stubd_bus_name(SandboxerSandboxService *self,
		   const GPid pid,
		   gchar **stubd_unique_name,
		   GError **error)
{ 
  SandboxerSandboxServiceClass *klass;
  int client_sfd;
  GIOChannel *io;
  gchar *msg;
  gchar **msg_parts;
  struct sockaddr_un client_addr;
  socklen_t client_addr_len= sizeof(struct sockaddr_un);
  GError *merror= NULL;
  gboolean ret= FALSE;

  g_assert(self); g_assert(stubd_unique_name);
  klass= SANDBOXER_SANDBOX_SERVICE_GET_CLASS(self); g_assert(klass);

  client_sfd= accept(klass->sfd, (struct sockaddr*)(&client_addr), &client_addr_len);
  do {
    if (-1 == client_sfd) break;
    g_debug("accepted");
    io= g_io_channel_unix_new(client_sfd); g_assert(io);
    if (G_IO_STATUS_NORMAL != g_io_channel_read_line(io,
						   &msg,
						   NULL,
						   NULL,
						     &merror)) break;
    msg= g_strchomp(msg);
    msg_parts= g_strsplit(msg, "=", 0); g_assert(msg_parts);
    if (pid != g_ascii_strtoll((const gchar*)msg_parts[0],
			       NULL,
			       10)) { g_debug("pid mismatch"); break; }
    if (':' != msg_parts[1][0]) break;
    *stubd_unique_name= g_strdup(msg_parts[1]);
    g_io_channel_shutdown(io, TRUE, &merror);
    if (merror) break;
    g_io_channel_unref(io);

    g_debug("stubd_unique_name: \"%s\"", *stubd_unique_name);
    g_free(msg);
    g_strfreev(msg_parts);
    ret= TRUE;
  } while(0);
  if (!ret) {
    g_set_error(error,
		SANDBOXER_SANDBOX_SERVICE_ERROR,
		SANDBOXER_SANDBOX_SERVICE_ERROR_FAIL,
		merror ? merror->message : g_strerror(errno));
  }
	    
  return ret;
}

static void
setup_stubd_socket(SandboxerSandboxServiceClass *klass)
{
  struct sockaddr_un my_addr;
  SandboxGlobalConfig *global_config;

  g_assert(klass);

  global_config= sandbox_global_config_new(NULL); g_assert(global_config);
  klass->stubd_socket_path= g_strdup_printf("%s/stubd.sock", 
					    global_config->jail_dir_path);
  
  my_addr.sun_family= AF_UNIX;
  my_addr.sun_path[UNIX_PATH_MAX - 1]= 0;
  strncpy(my_addr.sun_path, klass->stubd_socket_path, UNIX_PATH_MAX - 2);
     
  klass->sfd = socket(AF_UNIX, SOCK_STREAM, 0);
  if (-1 == klass->sfd) {
    g_error("socket: %s", g_strerror(errno));
  }
  
  if (g_file_test(klass->stubd_socket_path, G_FILE_TEST_EXISTS)) {
    if (g_unlink(klass->stubd_socket_path)) {
      g_error("unlink: %s", g_strerror(errno));
    }
  }
  if (-1 == bind(klass->sfd, (struct sockaddr *) &my_addr,
		 sizeof(struct sockaddr_un))) {
    g_error("bind: %s", g_strerror(errno));
  }
     
  if (-1 == listen(klass->sfd, LISTEN_BACKLOG)) {
    g_error("listen: %s", g_strerror(errno));
  }
}

static void 
stubd_reap_func(GPid pid,
		gint status,
		SandboxerServicePrivate *priv)
{
  g_assert(priv);
  gchar *sandbox_name;
  DBusGProxy *sandbox_proxy;
  
  if (0 != WEXITSTATUS(status)) {
    g_message("stubd (%d) exited with non-zero status %d!", pid, WEXITSTATUS(status));
  }
  sandbox_name= g_hash_table_lookup(priv->pid_sandbox, &pid);
  g_assert(sandbox_name);
  sandbox_proxy= g_hash_table_lookup(priv->sandbox_proxy, sandbox_name);
  g_assert(sandbox_proxy);
  g_hash_table_remove(priv->pid_sandbox, &pid);
  g_hash_table_remove(priv->sandbox_proxy, sandbox_name);  
}

static gboolean
get_process_stubd(SandboxerSandboxService *self, const gchar* path, 
		  gchar **process, gchar **secret_cookie, gchar **unique_name, 
		  SandboxConfig *conf,
		  GError **error)
{
  SandboxerServicePrivate *priv;
  SandboxerSandboxServiceClass *klass;
  DBusGProxy *sandbox_proxy;
  GError *merror= NULL;

  priv= SANDBOXER_SERVICE_GET_PRIVATE(self); g_assert(priv);
  klass= SANDBOXER_SANDBOX_SERVICE_GET_CLASS(self); g_assert(klass);
  
  sandbox_proxy= g_hash_table_lookup(priv->sandbox_proxy, conf->sandbox_name);
  if (sandbox_proxy) {
    g_debug("reusing sandbox: %s", conf->sandbox_name);
  } else {
    gchar *stubd_argv[2];
    GPtrArray *stubd_env;
    gchar *global_config_file_path;
    SandboxGlobalConfig *global_config;
    GPid child_pid;
    gint *pid_p;
    gchar *stubd_unique_name;
    
    global_config= sandbox_global_config_new(NULL); g_assert(global_config);
    g_debug("creating new sandbox: %s", conf->sandbox_name);
    stubd_argv[0]= g_strdup_printf("%s/stubd", global_config->lib_exec_path);
    stubd_argv[1]= NULL;
    global_config_file_path= global_config->global_config_file_path;
    g_assert(global_config_file_path);
    here();
    stubd_env= g_ptr_array_sized_new(4); g_assert(stubd_env);
    g_ptr_array_add(stubd_env, g_strdup_printf("STUBD_CONFIG=%s", global_config_file_path));
    g_ptr_array_add(stubd_env, g_strdup_printf("STUBD_SOCKET_PATH=%s", klass->stubd_socket_path));
    g_ptr_array_add(stubd_env, g_strdup_printf("STUBD_SANDBOX_NAME=%s", conf->sandbox_name));
    g_ptr_array_add(stubd_env, NULL);
    i_am_here();
    push_cred(0, 0);
    
    do {
      if(!g_spawn_async(NULL,
			stubd_argv,
			(gchar**)(stubd_env->pdata),
			G_SPAWN_DO_NOT_REAP_CHILD, 
			NULL,
			NULL,
			&child_pid,
			&merror)) break;
      pop_cred();
      g_ptr_array_free(stubd_env, TRUE);
      g_free(stubd_argv[0]);
      if (!get_stubd_bus_name(self, child_pid, &stubd_unique_name,
			      &merror)) break;
      sandbox_proxy= dbus_g_proxy_new_for_name(klass->connection,
					       stubd_unique_name,
					       sandboxer_sandbox_generate_dbus_path(child_pid),
					       SANDBOXER_DBUS_SANDBOX);
      if (!sandbox_proxy) break;
    } while(0);
    if (merror) {
      g_propagate_prefixed_error(error, merror, "stubd setup:");
      return FALSE;
    }
    
    pid_p= g_new(gint, 1);
    *pid_p= child_pid;
    g_hash_table_insert(priv->sandbox_proxy, conf->sandbox_name, sandbox_proxy);
    g_hash_table_insert(priv->pid_sandbox, pid_p, conf->sandbox_name);
    g_child_watch_add(child_pid, (GChildWatchFunc)stubd_reap_func, (gpointer)priv);
  }
  
  //common to both cases
  if (!com_intel_sandboxer_Sandbox_get_process (sandbox_proxy, 
						path, 
						process,
						secret_cookie, 
						unique_name,
						&merror)) {
    g_propagate_prefixed_error(error, merror, "get_process: ");
    return FALSE;
  }
  return TRUE;
}
