/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#ifndef _SANDBOXER_SANDBOX_SERVICE_H_
#define _SANDBOXER_SANDBOX_SERVICE_H_

#include <config.h>
#include <daemon/config_store.h>

#include <glib.h>
#include <glib-object.h>
#include <dbus/dbus-glib.h>

#define SANDBOXER_DBUS_SANDBOX_SERVICE_PATH  "/com/intel/sandboxer/SandboxService"
#define SANDBOXER_DBUS_SANDBOX_SERVICE       "com.intel.sandboxer.SandboxService"
        
#define SANDBOXER_TYPE_SANDBOX_SERVICE            (sandboxer_sandbox_service_get_type ())
#define SANDBOXER_SANDBOX_SERVICE(object)         (G_TYPE_CHECK_INSTANCE_CAST ((object), SANDBOXER_TYPE_SANDBOX_SERVICE, SandboxerSandboxService))
#define SANDBOXER_SANDBOX_SERVICE_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), SANDBOXER_TYPE_SANDBOX_SERVICE, SandboxerSandboxServiceClass))
#define SANDBOXER_IS_SANDBOX_SERVICE(object)      (G_TYPE_CHECK_INSTANCE_TYPE ((object), SANDBOXER_TYPE_SANDBOX_SERVICE))
#define SANDBOXER_IS_SANDBOX_SERVICE_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), SANDBOXER_TYPE_SANDBOX_SERVICE))
#define SANDBOXER_SANDBOX_SERVICE_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), SANDBOXER_TYPE_SANDBOX_SERVICE, SandboxerSandboxServiceClass))

#define SANDBOXER_SANDBOX_SERVICE_ERROR sandboxer_sandbox_service_error_quark ()

#define LISTEN_BACKLOG 10

G_BEGIN_DECLS

typedef struct _SandboxerSandboxService SandboxerSandboxService;
typedef struct _SandboxerSandboxServiceClass SandboxerSandboxServiceClass;

/** 
 * object representation of the SandboxerService (com.intel.sandboxer.SandboxService)
 */
struct _SandboxerSandboxService {
  GObject parent; /** GObject Parent instance */
};

/**
 * class representation of the SaneboxerService (com.intel.sandboxer.SandboxService)
 * com.intel.sandboxer.SandboxService exports two interfaces: 
 *   - com.intel.sandboxer.SandboxService
 *     -# GetProcess 
 *   - com.intel.sandboxer.Query
 *     -# ListAllProcesses
 *     -# FindProcessbyPid
 *     -# AddPackageFile
 *     -# RemovePackage
 * 
 */
struct _SandboxerSandboxServiceClass {
  GObjectClass parent; /** GObject parent class */

  DBusGConnection *connection;
  gchar *unique_name; /** The unique name of the dbus connection */
  gchar *stubd_socket_path; /** path of the socket to use for communicating connection info to stubd */
  DBusGProxy *proxy; 
  SandboxConfigStore *config_store; 
  gint sfd; /** socket file descriptor for stubd_socket_path */
};

/**
 * used as the code field of a GError
 */
typedef enum {
  SANDBOXER_SANDBOX_SERVICE_ERROR_LIST_PROCESSES,
  SANDBOXER_SANDBOX_SERVICE_ERROR_ADD_PACKAGE_FILE,
  SANDBOXER_SANDBOX_SERVICE_ERROR_REMOVE_PACKAGE,
  SANDBOXER_SANDBOX_SERVICE_ERROR_GET_PROCESS,
  SANDBOXER_SANDBOX_SERVICE_ERROR_FAIL
} SandboxerSandboxServiceError;

/**
 * get a #SandboxerProcess object reference.  The reference to the
 * #SandboxerProcess is by its dbus path.  Implements GetProcess of
 * the DBUS interface com.intel.sandboxer.SandboxService
 *
 * In the DBUS view of the flow of things, getProcess creates a
 * /com/intel/sandboxer/Process object on the bus.  This object can
 * have methods called on it using the return values from getProcess.
 * The example below shows this.
 * 
 * \msc
 * "DBUS Caller", "SandboxerService", "SandboxerProcess";
 * "DBUS Caller" => "SandboxerService" [label="getProcess(path)"];
 * "SandboxerService" => "SandboxerProcess" [label="<<create>>"];
 * "SandboxerService" >> "DBUS Caller" [label="(secret_cookie, process, unique_name)"];
 * "DBUS Caller" => "SandboxerProcess" [label="run(secret_cookie, argv)"];
 * ... ;
 * "SandboxerProcess" ->* [label="terminated(status)"];
 * \endmsc
 *
 *
 * \param[in]  obj             the #SandboxerSandboxService object pointer
 * \param[in]  path            the path to the executable you are getting a sandbox for
 * \param[out] process         the dbus object path the resulting #SandboxerProcess can be found at
 * \param[out] secret_cookie   the secret cookie to use when calling methonds on the #SandboxerProcess
 * \param[out] unique_name     the unique DBUS connection name to connect to for the #SandboxerProcess
 * \param[out] error           location for a GError or NULL
 *
 * \returns TRUE on success; FALSE on failure
 */
gboolean 
sandboxer_service_get_process(SandboxerSandboxService *obj, const gchar* path, 
			      gchar **process, gchar **secret_cookie, gchar **unique_name, 
			      GError **error);

/**
 * list all the #SandboxerProcess's that are active
 *
 * \param[in] obj the #SandboxerSandboxService object pointer
 *
 * \param[out] paths a location for a NULL terminated array of strings
 *                   containing the DBUS object paths of the
 *                   #SandboxerProcess's.  The value returned at that
 *                   location may be freed with g_strfreev
 *
 * \param[out] secret_cookies a location for a NULL terminated array
 *                            of strings containing the secret_cookies
 *                            for the #SandboxerProcess's.  The value
 *                            returned at that location may be freed
 *                            with g_strfreev
 * 
 * \param[out] unique_names a location for a NULL terminated array of
 *                          strings containing the DBUS unique
 *                          connection names for the
 *                          #SandboxerProcess's.  The value returned
 *                          at the location may be freed with
 *                          g_strfreev
 * \param[out] error a location for a GError or NULL
 *
 * \returns TRUE for success; FALSE for failure
 */
gboolean 
sandboxer_service_list_all_processes(SandboxerSandboxService *obj, 
				     gchar ***paths, gchar ***secret_cookies,
				     gchar ***unique_names, GError **error);
/**
 * find the DBUS object path, secret, cookie, and unique connection name for a given PID
 *
 * \param[in] obj a #SandboxerSandboxService
 * \param[in] pid the pid to look up
 * \param[out] process the DBUS object path of the #SandboxerProcess
 * \param[out] secret_cookie the secret cookie givein by getProcess() when the #SandboxerProcess was created
 * \param[out] unique_name the unique DBUS connection name that this #SandboxerProcess can be found at
 * \param[out] error a location for a GError or NULL
 *
 * \returns TRUE on success; FALSE on failure
 */
gboolean 
sandboxer_service_find_process_by_pid(SandboxerSandboxService *obj, 
				      const gint pid, 
				      gchar **process, gchar **secret_cookie, gchar **unique_name,
				      GError **error);

//FIXME: add kill everybody method (for sandboxerd shutdown)

/**
 * add a .package file to the #SandboxConfigStore
 *
 * \param[in] self  a #SandboxerSandboxService
 * \param[in] path  the path to a .package file on disk
 * \param[out] error  location for a GError or NULL
 *
 * \returns TRUE on success; FALSE on failure
 */
gboolean 
sandboxer_service_add_package_file(SandboxerSandboxService *self, 
				   const gchar *path, GError **error);

/**
 * remove a .package file from the #SandboxConfigStore
 *
 * \param[in]  self  a #SandboxerSandboxService
 * \param[in]  package_name the PackageName value supplied in the .package file
 * \param[out] error location for a GError or NULL
 *
 * \returns TRUE on success; FALSE on failure
 */
gboolean
sandboxer_service_remove_package(SandboxerSandboxService *self,
				 const gchar *package_name, GError **error);

SandboxerSandboxService *sandboxer_sandbox_service_new (void);

GType sandboxer_sandbox_service_get_type (void);

#ifdef UNITTEST
GSList *
sandboxer_service_give_process_list(SandboxerSandboxService *obj);
#endif

G_END_DECLS

#endif
