/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <daemon/sandboxer-helper.h>
#include <daemon/sandboxer-service.h>
#include <daemon/config_store.h>
#include <daemon/plugin.h>
#include <daemon/plugin_impl.h>
#include <daemon/prctl_fixup.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/prctl.h>
#include <syslog.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <glib.h>
#include <glib/gthread.h>
#include <gmodule.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-bindings.h>

#define DEFAULT_CONFIG_FILE "/etc/sandboxer/sandboxerd.conf"

/*    globals      */

static SandboxGlobalConfig *global_config;
static gboolean daemonize= FALSE;
static gboolean log_to_syslog= FALSE;
gchar *config_file; /* location of config file to use */

static GOptionEntry entries[] =
  {
    {"daemonize", 'd', 0, G_OPTION_ARG_NONE,     &daemonize,     "Detatch from console and demonize", NULL},
    {"syslog",    's', 0, G_OPTION_ARG_NONE,     &log_to_syslog, "Send logging output to syslog", NULL},
    {"config",    'c', 0, G_OPTION_ARG_FILENAME, &config_file,   "Location of config file to use", NULL},
    {NULL}
  };

static gboolean
ensure_pid_file_dir(void);


//sends logs sent to it to the bit-bucket
static void
eat_log(const gchar *log_domain,
	GLogLevelFlags log_level,
	const gchar *message,
	gpointer unused)
{
  ;
}

//FIXME: check configuration file to figure out logging
static void
glog_to_syslog(const gchar *log_domain,
	       GLogLevelFlags log_level,
	       const gchar *message,
	       gpointer unused)
{
  int syslog_priority;
  char *level;

  switch (G_LOG_LEVEL_MASK & log_level) 
    {
    case G_LOG_LEVEL_ERROR:
      syslog_priority= LOG_ERR;
      level= "ERROR";
      break;
      
    case G_LOG_LEVEL_CRITICAL:
      syslog_priority= LOG_CRIT;
      level = "CRITICAL";
      break;

    case G_LOG_LEVEL_WARNING:
      syslog_priority= LOG_WARNING;
      level = "WARNING";
      break;

    case G_LOG_LEVEL_MESSAGE:
      syslog_priority= LOG_NOTICE;
      level = "NOTICE";
      break;
      
    case G_LOG_LEVEL_INFO:
      syslog_priority= LOG_INFO;
      level = "INFO";

      break;

    case G_LOG_LEVEL_DEBUG:
      syslog_priority= LOG_DEBUG;
      level = "DEBUG";
      break;
      
    default:
      /* this should not happen */
      syslog_priority= LOG_DEBUG;
      level = "DEBUG";
      syslog(LOG_MAKEPRI(LOG_DAEMON, syslog_priority), " %s - Got bizzare message level: 0x%X", level, log_level);
      break;      
    }

  if (log_domain) {
    syslog(LOG_MAKEPRI(LOG_DAEMON, syslog_priority), " %s - %s: %s", level, log_domain, message);
  } else {
    syslog(LOG_MAKEPRI(LOG_DAEMON, syslog_priority), " %s - %s", level, message);
  }
}

static void 
cleanup(int unused)
{
  g_message("exiting");
  if (global_config)
    unlink(global_config->pid_file_path);
  exit(EXIT_SUCCESS);
}

/*
 * ensure_pid_file_dir: make sure the directory that is the part of
 * pid_file_path, exists.  If not, create it.
 *
 * Returns: %TRUE for success; %FALSE for failure
 */
static gboolean
ensure_pid_file_dir(void)
{
  SandboxGlobalConfig *global_config;
  gchar *dupd;
  guint i;
  gboolean ret;
  
  global_config= sandbox_global_config_new(NULL); g_assert(global_config);
  g_assert(global_config->pid_file_path);
  ret= FALSE;
  dupd= g_strdup(global_config->pid_file_path); g_assert(dupd);
  for (i= strlen(dupd) - 1;
       dupd[i] != '/' && i > 0;
       i--);
  do {
    if (0 == i) break;
    dupd[i]= 0;
  
    if (-1 == g_mkdir_with_parents(dupd, S_IRWXU | S_IRGRP)) break;
    if (chown(dupd, global_config->sandbox_user_uid, global_config->sandbox_user_gid)) break;
    ret= TRUE;
  } while(0);
  g_free(dupd);

  return ret;
}


static void
write_pidfile(void)
{
  GIOChannel *pidfile;
  GIOStatus status;
  gchar *pid_string;
  GError *error= NULL;
  sigset_t mask;

  pid_string= g_strdup_printf("%d\n", getpid());
  sigemptyset(&mask);
  sigaddset(&mask, SIGTERM);
  sigaddset(&mask, SIGINT);
  pthread_sigmask(SIG_BLOCK, &mask, NULL);
  
  do {
    if (!ensure_pid_file_dir()) break;
    pidfile= g_io_channel_new_file(global_config->pid_file_path, "w", &error);
    if (NULL == pidfile)  break;
    status= g_io_channel_write_chars(pidfile, pid_string, strlen(pid_string), NULL, &error);
    if (G_IO_STATUS_NORMAL != status)  break;      
    status= g_io_channel_shutdown(pidfile, TRUE, &error);
    if (G_IO_STATUS_NORMAL != status)  break;      
  } while (0);
  if (error)
    g_error("%s: %s", global_config->pid_file_path, error ? error->message : "unknown");
   
  signal(SIGTERM, cleanup);
  signal(SIGINT, cleanup);

  pthread_sigmask(SIG_UNBLOCK, &mask, NULL);
}

int main (int argc, char **argv)
{
  GMainLoop *loop;
  SandboxerSandboxService* sandserv;  
  GOptionContext *context;
  SandboxConfigStore *config_store;
  GError *error = NULL;
 
  /* set slack to the maximum */
  prctl(PR_SET_TIMERSLACK, G_MAXUINT32, 0, 0, 0);

  g_type_init ();        
  if (!g_thread_supported ())
    g_thread_init (NULL);
  dbus_g_thread_init ();

  g_log_set_always_fatal (G_LOG_LEVEL_ERROR    |
			  G_LOG_LEVEL_CRITICAL |
			  G_LOG_LEVEL_WARNING);

  context= g_option_context_new(" - server for sandboxing applications");
  g_assert(context);
  g_option_context_add_main_entries (context, entries, NULL);
  if (!g_option_context_parse (context, &argc, &argv, &error)) {
    g_error("couldn't parse: %s", error->message);
  }

  if (log_to_syslog) {
    openlog (argv[0], LOG_PID, LOG_DAEMON);
    g_log_set_handler(NULL, G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION,
		      glog_to_syslog, NULL);  
  }

  if (!config_file) {
    config_file= DEFAULT_CONFIG_FILE;
  }
  g_debug("config_file=%s", config_file);
  global_config= sandbox_global_config_new(config_file);
  if (!global_config) {
    g_error("error reading config file");
  }

  g_debug("global_config->log_level= 0x%X", global_config->log_level);
  if (global_config->log_level)
    g_log_set_handler(NULL, global_config->log_level, eat_log, NULL);
  
  config_store= sandbox_config_store_new(global_config->package_file_directory);
  g_assert(config_store);

  if (!is_root()) {
    g_error("Refusing to run as %d.  Please run as root", geteuid());
  }

  if (daemonize) {
    if (g_file_test(global_config->pid_file_path, G_FILE_TEST_EXISTS)) {
      g_error("refusing to run because %s exists", global_config->pid_file_path);
    }
    if (-1 == daemon(0, 0)) {
      g_error("daemonize returned -1");
    }
  } else {
    g_message("Not daemonizing");
  }
  write_pidfile();
  g_message("%s started", g_get_prgname()); 
  i_am_here();
  push_cred(global_config->sandbox_user_uid, global_config->sandbox_user_gid);

  plugin_system_init();
  load_plugins();

  sandserv= sandboxer_sandbox_service_new();
  g_assert(sandserv);

  loop = g_main_loop_new (NULL, FALSE); g_assert(loop);  
  g_main_loop_run (loop);
    
  return 0; /* never happens */
}

