/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <daemon/sandboxer-helper.h>
#include <daemon/plugin.h>
#include <daemon/process.h>

static GPtrArray *dbus_keys= NULL;
  
/*
 * NOTE:
 *  this plugin makes the assumption that you have 
 *  <allow user="*"/> in either your dbus session.conf
 *  or session-local.conf
 *
 *  Additionally, this assumes that dbus has been
 *  autolaunched and the credentials are in 
 *  some file as a series of environment variables
 *
 */

//TODO: could be more memory efficient
static gboolean 
generic_class_init(SandboxerProcessClass *klass, GError **error)
{
  const gchar *dbus_group= "dbus";
  gchar *session_dir_path;
  gboolean ret;
  GDir *session_dir;
  gchar *dent;
  gchar *full_file_path;
  gchar *file_contents;
  SandboxGlobalConfig *global_config;
  GKeyFile *keyfile;

  if (dbus_keys) return TRUE; 

  global_config= sandbox_global_config_new(NULL);
  g_assert(global_config);
  
  keyfile= g_key_file_new(); g_assert(keyfile);
  
  g_assert(klass);
  do {
    ret= FALSE;
    
    if (!g_key_file_load_from_file(keyfile, global_config->global_config_file_path,
				   G_KEY_FILE_NONE, error)) break;
    session_dir_path= g_key_file_get_value(keyfile, dbus_group, "session_path", error);
    g_key_file_free(keyfile);
    if (!session_dir_path) break;
    g_debug("session_dir_path=%s", session_dir_path);

    push_cred(0,0);
    session_dir= g_dir_open(session_dir_path, 0, error);
    if (!session_dir) { pop_cred(); break; }
    
    /* we actually only take the first entry in the session bus dir
       if you have more than one session bus, something is wrong */
    dent= (gchar*)g_dir_read_name(session_dir);
    if (!dent) break;
    full_file_path= g_strdup_printf("%s/%s", session_dir_path, dent);

    g_debug("full_file_path=%s", full_file_path);

    if (!g_file_get_contents(full_file_path, &file_contents, NULL, error)) { pop_cred(); break; }
    pop_cred();

    g_assert(file_contents);    
    dbus_keys= g_ptr_array_new();

    //TODO: subroutine and use regex instead
    {
      gchar **lines;
      gchar **sp;

      lines= g_strsplit(file_contents, "\n", 0);
      for (sp = lines ; *sp ; sp++) {

	g_debug("line is: \"%s\"", *sp);

	/* technically, this test isn't quite correct,
	 * it looks for the last = in the string and
	 * makes sure that it doesn't start with #
	 * lines of the format:
	 *   foo=bar # comment
	 * will be inserted as FOO=bar # comment
	 */
	if ((g_strrstr(*sp, "=")) && (*sp[0] != '#')) {
	  gchar *key_dup;
	  
	  g_debug("inserting \"%s\"", *sp);
	  key_dup= g_strdup(*sp);
	  g_ptr_array_add(dbus_keys, (gpointer)key_dup);	 
	}
      }
      g_strfreev(lines);      
    }

    g_free(file_contents);
    g_free(full_file_path);
    g_dir_close(session_dir);
    
    ret= TRUE;
  } while(0);
  
  return ret;
}

static void
for_each_callback(gchar* key, SandboxerProcess *self)
{
  g_assert(self);
  g_assert(key);
  
  g_debug("adding \"%s\" to env", key);
  sandboxer_process_add_to_env_pair(self, key);
}

static gboolean
process_pre_fork(SandboxerProcess *self, GError **error)
{
  SandboxerProcessClass* klass;

  g_assert(self);
  klass= SANDBOXER_PROCESS_GET_CLASS(self); g_assert(klass);
  g_assert(dbus_keys);
  g_ptr_array_foreach(dbus_keys, (GFunc)for_each_callback, 
		      (gpointer)self);
 
  return TRUE;
}

static const SandboxerPluginDescriptor plugin_descriptor= {
  .plugin_name= "dbus",
  .minor_version= PLUGIN_MINOR_VERSION,
  .major_version= PLUGIN_MAJOR_VERSION,
  .process_class_init_hook= (SandboxerProcessClassHookFunc)generic_class_init,
  .process_hooks= {
    NULL,    /* INSTANCE_INIT     */
    process_pre_fork,
    NULL,    /* PRE_UNSHARE       */
    NULL,    /* PRE_PIVOT_ROOT    */
    NULL,    /* PRE_UMOUNT        */
    NULL,    /* PRE_SETUID        */
    NULL,    /* PRE_EXEC          */
    NULL,    /* INSTANCE_FINALIZE */
  },
  .sandbox_class_init_hook= (SandboxerSandboxClassHookFunc)generic_class_init,
  .close_hook= NULL, /* close function */
  .padding= PADDING_INIT
};


const gchar* g_module_check_init(GModule *module)
{
  g_assert(module);
  if (sandboxer_register_plugin(&plugin_descriptor, module)) {
    g_debug("yes"); 
  } else {
    g_debug("no");
  }
  return NULL;
}
