/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <plugins/env_resolver.h>
#include <daemon/plugin.h>
#include <daemon/process.h>
#include <daemon/config_store.h>

//uncomment below to turn on some debugging information for this plugin
//#define ENVPLUGINDEBUG 1

#ifdef ENVPLUGINDEBUG
#include <daemon/sandboxer-helper.h>
#define env_debug(...) g_debug(__VA_ARGS__)
#else
#define env_debug(...) ;
#endif

#define PLUGIN_GROUP    "env_plugin"

//FIXME duplicate vars (PATH) have trouble.  FIX not completely obvious


static gboolean 
process_class_init(SandboxerProcessClass *klass, GError **error);

static void
add_to_env_callback(EnvPair *env_pair,
		    SandboxerProcess *self);

static void
env_to_pairs_callback(gchar *key_value, GQueue *queues[]);

static void
free_env_pair_callback(EnvPair *env_pair, gpointer unused);

static void
copy_env_queue_callback(EnvPair *env_pair, GQueue *q);

static gboolean
process_pre_fork(SandboxerProcess *self, GError **error);

static void
env_to_queued_pairs(GPtrArray *env, GQueue *resolved_pairs, GQueue *unresolved_pairs);



const gchar* g_module_check_init(GModule *module);

static gboolean
env_pairs_from_file(const gchar *filename,
		    GQueue *resolved_pairs, GQueue *unresolved_pairs, 
		    GError **error);


#ifdef ENVPLUGINDEBUG
static void
dump_env_callback(EnvPair *env_pair,
		  void *unused);
#endif



static const SandboxerPluginDescriptor plugin_descriptor= {
  .plugin_name= "env",
  .major_version= PLUGIN_MAJOR_VERSION,
  .minor_version= PLUGIN_MINOR_VERSION,
  .process_class_init_hook= process_class_init,
  .process_hooks= {
    NULL,    /* INSTANCE_INIT     */
    process_pre_fork,
    NULL,    /* PRE_UNSHARE       */
    NULL,    /* PRE_PIVOT_ROOT    */
    NULL,    /* PRE_UMOUNT        */
    NULL,    /* PRE_SETUID        */
    NULL,    /* PRE_EXEC          */
    NULL,    /* INSTANCE_FINALIZE */
  },
  .close_hook= NULL, /* close function */
  .padding= PADDING_INIT
};
static GQueue *unresolved_pairs;
static GQueue *resolved_pairs;


/*
 * env_pairs_from_file: look in a keyfile and grab out environment variables 
 *                      for this plugin.
 *
 * @filename: the pathname to a #GKeyFile containing a group PLUGIN_GROUP
 *            and keys, where "key=value"
 *
 * @resolved_pairs: a #GQueue to put #EnvPair's in if they have no interpolation
 *
 * @resolved_pairs: a #GQueue to put #EnvPair's in if they have no interpolation
 *
 * @error: a location to return an error or %NULL
 *
 * Returns: %TRUE for success; %FALSE for error
 *
 */
static gboolean
env_pairs_from_file(const gchar *filename,
		    GQueue *resolved_pairs, GQueue *unresolved_pairs, 
		    GError **error)
{
  GKeyFile *keyfile;
  gchar **env_keys;
  gsize num_env_keys;
  guint i;
  GError *merror= NULL;

  g_assert(filename); g_assert(resolved_pairs); g_assert(unresolved_pairs);
  keyfile= g_key_file_new();
  do {
    if (!g_key_file_load_from_file(keyfile,
				   filename,
				   G_KEY_FILE_NONE,
				   &merror)) break;
    env_keys= g_key_file_get_keys(keyfile,
				  PLUGIN_GROUP,
				  &num_env_keys,
				  &merror);
    if (!env_keys) break;
  } while(0);
  if (merror) {
    g_propagate_prefixed_error(error,
			       merror,
			       "env_plugin: ");
    g_error_free(merror);
    return FALSE;
  }
  
  for (i= 0; i < num_env_keys; i++) {
    EnvPair *env_pair;
    gchar *value;
    
    value= g_key_file_get_string(keyfile,
				 PLUGIN_GROUP,
				 env_keys[i],
				 &merror);
    if (!value) {
      g_propagate_prefixed_error(error,
				 merror,
				 "env_plugin: ");
      g_error_free(merror);
      return FALSE;
    }
    env_pair= env_pair_new(env_keys[i], value); g_assert(env_pair);
    
    if (g_strrstr(value, "$")) {
      env_debug("adding %s=%s to unresolved list, since it has $ in it",
	      env_pair->key, env_pair->value);
      g_queue_push_tail(unresolved_pairs, (gpointer)env_pair);
    } else {
      env_debug("adding %s=%s to resolved list since it does not have $ in it",
	      env_pair->key, env_pair->value);
      g_queue_push_tail(resolved_pairs, (gpointer)env_pair);
    }
    g_free(value);
  }
  g_strfreev(env_keys);
  g_key_file_free(keyfile);

  return TRUE;
}

static gboolean 
process_class_init(SandboxerProcessClass *klass, GError **error)
{
  SandboxGlobalConfig *global_config;
  GError *merror= NULL;
  
  g_assert(klass);
  global_config= sandbox_global_config_new(NULL);
  g_assert(global_config);

  unresolved_pairs=     g_queue_new();
  resolved_pairs=       g_queue_new();

  if (!env_pairs_from_file(global_config->global_config_file_path,
			   resolved_pairs, unresolved_pairs, 
			   error))
    return FALSE;


  if (!resolve(resolved_pairs, unresolved_pairs, 
	       resolved_pairs, NULL,
	       &merror)) {
    g_propagate_prefixed_error(error, merror, "env_plugin: ");
    return FALSE;
  }
  return TRUE;
}


static void
add_to_env_callback(EnvPair *env_pair,
		    SandboxerProcess *self)
{
  static guint i= 0;
  
  i++;
  env_debug("%s #%d", __FUNCTION__, i);
  g_assert(self);
  if (env_pair) {
    g_assert(env_pair->key); g_assert(env_pair->value);
    env_debug("adding \"%s\"=\"%s\" to env", env_pair->key, env_pair->value);
    sandboxer_process_add_to_env(self, env_pair->key, env_pair->value);
  } else {
    env_debug("env_pair is NULL!");
  }
}

#ifdef ENVPLUGINDEBUG
static void
dump_env_callback(EnvPair *env_pair,
		  void *unused)
{
  g_assert(env_pair); g_assert(env_pair); 
  g_assert(env_pair->key); g_assert(env_pair->value);
  env_debug("key=\"%s\"\tvalue=\"%s\"", env_pair->key, env_pair->value);
}
#endif

static void
env_to_pairs_callback(gchar *key_value, GQueue *queues[])
{
  EnvPair *env_pair;
  gchar **env_split;

  g_assert(queues); g_assert(queues[0]); g_assert(queues[1]);
  if (NULL == key_value) return;

  env_split= g_strsplit(key_value, "=", 2);
  if (NULL == env_split[1]) {
    g_warning("environment variable string \"%s\" does not have a = in it", key_value);
  }  
  env_pair= env_pair_new(env_split[0], env_split[1]); g_assert(env_pair);
  g_strfreev(env_split);
  if (g_strrstr(env_pair->value, "$")) {
    env_debug("adding self->env element %s to unresolved pairs, since it has a $ in it",
	    key_value);
    g_queue_push_tail(queues[1], (gpointer)env_pair);
  } else {
    env_debug("adding self->env element %s to resolved pairs, since it does not have a $ in it",
	    key_value);
    g_queue_push_tail(queues[0], (gpointer)env_pair);
  }
}

static void
free_env_pair_callback(EnvPair *env_pair, gpointer unused)
{
  env_pair_free(env_pair);
}

static void
copy_env_queue_callback(EnvPair *env_pair, GQueue *q)
{
  g_assert(env_pair); g_assert(q);
  g_queue_push_tail(q, env_pair_copy(env_pair));
}

static void
env_to_queued_pairs(GPtrArray *env, GQueue *resolved_pairs, GQueue *unresolved_pairs)
{
  GQueue *queues[2];
  
  g_assert(env); g_assert(resolved_pairs); g_assert(unresolved_pairs);
  queues[0]= resolved_pairs; queues[1]= unresolved_pairs;
  g_ptr_array_foreach(env, (GFunc)env_to_pairs_callback, (gpointer)queues);  
}



//FIXME: this is sort of awkward.  We should perhaps add sconfig to 
// a process property or perhaps private property and and avoid a double-hash lookup?
static gboolean
process_pre_fork(SandboxerProcess *self, GError **error)
{
  SandboxConfig *sconfig;
  SandboxConfigStore *config_store;
  SandboxerProcessClass *klass;
  GQueue garbage_pairs= G_QUEUE_INIT;
  GQueue my_resolved_pairs= G_QUEUE_INIT;
  GQueue my_unresolved_pairs= G_QUEUE_INIT;
  
  g_assert(self);
  klass= SANDBOXER_PROCESS_GET_CLASS(self); g_assert(klass);

  config_store= sandbox_config_store_new(NULL); g_assert(config_store);
  sconfig= sandbox_config_store_lookup_path(config_store, self->path); 
  g_assert(sconfig);
  
  g_queue_foreach(resolved_pairs, 
		  (GFunc)copy_env_queue_callback, 
		  &my_resolved_pairs);
  g_queue_foreach(unresolved_pairs, 
		  (GFunc)copy_env_queue_callback, 
		  &my_unresolved_pairs);
  

  if (sconfig->extra_config_groups) {
    if (!env_pairs_from_file(sconfig->package_file_path,
			     &my_resolved_pairs, &my_unresolved_pairs, 
			     error))
      return FALSE;
  }

  env_to_queued_pairs(self->env, &my_resolved_pairs, &my_unresolved_pairs);

#ifdef ENVPLUGINDEBUG
  env_debug("resolved pairs:");   g_queue_foreach(&my_resolved_pairs, (GFunc)dump_env_callback, NULL);
  env_debug("unresolved_pairs:"); g_queue_foreach(&my_unresolved_pairs, (GFunc)dump_env_callback, NULL);
#endif 

  if (!resolve(&my_resolved_pairs, &my_unresolved_pairs,
	       &my_resolved_pairs, &garbage_pairs,
	       error)) {
    return FALSE;
  }

#ifdef ENVPLUGINDEBUG
  env_debug("post resolve");
  env_debug("resolved pairs:"); g_queue_foreach(&my_resolved_pairs, (GFunc)dump_env_callback, NULL);
  env_debug("unresolved_pairs:"); g_queue_foreach(&my_unresolved_pairs, (GFunc)dump_env_callback, NULL);
  env_debug("garbage_pairs:"); g_queue_foreach(&garbage_pairs, (GFunc)dump_env_callback, NULL);
#endif

  g_ptr_array_free(self->env, TRUE);
  self->env= g_ptr_array_sized_new(g_queue_get_length(resolved_pairs)); 
  g_queue_foreach(&my_resolved_pairs, (GFunc)add_to_env_callback, self); 
  /* NOTE: unresolved items are not included in env.  In sh, unresolved items
   *       are actually set as "", but we decline to do so
   */
  g_queue_foreach(&garbage_pairs, (GFunc)free_env_pair_callback, NULL);
  g_queue_foreach(&my_resolved_pairs, (GFunc)free_env_pair_callback, NULL);
  g_queue_foreach(&my_unresolved_pairs, (GFunc)free_env_pair_callback, NULL);
  return TRUE;
}

const gchar* g_module_check_init(GModule *module)
{
  g_assert(module);
  if (sandboxer_register_plugin(&plugin_descriptor, module)) {
    env_debug("yes"); 
  } else {
    env_debug("no");
  }
  return NULL;
}
