/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <plugins/env_resolver.h>


typedef struct _EvalContext EvalContext;
struct _EvalContext {
  GHashTable *env_hash;
  gboolean not_found;
};


static void
resolve_hash_insert_callback(EnvPair *pair, GHashTable *env_hash);

static gboolean
regex_eval_callback(const GMatchInfo *match_info,
		    GString *result,
		    EvalContext *context);




static void
resolve_hash_insert_callback(EnvPair *pair, GHashTable *env_hash)
{
  g_assert(env_hash); g_assert(pair);

  g_hash_table_insert(env_hash, pair->key, pair->value);
}

static gboolean
regex_eval_callback(const GMatchInfo *match_info,
		    GString *result,
		    EvalContext *context)
{
  gchar *value;
  gchar *key;

  g_assert(match_info);
  g_assert(result);
  g_assert(context);

  key= g_match_info_fetch(match_info, 0);
  value= g_hash_table_lookup(context->env_hash, key + 1); /* take off the leadbing $ */
  if (value) {
    g_string_append(result, value);
    context->not_found= FALSE;
    return TRUE;
  }
  context->not_found= TRUE;
  return FALSE;
}

/**
 * resolve: Resolves interpolated environment variables
 *
 * @resolved_pairs: a #GQueue containing #EnvPairs that are non-interpolated
 *
 * @unresolved_pairs: a #GQueue containing #EnvPairs that are interpolated
 *
 * @newly_resolved_pairs: a #GQeue that will be populated with #EnvPairs that are no
 *                        longer interpolated.  These #EnvPairs may be freed with env_pair_free()
 *
 * @garbage_pairs: a #GQueue with the "waste" #EnvPairs that were resolved.  If %NULL, then leaks garbage.
 *
 * @error: return location for a #GError
 *
 * Return value: %TRUE for success; %FALSE for failure
 *
 */
//FIXME: could be made more efficient by recycling GList element
gboolean
resolve(GQueue *resolved_pairs, GQueue *unresolved_pairs,
	GQueue *newly_resolved_pairs, GQueue *garbage_pairs,
	GError **error)
{
  EvalContext context;
  guint repetitions= 0;
  GRegex *regex= NULL;
  GError *merror= NULL;

  g_assert(resolved_pairs); g_assert(unresolved_pairs);
  g_assert(newly_resolved_pairs);

  context.env_hash= g_hash_table_new_full(g_str_hash, g_str_equal,
					  NULL, NULL);
  g_assert(context.env_hash);
  g_queue_foreach(resolved_pairs,
		  (GFunc)resolve_hash_insert_callback,
		  (gpointer)context.env_hash);
  do {
    EnvPair *unresolved_pair;
    gchar *resolved_string;
    
    unresolved_pair= (EnvPair*)g_queue_pop_head(unresolved_pairs);
    if (!unresolved_pair) break;
    regex= g_regex_new("\\$[\\w]+", 0, 0, &merror);
    if (merror) break;
    resolved_string= g_regex_replace_eval(regex, unresolved_pair->value,
					  -1, 0, 0, 
					  (GRegexEvalCallback)regex_eval_callback,
					  &context,
					  &merror);
    if (merror) break;
    if (context.not_found) {
      g_queue_push_tail(unresolved_pairs, (gpointer)unresolved_pair);
      repetitions++;
    } else {
      EnvPair *resolved_pair;

      resolved_pair= env_pair_new(unresolved_pair->key, resolved_string);
      g_hash_table_insert(context.env_hash, resolved_pair->key, resolved_pair->value);
      g_queue_push_head(newly_resolved_pairs, (gpointer)resolved_pair);
      if (garbage_pairs)
	g_queue_push_head(garbage_pairs, (gpointer)unresolved_pair);
      repetitions= 0;
    }
    g_free(resolved_string);
  } while (repetitions <= g_queue_get_length(unresolved_pairs)); 
  if (merror) {
    g_propagate_prefixed_error(error, merror, "resolve");
    g_clear_error(&merror);
    return FALSE;
  }

  if (regex) g_regex_unref(regex);
  g_hash_table_destroy(context.env_hash);
  return TRUE;
}

EnvPair* 
env_pair_new(gchar *key, gchar *value)
{
  EnvPair *self;
  
  g_assert(key); g_assert(value);
  self= g_new(EnvPair, 1); g_assert(self);
  self->key= g_strdup(key);
  self->value= g_strdup(value);

  return self;
}

EnvPair*
env_pair_copy(EnvPair *env_pair)
{
  g_assert(env_pair);
  return env_pair_new(env_pair->key, env_pair->value);
}

void 
env_pair_free(EnvPair *self)
{
  if (self) {
    g_free(self->key);
    g_free(self->value);
    g_free(self);
  }
}
