/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <daemon/sandboxer-helper.h>
#include <daemon/plugin.h>
#include <daemon/process.h>
#include <daemon/base.h>

#include <glob.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <pwd.h>

//uncomment below to turn on some debugging information for this plugin
//#define XAUTHPLUGINDEBUG 1
#ifdef XAUTHPLUGINDEBUG
#define xauth_debug(...) g_debug(__VA_ARGS__)
#else
#define xauth_debug(...) ;
#endif

typedef struct _FileBlob {
  gchar *data;
  gsize length;
  time_t mtime;
} FileBlob;


static gchar *xauth_filename;
static gchar *iceauth_filename;
static gchar *display;
static FileBlob *xauth;
static FileBlob *iceauth;



static gboolean
get_mtime(const gchar *filename, time_t *mtime, GError **error);


static FileBlob*
file_blob_new_from_file(const gchar *filename, GError **error);

static void 
file_blob_free(FileBlob *self);

static FileBlob*
file_blob_new_if_newer(FileBlob *self, const gchar *filename, GError **error);

static gboolean
file_blob_to_file(FileBlob *self, const gchar *filename, GError **error);

static gboolean
load_config_data(SandboxGlobalConfig *global_config, GError **error);

static gboolean
generic_pre_unshare(gpointer unused, GError **error);

static gboolean 
process_pre_exec(SandboxerProcess *self, GError **error);

static gboolean
process_pre_fork(SandboxerProcess *self, GError **error);




static gboolean
get_mtime(const gchar *filename, time_t *mtime, GError **error)
{
  struct stat buf;
  gboolean ret;

  g_assert(filename);
  g_assert(mtime);

  push_cred(0,0);
  if (g_lstat(filename, &buf)) {
    g_set_error(error, 
		SANDBOXER_PROCESS_ERROR,
		SANDBOXER_PROCESS_ERROR_FAILED,
		"lstat of %s: %s", filename, g_strerror(errno));
    ret= FALSE;
  } else {    
    *mtime= buf.st_mtime;
    xauth_debug("mtime of %s is %d", filename, (int)buf.st_mtime);
    ret= TRUE;
  }
  pop_cred();
  return ret;  
}

static gchar*
first_file_from_glob(const gchar *glob_string, GError **error)
{
  glob_t glob_buf;
  gchar *ret;
  
  g_assert(glob_string);
  switch(glob(glob_string, 0, NULL, &glob_buf)) {

  case GLOB_NOSPACE:
  case GLOB_ABORTED:
    //FIXME: set error
    ret= NULL;
    break;
    
  case GLOB_NOMATCH:
    //FIXME: set error
    ret= NULL;
    break;
    
  default: /* 0 -- see GLOB(3) */
    g_assert(glob_buf.gl_pathv[0]); /* there should always be one match */
    ret= g_strdup(glob_buf.gl_pathv[0]);
    xauth_debug("xauth file= %s", ret);
    globfree(&glob_buf);
    break;
  }
  return ret;
}

static gboolean
load_config_data(SandboxGlobalConfig *global_config, GError **error)
{
  const gchar *xauth_group= "xauth";
  const gchar *xauth_key= "xauth_file";
  const gchar *iceauth_key= "iceauth_file";
  GKeyFile *keyfile;
  GError *merror= NULL;
  
  xauth_debug("load_config_data entry");
  g_assert(global_config);
  keyfile= g_key_file_new(); g_assert(keyfile);

  push_cred(0,0);
  do {
    if (!g_key_file_load_from_file(keyfile, global_config->global_config_file_path,
				   G_KEY_FILE_NONE, &merror)) break;

    if (g_key_file_has_key(keyfile, xauth_group, xauth_key, &merror)) {
      gchar *key;

      key= g_key_file_get_string(keyfile, xauth_group, xauth_key, &merror);
      if (merror) break;

      xauth_filename= first_file_from_glob(key, &merror);
      xauth_debug("xauth filename is %s", xauth_filename);
      g_free(key);
      }
    if (merror) break;
    
    if (g_key_file_has_key(keyfile, xauth_group, iceauth_key, &merror))
      iceauth_filename= g_key_file_get_string(keyfile, xauth_group, iceauth_key, &merror);
    if (merror) break;

    display= g_key_file_get_value(keyfile, xauth_group, "display", &merror);
    if (!display) break;

    xauth_debug("xauth_file= %s", xauth_filename);
    xauth_debug("iceauth_file= %s", iceauth_filename);
    xauth_debug("display= %s", display);

    g_key_file_free(keyfile);
  } while(0);
  pop_cred();
  if (merror) {
    g_propagate_error(error, merror);
    xauth_debug("load_config_data exit: failure");
    return FALSE;
  }
  xauth_debug("load_config_data exit: success");
  return TRUE;
}

static gboolean
generic_pre_unshare(gpointer unused, GError **error)
{
  GError *merror= NULL;

  xauth_debug("generic_pre_unshare entry");
  if (iceauth_filename) {
    iceauth= file_blob_new_if_newer(iceauth, iceauth_filename, &merror);
    if (merror) {
      g_propagate_error(error, merror);
      xauth_debug("generic_pre_unshare exit: failure");
      return FALSE;
    }
  }
  if (xauth_filename) {
    //xauth= xauth_new_if_newer(xauth, xauth_filename, display, &merror);
    xauth= file_blob_new_if_newer(xauth, xauth_filename, &merror);
    if (merror) {
      g_propagate_error(error, merror);
      xauth_debug("generic_pre_unshare exit: failure");
      return FALSE;
    }    
  }
  xauth_debug("iceauth is %sset", iceauth ? "" : "un");
  xauth_debug("xauth   is %sset", xauth ? "" : "un");
  xauth_debug("generic_pre_unshare exit: success");
  return TRUE;
}

static FileBlob*
file_blob_new_from_file(const gchar *filename, GError **error)
{
  FileBlob *self;
  gboolean err;

  g_assert(filename);
  self= g_slice_new(FileBlob);
  err= TRUE;
  do {
    if (!get_mtime(filename, &(self->mtime), error)) break;
    if (!g_file_get_contents(filename, &(self->data), &(self->length), error)) break;

    err= FALSE;
  } while (0);
  if (err) {
    g_slice_free(FileBlob, self);
    return NULL;
  }
  return self;
}

static FileBlob*
file_blob_new_if_newer(FileBlob *self, const gchar *filename, GError **error)
{
  time_t mtime;

  g_assert(filename);
  if (!self)
    return file_blob_new_from_file(filename, error);
  
  if (!get_mtime(filename, &mtime, error))
    return NULL;

  if (mtime > self->mtime) {
    file_blob_free(self);
    return file_blob_new_from_file(filename, error);
  }
  
  return self;
}

static gboolean
file_blob_to_file(FileBlob *self, const gchar *filename, GError **error)
{
  g_assert(self);
  g_assert(filename);
  
  return g_file_set_contents(filename, self->data, self->length, error);
}

static void 
file_blob_free(FileBlob *self)
{
  if (self) {
    if (self->data) g_free(self->data);
    g_slice_free(FileBlob, self);
  }
}


static gboolean 
process_class_init(SandboxerProcessClass *klass, GError **error)
{
  xauth_debug("process_class_init entry");
  g_assert(klass);

  return load_config_data(klass->parent.global_config, error);
}

static gboolean 
sandbox_class_init(SandboxerSandboxClass *klass, GError **error)
{
  xauth_debug("sandbox_class_init entry");
  g_assert(klass);
  
  return load_config_data(klass->parent.global_config, error);
}

/* recall in PRE_EXEC, we are running as the right UID */
static gboolean 
process_pre_exec(SandboxerProcess *self, GError **error)
{
  struct passwd *pw_buf;
  GError *merror= NULL;
  
  xauth_debug("process_pre_exec entry");
  g_assert(self);
  
  pw_buf= getpwuid(self->uid);
  if (!pw_buf) {
    merror= g_error_new(SANDBOXER_PROCESS_ERROR,
			SANDBOXER_PROCESS_ERROR_FAILED,
			"couldn't find passwd entry for %d", self->uid);
  }  
  
  if (!merror && xauth) {
    gchar *filename;

    filename= g_strdup_printf("%s/.Xauthority", pw_buf->pw_dir);    
    file_blob_to_file(xauth, filename, &merror);
    g_free(filename);    
  }
  if (!merror && iceauth) {
    gchar *filename;
    
    filename= g_strdup_printf("%s/.ICEauthority", pw_buf->pw_dir);
    file_blob_to_file(iceauth, filename, &merror);
    g_free(filename);
  }
  
  if (merror) {
    g_propagate_error(error, merror);
    xauth_debug("process_pre_exec exit: failure");
    return FALSE;
  }

  xauth_debug("process_pre_exec exit: success");
  return TRUE;
}
 

/* we need to set $DISPLAY */
static gboolean
process_pre_fork(SandboxerProcess *self, GError **error)
{
  xauth_debug("process_pre_fork entry");
  g_assert(self);

  g_assert(display); xauth_debug("DISPLAY=%s", display);
  sandboxer_process_add_to_env(self, "DISPLAY", display);

  xauth_debug("process_pre_fork exit");
  return TRUE;
}

static const SandboxerPluginDescriptor plugin_descriptor= {
  .plugin_name= "xauth",
  .major_version= PLUGIN_MAJOR_VERSION,
  .minor_version= PLUGIN_MINOR_VERSION,
  .process_class_init_hook= process_class_init,
  .sandbox_class_init_hook= sandbox_class_init,
  .process_hooks= {
    NULL,    /* INSTANCE_INIT     */
    process_pre_fork,
    (SandboxerProcessHookFunc)generic_pre_unshare,
    NULL,    /* PRE_PIVOT_ROOT    */
    NULL,    /* PRE_UMOUNT        */
    NULL,    /* PRE_SETUID        */
    process_pre_exec,
    NULL,    /* INSTANCE_FINALIZE */    
  },
  .sandbox_hooks= {
    NULL,    /* INSTANCE_INIT     */
    NULL,    /* PRE_FORK          */
    (SandboxerSandboxHookFunc)generic_pre_unshare,
    NULL,    /* PRE_PIVOT_ROOT    */
    NULL,    /* PRE_UMOUNT        */
  },
  .close_hook= NULL, /* close function */
  .padding= PADDING_INIT
};


const gchar* g_module_check_init(GModule *module)
{
  g_assert(module);
  if (sandboxer_register_plugin(&plugin_descriptor, module)) {
    xauth_debug("xauth plugin loaded"); 
  } else {
    xauth_debug("xauth plugin refused to load");
  }
  return NULL;
}
