/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <test/check_client_server.h>
#include <test/config_file_helper.h>
#include <daemon/sandboxer-helper.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <glib.h>

#define MS G_USEC_PER_SEC / 1000
#define TIMES 3

extern uid_t sandboxer_uid;
extern gid_t sandboxer_gid;
extern uid_t tester_uid;
extern gid_t tester_gid;

pid_t server_pid;

static void
test_configs_start_server(void);

static void
start_server(void);

static void
kill_server(void);

static const gchar *
get_sandboxer(void);



/*
 * NOTE:
 *  this is an unchecked fixture meaning that
 *  it is in the same process as the test 
 *  runners.  If it crashes, the tests
 *  will crash.  
 *
 */
static void
start_server(void)
{
  gchar *package_file_dir;
  char* argv[4];
  uid_t saved;

  argv[0]= g_strdup_printf("%s/sandboxerd", get_bindir());
  argv[1]= "-c";
  argv[2]= (gchar*)get_sandboxerd_conf();
  argv[3]= NULL;
  
  push_cred(0, 0);
  nukem();
  server_pid= fork();
  switch(server_pid)
    {
    case 0:
      server_pid= getpid(); g_debug("server_pid=%d", server_pid);

      i_am_here();
      execv(argv[0], (char**)argv);
      fail("control should not reach here: %s", g_strerror(errno));
      break;
      
    case -1:
      fail("couldn't fork");
      break;
      
    default:
      g_usleep(100*MS);
      g_debug("***** (START) server pid is %d *********", server_pid);
      break;
    }  
  pop_cred();
}


static void
kill_server(void)
{
  int exit_status;
  int ret;
  uid_t saved;

  g_debug("***** (KILL) server pid is %d *********", server_pid);



  push_cred(0, 0);
  ret= waitpid(server_pid, &exit_status, WNOHANG);
  fail_if(-1 == server_pid, "waitpid error: %s", g_strerror(errno));
  if (server_pid == ret) {
    fail_if(WIFEXITED(exit_status), "server exited with status: %d", WEXITSTATUS(exit_status));
    fail_if(WIFSIGNALED(exit_status), "server died with signal: %d", WTERMSIG(exit_status));
  }
  fail_unless(0 == kill(server_pid, SIGKILL), "couldn't kill %d", server_pid);
  fail_unless(server_pid == waitpid(server_pid, &exit_status, 0), "wait returned error");
  fail_unless(0 == WEXITSTATUS(exit_status), "server returned %d", WEXITSTATUS(exit_status));

  nukem();

  ret= waitpid(server_pid, &exit_status, WNOHANG);
  fail_if(0 == ret, "server is still running");  

  pop_cred();
}

//throw it in another process b/c it runs as an unchecked fixture- that is in this process
//FIXME: what does the above mean?
static void
test_configs_start_server(void)
{
  setup_tmpdir_and_config_store();
  start_server();
}

static const gchar *
get_sandboxer(void)
{
  static gchar *sandboxer= NULL;

  if (!sandboxer) {
    sandboxer= g_strdup_printf("%s/sandboxer", get_bindir());
  }
  return sandboxer;
}


START_TEST(client_server_return_value)
{
  int ret;
  guint i;
  gchar *command_line;

  command_line= g_strdup_printf("%s/sandboxer %s/return_value 45", get_bindir(), get_tmpdir());

  for (i= 0; i < TIMES; i++) { 
    g_debug("i= %d (%s)", i, __FUNCTION__);
    push_cred(tester_uid, tester_gid);
    ret= system(command_line);
    pop_cred();
    fail_unless(45 == WEXITSTATUS(ret), "%d is not 45", ret);
  }
  g_free(command_line);
}
END_TEST

START_TEST(client_server_add_package)
{
  const gchar *new= "/tmp/coreutils.package";
  gchar *orig;
  gchar *sandboxer;
  gchar *add_package;
  gchar *command_line;
  int ret;

  orig= g_strdup_printf("%s/coreutils.package", get_packagedir());
  sandboxer= g_strdup_printf("%s/sandboxer", get_bindir());
  add_package= g_strdup_printf("%s/sandboxer_add_package", get_bindir());
  fail_unless(g_file_test(orig, G_FILE_TEST_EXISTS), "doesn't exist?!");

  fail_unless(g_file_test(orig, G_FILE_TEST_EXISTS), "doesn't exist?!");  
  ret= system(g_strdup_printf("mv %s %s", orig, new));
  fail_unless(0 == WEXITSTATUS(ret), "mv returned %d", WEXITSTATUS(ret));
  
  start_server();
  command_line= g_strdup_printf("%s %s", add_package, new);
  push_cred(tester_uid, tester_gid);
  ret= system(command_line);
  fail_unless(0 == WEXITSTATUS(ret), "add_package returned %d", WEXITSTATUS(ret));
  pop_cred();
  g_free(command_line);
  
  command_line= g_strdup_printf("%s /bin/true", sandboxer);
  ret= system(command_line);
  fail_unless(0 == WEXITSTATUS(ret), "true did not return true");

  kill_server();
}
END_TEST

START_TEST(client_server_remove_package)
{
  gchar *orig;
  gchar *remove_package;
  gchar *command_line;
  int ret;

  orig= g_strdup_printf("%s/coreutils.package", get_packagedir());
  fail_unless(g_file_test(orig, G_FILE_TEST_EXISTS), "doesn't exist?!");

  remove_package= g_strdup_printf("%s/sandboxer_remove_package", get_bindir());
  command_line= g_strdup_printf("%s coreutils", remove_package);

  start_server();

  push_cred(tester_uid, tester_gid);
  ret= system(command_line);
  fail_unless(0 == WEXITSTATUS(ret), "remove_package returned %d", WEXITSTATUS(ret));
  pop_cred();
  fail_if(TRUE == g_file_test(orig, G_FILE_TEST_EXISTS), "file exists!"); 
  
  kill_server();
  g_free(orig); g_free(remove_package); g_free(command_line);
}
END_TEST

START_TEST(client_server_run_date)
{ 
  int ret;
  guint i;

  push_cred(tester_uid, tester_gid);
  for (i= 0; i < TIMES; i++) {
    gchar *command_line;

    command_line= g_strconcat(get_sandboxer(), " /bin/date", NULL);
    ret= system(command_line);
    fail_unless(0 == WEXITSTATUS(ret), "%d is not 0", ret); 
    
    g_free(command_line);
  }
  pop_cred();
  
}
END_TEST

START_TEST(client_server_run_date_uname)
{
 int ret;
 guint i;
 gchar *date_command_line;
 gchar *uname_command_line;

 date_command_line= g_strconcat(get_sandboxer(), " /bin/date", NULL);
 uname_command_line= g_strconcat(get_sandboxer(), " /bin/uname -a", NULL);
 
 push_cred(tester_uid, tester_gid);
 for (i= 0; i < TIMES; i++) {
   ret= system(date_command_line);
   fail_unless(0 == WEXITSTATUS(ret), "%d is not 0", ret); 
   ret= system(uname_command_line);
   fail_unless(0 == WEXITSTATUS(ret), "%d is not 0", ret); 
 }
 pop_cred();
}
END_TEST

START_TEST(client_server_run_multiple)
{
  gchar *date_command_line;
  gchar *true_command_line;
  gchar *false_command_line;
  gchar *return_value_command_line;
  int ret;
  guint i;

  date_command_line= g_strconcat(get_sandboxer(), " /bin/date", NULL);
  true_command_line= g_strconcat(get_sandboxer(), " /bin/true", NULL);
  false_command_line= g_strconcat(get_sandboxer(), " /bin/false", NULL);
  return_value_command_line= g_strconcat(get_sandboxer(), " ", get_tmpdir(), "/return_value 45", NULL);

  g_debug(" ******  run_multiple ******");
  push_cred(tester_uid, tester_gid);
  
  for (i= 0; i < TIMES; i++) {
    ret= system(date_command_line);
    fail_unless(0 == WEXITSTATUS(ret), "date returned %d instead of 0", ret); 
    ret= system(true_command_line);
    fail_unless(0 == WEXITSTATUS(ret), "true did not return true");
    ret= system(false_command_line);
    fail_unless(1 == WEXITSTATUS(ret), "false did not return false");
    ret= system(return_value_command_line);
    fail_unless(45 == WEXITSTATUS(ret), "return value retruned %d instead of 45", ret);
  }
  sleep(10);
  pop_cred();
}
END_TEST

START_TEST(server_setup_teardown)
{
  printf("test start\n");
  sleep(3); /* server should not be dead when we run */
  printf("test end\n");
}
END_TEST

Suite *
client_server_suite(void)
{
  Suite *s;
  TCase *tc;

  s= suite_create("Client/Server");

  tc = tcase_create ("Testing");
  tcase_add_test (tc, server_setup_teardown);
  tcase_add_unchecked_fixture(tc, test_configs_start_server, kill_server);
  suite_add_tcase (s, tc);

  tc = tcase_create ("Sandboxer");
  tcase_add_unchecked_fixture(tc, test_configs_start_server, kill_server);
  tcase_add_test (tc, client_server_return_value);
  suite_add_tcase (s, tc);
 
  tc = tcase_create ("Add Package");
  tcase_add_checked_fixture(tc, tmpdir_setup, tmpdir_teardown);
  tcase_add_test (tc, client_server_add_package);
  suite_add_tcase (s, tc);

  tc = tcase_create ("Remove Package");
  tcase_add_checked_fixture(tc, tmpdir_setup, tmpdir_teardown);
  tcase_add_test (tc, client_server_remove_package);
  suite_add_tcase (s, tc);

  tc = tcase_create ("Sandboxer Shared Sandbox");
  tcase_add_unchecked_fixture(tc, test_configs_start_server, kill_server);
  tcase_add_test (tc, client_server_run_date);
  tcase_add_test (tc, client_server_run_date_uname);
  suite_add_tcase (s, tc);

  tc = tcase_create ("Sandboxer Multiple");
  tcase_add_unchecked_fixture(tc, test_configs_start_server, kill_server);
  tcase_add_test (tc, client_server_run_multiple);
  suite_add_tcase (s, tc);

  return s;
}

