/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <daemon/config_store.h>
#include <test/check_config_store.h>
#include <test/config_file_helper.h>
#include <unistd.h>
#include <sys/stat.h>

#define TEST_CONFIG_PATH "/tmp/a_test_thingy.package"
#define TEST_CONFIG_CONTENTS			\
  "[Sandbox]\n"					\
  "SandboxName=a_test_thingy\n"			\
  "PackageName=a_test_thingy_package\n"		\
  "Users=nobody\n"				\
  "ExecPaths=/a_test_thingy\n"


#define EMPTY_DIR "./empty_dir"
#define fail_str_ne(x,y)                        \
  {                                             \
    fail_unless(g_str_equal(x, y),              \
		"\"%s\" is not \"%s\"", x, y);	\
  }




static void
ensure_empty_dir(void);


static void
ensure_empty_dir(void)
{
  if (!g_file_test(EMPTY_DIR, G_FILE_TEST_EXISTS)) {
    fail_unless(0 == g_mkdir(EMPTY_DIR, S_IRWXU | S_IRWXG | S_IRWXO), "g_mkdir returned error!");
  } 
}



START_TEST(test_sandbox_config_new)
{
  SandboxConfig *sconfig;

  sconfig= sandbox_config_new();
  fail_if(NULL == sconfig, "sandbox_config_new returned NULL!");
}
END_TEST

/* this test doesn't actually have any tests
 * but we just want to make sure it doesn't
 * segfault
 */
START_TEST(test_sandbox_config_free)
{
  SandboxConfig *sconfig;

  sconfig= sandbox_config_new();
  sandbox_config_free(sconfig);
}
END_TEST


START_TEST(test_sandbox_config_new_from_file_valid1)
{
  SandboxConfig *sconfig;
  GError *error= NULL;
  gchar *path;
  
  path= g_strdup_printf("%s/test1.package", get_packagedir());
  mark_point();
  sconfig= sandbox_config_new_from_file(path, &error);
  mark_point();
  fail_unless(NULL == error, "returned error");
  mark_point();
  fail_if(NULL == sconfig, "returned NULL, with non-NULL error");
  fail_str_ne(sconfig->package_name, "test1_package_name");
  fail_str_ne(sconfig->sandbox_name, "test1_sandbox");
  fail_str_ne(sconfig->package_file_path, (gchar *)path);
  fail_unless(3 == sconfig->num_uids, "%d is not 3", sconfig->num_uids);
  fail_unless(2 == sconfig->num_exec_paths, "%d is not 2", sconfig->num_exec_paths);
}
END_TEST

START_TEST(test_sandbox_config_store_new_empty_path)
{
  SandboxConfigStore *scs;
  
  scs= sandbox_config_store_new(EMPTY_DIR);
  fail_if(NULL == scs, "sandbox_config_store_new returned NULL");
}
END_TEST

START_TEST(test_sandbox_config_store_new_repeated)
{
  SandboxConfigStore *scs1, *scs2;
  
  scs1= sandbox_config_store_new(EMPTY_DIR);
  scs2= sandbox_config_store_new(NULL);
  fail_unless(scs1 == scs2, "sandbox_config_store_new doesn't behave as singleton");
}
END_TEST

START_TEST(test_sandbox_config_store_valid_files)
{
  SandboxConfigStore *scs;

  scs= sandbox_config_store_new(get_packagedir());
  fail_if(NULL == scs, "sandbox_config_store_new returned NULL");
}
END_TEST

START_TEST(test_sandbox_config_store_lookup_path_incorrect_path)
{
  SandboxConfigStore *scs;
  SandboxConfig *sconfig;

  scs= sandbox_config_store_new(get_packagedir());
  sconfig= sandbox_config_store_lookup_path(scs, "a tottally bogus path");
  fail_unless(NULL == sconfig, "failed to return NULL for bogus path");
}
END_TEST

START_TEST(test_sandbox_config_store_lookup_path_correct_path)
{
  SandboxConfigStore *scs;
  SandboxConfig *sconfig;

  scs= sandbox_config_store_new(get_packagedir());
  sconfig= sandbox_config_store_lookup_path(scs, "/bin/true");
  fail_if(NULL == sconfig, "lookup_path returned NULL for a correct path");
  fail_str_ne(sconfig->package_name, "coreutils");
  fail_unless(sconfig == sandbox_config_store_lookup_path(scs, "/bin/false"), "lookup of second path did not give same result");
}
END_TEST

START_TEST(test_sandbox_config_store_lookup_num_sandboxes_more_than_one)
{
  SandboxConfigStore *scs;
  int num_sandboxes;
  
  mark_point();
  scs= sandbox_config_store_new(get_packagedir());
  mark_point();
  num_sandboxes= sandbox_config_store_lookup_num_sandboxes(scs,
							   "test1_sandbox");
  mark_point();
  fail_unless(2 == num_sandboxes, "2!=%d", num_sandboxes);
}
END_TEST

START_TEST(test_sandbox_config_store_lookup_num_sandboxes_none)
{
  SandboxConfigStore *scs;
  int num_sandboxes;
  
  scs= sandbox_config_store_new(get_packagedir());
  num_sandboxes= sandbox_config_store_lookup_num_sandboxes(scs,
							   "random_name");
  fail_unless(0 == num_sandboxes, "0!=%d", num_sandboxes);
}
END_TEST

START_TEST(test_sandbox_config_store_lookup_num_sandboxes_one)
{
  SandboxConfigStore *scs;
  int num_sandboxes;
  
  scs= sandbox_config_store_new(get_packagedir());
  num_sandboxes= sandbox_config_store_lookup_num_sandboxes(scs,
							   "test2_sandbox");
  fail_unless(1 == num_sandboxes, "1!=%d", num_sandboxes);
}
END_TEST

START_TEST(test_sandbox_config_store_lookup_sandbox_valid)
{
  SandboxConfigStore *scs;
  SandboxConfig **sconfigs;
  guint length;

  scs= sandbox_config_store_new(get_packagedir());
  sconfigs= sandbox_config_store_lookup_sandbox(scs, "test1_sandbox", &length);
  fail_if(NULL == sconfigs, 
	  "sandbox_config_store_lookup_sandbox returned NULL!");
  fail_unless(2 == length, "num sandboxes is not 2");
}
END_TEST

START_TEST(test_sandbox_config_store_lookup_sandbox_invalid)
{
  SandboxConfigStore *scs;
  SandboxConfig **sconfigs;
  guint length= 1234;

  scs= sandbox_config_store_new(get_packagedir());
  sconfigs= sandbox_config_store_lookup_sandbox(scs, "random_name", &length);
  fail_unless(NULL == sconfigs, "got valid pointer for invalid name input");
  fail_unless(0 == length, "length is not 0");
}
END_TEST

START_TEST(test_sandbox_global_config_new)
{
  SandboxGlobalConfig *global_config;

  global_config= sandbox_global_config_new(get_sandboxerd_conf());
  fail_if(NULL == global_config, "sandbox_global_config_new returned NULL");
  fail_str_ne(global_config->global_config_file_path, get_sandboxerd_conf());
  //TODO: check sandboxer_user_uid
  fail_unless(NULL == *(global_config->enabled_plugins), 
	      "plugins are enabled, when shouldn't be");
  fail_str_ne(global_config->plugins_path, g_strdup_printf("%s/plugins", get_tmpdir()));
  fail_str_ne(global_config->package_file_directory, get_packagedir());
}
END_TEST


START_TEST(test_sandbox_config_store_add_file_valid)
{
  SandboxConfig *sconfig;
  SandboxConfigStore *scs;
  GError *error= NULL;
  gboolean ret;

  scs= sandbox_config_store_new(NULL);
  fail_if(NULL == scs, "sandbox_config_store_new returned NULL");
  fail_unless(g_file_set_contents(TEST_CONFIG_PATH, TEST_CONFIG_CONTENTS, -1, NULL),
	      "failed to write %s", TEST_CONFIG_PATH);

  ret= sandbox_config_store_add_file(scs, 
				     TEST_CONFIG_PATH,
				     &error);
  mark_point();
  fail_unless(TRUE == ret, "sandbox_config_store_add_file returned FALSE: %s",
	      error ? error->message : "");
  sconfig= sandbox_config_store_lookup_path(scs, "/a_test_thingy");
  fail_if(NULL == sconfig, "couldn't lookup recently added package");
}
END_TEST

START_TEST(test_sandbox_config_store_remove_package_present)
{
  SandboxConfig *sconfig;
  SandboxConfigStore *scs;
  GError *error= NULL;
  gboolean ret;

  scs= sandbox_config_store_new(NULL);
  fail_if(NULL == scs, "sandbox_config_store_new returned NULL");
  ret= sandbox_config_store_remove_package(scs, "coreutils", &error);
  fail_unless(ret, "config_store_remove_packages returned FALSE: %s", 
	      error ? error->message : "");
  fail_unless(NULL == error, "no error but error is not NULL");
  fail_unless(NULL == sandbox_config_store_lookup_path(scs, "/bin/true"),
	      "could still find /bin/true");
  fail_if(g_file_test(g_strdup_printf("%s/coreutils.package", get_packagedir()), G_FILE_TEST_EXISTS),
	  "coreutils.package still exists");
    
}
END_TEST

START_TEST(test_sandbox_config_store_remove_package_not_present)
{
  SandboxConfig *sconfig;
  SandboxConfigStore *scs;
  GError *error= NULL;
  gboolean ret;

  scs= sandbox_config_store_new(NULL);
  fail_if(NULL == scs, "sandbox_config_store_new returned NULL");
  ret= sandbox_config_store_remove_package(scs, "don't exist", &error);
  fail_if(ret, "config_store_remove_packages returned TRUE");
  fail_if(NULL == error, "error was non-NULL");
  fail_unless(SANDBOXER_CONFIG_STORE_ERROR == error->domain, "was not of correct error domain");
}
END_TEST

//FIXME: nuke
START_TEST(do_nothing)
{
  g_debug("nothing");
}
END_TEST


Suite *
config_suite(void)
{
  Suite *s;
  TCase *tc;

  s= suite_create("Config Store");
  tc= tcase_create("Sandbox Config New/Free");
  tcase_add_test (tc, test_sandbox_config_new);
  tcase_add_test (tc, test_sandbox_config_free);
  suite_add_tcase (s, tc);

  tc= tcase_create("Sandbox Config New From File");
  tcase_add_checked_fixture(tc, tmpdir_setup, tmpdir_teardown);
  tcase_add_test (tc, test_sandbox_config_new_from_file_valid1);
  suite_add_tcase (s, tc);
  
  tc= tcase_create("Sandbox Global Config New");
  tcase_add_checked_fixture(tc, tmpdir_setup, tmpdir_teardown);
  tcase_add_test(tc, test_sandbox_global_config_new);
  suite_add_tcase(s, tc);
  

  tc= tcase_create("Sandbox Config Store New");
  tcase_add_checked_fixture(tc, ensure_empty_dir, NULL);
  tcase_add_checked_fixture(tc, tmpdir_setup, tmpdir_teardown);
  tcase_add_test(tc, test_sandbox_config_store_new_empty_path);
  tcase_add_test(tc, test_sandbox_config_store_new_repeated);
  tcase_add_test(tc, test_sandbox_config_store_valid_files);
  suite_add_tcase(s, tc);
  
  tc= tcase_create("Sandbox Config Store Lookup");
  tcase_add_checked_fixture(tc, tmpdir_setup, tmpdir_teardown);
  tcase_add_test(tc, test_sandbox_config_store_lookup_path_incorrect_path);
  tcase_add_test(tc, test_sandbox_config_store_lookup_path_correct_path);
  suite_add_tcase(s, tc);

  tc= tcase_create("Sandbox Config Store Lookup Sandbox");
  tcase_add_checked_fixture(tc, tmpdir_setup, tmpdir_teardown);
  tcase_add_test(tc, test_sandbox_config_store_lookup_sandbox_valid);
  tcase_add_test(tc, test_sandbox_config_store_lookup_sandbox_invalid);
  suite_add_tcase(s, tc);

  tc= tcase_create("Sandbox Config Store Lookup Number of Sandboxes");
  tcase_add_checked_fixture(tc, tmpdir_setup, tmpdir_teardown);
  tcase_add_test(tc, test_sandbox_config_store_lookup_num_sandboxes_more_than_one);
  tcase_add_test(tc, test_sandbox_config_store_lookup_num_sandboxes_none);
  tcase_add_test(tc, test_sandbox_config_store_lookup_num_sandboxes_one);
  suite_add_tcase(s, tc);

  tc= tcase_create("Sandbox Config Store Add File");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, tmpdir_teardown);  
  tcase_add_test(tc, test_sandbox_config_store_add_file_valid);
  suite_add_tcase(s, tc);

  tc= tcase_create("Sandbox Config Store Remove Package");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, tmpdir_teardown);  
  tcase_add_test(tc, test_sandbox_config_store_remove_package_present);
  suite_add_tcase(s, tc);

  tc= tcase_create("test tmpdir thing");
  tcase_add_checked_fixture (tc, tmpdir_setup, tmpdir_teardown);  
  tcase_add_test(tc, do_nothing);
  suite_add_tcase(s, tc);
  

  return s;
}
