/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <test/check_env_resolver.h>
#include <plugins/env_resolver.h>

static gboolean
env_pair_equal(EnvPair *a, EnvPair *b);

static gchar*
env_pair_to_string(EnvPair *self);

static gint
env_pair_compare(EnvPair *a, EnvPair *b);



static gboolean
env_pair_equal(EnvPair *a, EnvPair *b)
{
  g_assert(a); g_assert(b);
  if (g_str_equal(a->key, b->key) && 
      g_str_equal(a->value, b->value))
    return TRUE;
  return FALSE;
}

static gchar*
env_pair_to_string(EnvPair *self)
{
  g_assert(self);
  return g_strdup_printf("%s=%s", self->key, self->value);
}

static gint
env_pair_compare(EnvPair *a, EnvPair *b)
{
  gchar *concata, *concatb;
  gint ret;
  
  concata= env_pair_to_string(a);
  concatb= env_pair_to_string(b);
  ret= g_strcmp0(concata, concatb);
  g_free(concata); g_free(concatb);
  return ret;
}



START_TEST(test_resolve_no_unresolved_pairs)
{
  GQueue *resolved_pairs, *unresolved_pairs;
  GQueue *newly_resolved_pairs, *garbage_pairs;
  GError *error= NULL;

  resolved_pairs= g_queue_new();
  unresolved_pairs= g_queue_new();
  newly_resolved_pairs= g_queue_new();
  garbage_pairs= g_queue_new();

  g_queue_push_head(resolved_pairs, env_pair_new("KEY1", "VALUE1"));
  g_queue_push_head(resolved_pairs, env_pair_new("KEY2", "VALUE2"));
  g_queue_push_head(resolved_pairs, env_pair_new("KEY3", "VALUE3"));
  g_queue_push_head(resolved_pairs, env_pair_new("KEY4", "VALUE4"));
  
  fail_unless(resolve(resolved_pairs, unresolved_pairs,
		      newly_resolved_pairs, garbage_pairs,
		      &error),
	      "resolve_pairs returned FALSE: %s", error ? error->message : error);
  fail_unless(g_queue_is_empty(unresolved_pairs), "unresolved pairs is not empty");
  fail_unless(g_queue_is_empty(newly_resolved_pairs), "newly pairs is not empty");
  fail_unless(g_queue_is_empty(garbage_pairs), "garbage pairs is not empty");  
}
END_TEST

START_TEST(test_resolve_simple_pairs)
{
  GQueue *resolved_pairs, *unresolved_pairs;
  GQueue *newly_resolved_pairs, *garbage_pairs;
  EnvPair *resolved_pair;
  EnvPair correct_pair= {"KEY5", "VALUE3"};
  GError *error= NULL;

  resolved_pairs= g_queue_new();
  unresolved_pairs= g_queue_new();
  newly_resolved_pairs= g_queue_new();
  garbage_pairs= g_queue_new();

  g_queue_push_head(resolved_pairs, env_pair_new("KEY1", "VALUE1"));
  g_queue_push_head(resolved_pairs, env_pair_new("KEY2", "VALUE2"));
  g_queue_push_head(resolved_pairs, env_pair_new("KEY3", "VALUE3"));
  g_queue_push_head(resolved_pairs, env_pair_new("KEY4", "VALUE4"));

  g_queue_push_head(unresolved_pairs, env_pair_new("KEY5", "$KEY3"));

  fail_unless(resolve(resolved_pairs, unresolved_pairs,
		      newly_resolved_pairs, garbage_pairs,
		      &error),
	      "resolve_pairs returned FALSE: %s", error ? error->message : error);

  fail_unless(g_queue_is_empty(unresolved_pairs), "unresolved pairs is not empty");
  
  fail_unless(1 == g_queue_get_length(newly_resolved_pairs), 
	      "newly resolved pairs does not have 1 entry");
  resolved_pair= (EnvPair*)g_queue_pop_tail(newly_resolved_pairs); 
  fail_if(NULL == resolved_pair, "resolved pair is NULL?!");

  fail_unless(env_pair_equal(&correct_pair, resolved_pair),
	      "%s was not %s", 
	      env_pair_to_string(resolved_pair),
	      env_pair_to_string(&correct_pair));
  fail_if(g_queue_is_empty(garbage_pairs), "garbage pairs is empty");  
}
END_TEST

START_TEST(test_reolved_simple_pairs_dep)
{
  GQueue *resolved_pairs, *unresolved_pairs;
  GQueue *newly_resolved_pairs, *garbage_pairs;
  EnvPair correct_pair1= {"KEY5", "VALUE3"};
  EnvPair correct_pair2= {"KEY6", "VALUE3"};
  GError *error= NULL;

  resolved_pairs= g_queue_new();
  unresolved_pairs= g_queue_new();
  newly_resolved_pairs= g_queue_new();
  garbage_pairs= g_queue_new();

  g_queue_push_head(resolved_pairs, env_pair_new("KEY1", "VALUE1"));
  g_queue_push_head(resolved_pairs, env_pair_new("KEY2", "VALUE2"));
  g_queue_push_head(resolved_pairs, env_pair_new("KEY3", "VALUE3"));
  g_queue_push_head(resolved_pairs, env_pair_new("KEY4", "VALUE4"));

  g_queue_push_head(unresolved_pairs, env_pair_new("KEY5", "$KEY3"));
  g_queue_push_head(unresolved_pairs, env_pair_new("KEY6", "$KEY5"));

  fail_unless(resolve(resolved_pairs, unresolved_pairs,
		      newly_resolved_pairs, garbage_pairs,
		      &error),
	      "resolve_pairs returned FALSE: %s", error ? error->message : error);

  fail_unless(g_queue_is_empty(unresolved_pairs), "unresolved pairs is not empty");
  fail_if(NULL == g_queue_find_custom(newly_resolved_pairs, 
				      &correct_pair1, 
				      (GCompareFunc)env_pair_compare),
	  "%s is not in resolved queue", env_pair_to_string(&correct_pair1));
  fail_if(NULL == g_queue_find_custom(newly_resolved_pairs, 
				      &correct_pair2, 
				      (GCompareFunc)env_pair_compare),
	  "%s is not in resolved queue", env_pair_to_string(&correct_pair2));

  fail_unless(2 == g_queue_get_length(garbage_pairs), "weren't 2 items in garbage queue");
}
END_TEST

START_TEST(test_resolved_interpolated_pairs_nodep)
{
  GQueue *resolved_pairs, *unresolved_pairs;
  GQueue *newly_resolved_pairs, *garbage_pairs;
  EnvPair correct_pair= {"KEY5", "VALUE3:foo"};
  GError *error= NULL;

  resolved_pairs= g_queue_new();
  unresolved_pairs= g_queue_new();
  newly_resolved_pairs= g_queue_new();
  garbage_pairs= g_queue_new();

  g_queue_push_head(resolved_pairs, env_pair_new("KEY1", "VALUE1"));
  g_queue_push_head(resolved_pairs, env_pair_new("KEY2", "VALUE2"));
  g_queue_push_head(resolved_pairs, env_pair_new("KEY3", "VALUE3"));
  g_queue_push_head(resolved_pairs, env_pair_new("KEY4", "VALUE4"));

  g_queue_push_head(unresolved_pairs, env_pair_new("KEY5", "$KEY3:foo"));

  fail_unless(resolve(resolved_pairs, unresolved_pairs,
		      newly_resolved_pairs, garbage_pairs,
		      &error),
	      "resolve_pairs returned FALSE: %s", error ? error->message : error);

  fail_unless(g_queue_is_empty(unresolved_pairs), "unresolved pairs is not empty");
  fail_if(NULL == g_queue_find_custom(newly_resolved_pairs, 
				      &correct_pair, 
				      (GCompareFunc)env_pair_compare),
	  "%s is not in resolved queue", env_pair_to_string(&correct_pair));
  fail_unless(1 == g_queue_get_length(garbage_pairs), "weren't 1 items in garbage queue");
}
END_TEST

START_TEST(test_resolved_interpolated_pairs_dep)
{
  GQueue *resolved_pairs, *unresolved_pairs;
  GQueue *newly_resolved_pairs, *garbage_pairs;
  EnvPair correct_pair1= {"KEY5", "VALUE3:foo"};
  EnvPair correct_pair2= {"KEY6", "VALUE3:foo:bar"};
  GError *error= NULL;

  resolved_pairs= g_queue_new();
  unresolved_pairs= g_queue_new();
  newly_resolved_pairs= g_queue_new();
  garbage_pairs= g_queue_new();

  g_queue_push_head(resolved_pairs, env_pair_new("KEY1", "VALUE1"));
  g_queue_push_head(resolved_pairs, env_pair_new("KEY2", "VALUE2"));
  g_queue_push_head(resolved_pairs, env_pair_new("KEY3", "VALUE3"));
  g_queue_push_head(resolved_pairs, env_pair_new("KEY4", "VALUE4"));

  g_queue_push_head(unresolved_pairs, env_pair_new("KEY5", "$KEY3:foo"));
  g_queue_push_head(unresolved_pairs, env_pair_new("KEY6", "$KEY5:bar"));

  fail_unless(resolve(resolved_pairs, unresolved_pairs,
		      newly_resolved_pairs, garbage_pairs,
		      &error),
	      "resolve_pairs returned FALSE: %s", error ? error->message : error);

  fail_unless(g_queue_is_empty(unresolved_pairs), "unresolved pairs is not empty");
  fail_if(NULL == g_queue_find_custom(newly_resolved_pairs, 
				      &correct_pair1, 
				      (GCompareFunc)env_pair_compare),
	  "%s is not in resolved queue", env_pair_to_string(&correct_pair1));
  fail_if(NULL == g_queue_find_custom(newly_resolved_pairs, 
				      &correct_pair2, 
				      (GCompareFunc)env_pair_compare),
	  "%s is not in resolved queue", env_pair_to_string(&correct_pair2));

  fail_unless(2 == g_queue_get_length(garbage_pairs), "weren't 2 items in garbage queue");
}
END_TEST



Suite *
env_resolver_suite(void)
{
  Suite *s;
  TCase *tc;

  s= suite_create("env resolver");
  

  tc = tcase_create ("resolve");
  tcase_add_test(tc, test_resolve_no_unresolved_pairs);
  tcase_add_test(tc, test_resolve_simple_pairs);
  tcase_add_test(tc, test_reolved_simple_pairs_dep);
  tcase_add_test(tc, test_resolved_interpolated_pairs_nodep);
  tcase_add_test(tc, test_resolved_interpolated_pairs_dep);
  suite_add_tcase(s, tc);

  return s;

}
