/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <test/check_helper.h>
#include <daemon/sandboxer-helper.h>
#include <errno.h>

extern gid_t tester_gid;
extern uid_t tester_uid;
extern gid_t sandboxer_gid;
extern uid_t sandboxer_uid;

START_TEST(test_is_root_really_root)
{
  uid_t ruid, euid, suid;

  i_am_here();
  getresuid(&ruid, &euid, &suid);
  fail_unless(0 == ruid, "ruid is not 0");
  fail_unless(0 == euid, "euid is not 0");
  fail_unless(0 == suid, "suid is not 0");
  fail_unless(is_root(), "we are root, but is_root reports we are not");
}
END_TEST

START_TEST(test_is_root_notreally_root)
{
  const uid_t not_root= 1000;
  uid_t ruid, euid, suid;

  fail_if(setresuid(not_root, not_root, -1), "couldnt' turn into %d", not_root);
  fail_if(is_root(), "is_root says we are root, but we are not");
}
END_TEST

START_TEST(test_push_from_root)
{
  uid_t ruid, euid, suid;
  gid_t rgid, egid, sgid;

  fail_unless(0 == getuid(), "you must be root to run this test");
  push_cred(tester_uid, tester_gid);
  fail_unless(0 == getresuid(&ruid, &euid, &suid),
	      "getresuid: %s", g_strerror(errno));
  fail_unless(0 == getresgid(&rgid, &egid, &sgid),
	      "getresgid: %s", g_strerror(errno));

  fail_unless(tester_uid == ruid, "ruid is %d instead of %d", ruid, tester_uid);
  fail_unless(tester_uid == euid, "euid is %d instead of %d", euid, tester_uid);
  fail_unless(         0 == suid, "suid is %d instead of %d", suid, 0);
  fail_unless(tester_gid == rgid, "rgid is %d instead of %d", rgid, tester_gid);
  fail_unless(tester_gid == egid, "egid is %d instead of %d", egid, tester_gid);
  fail_unless(         0 == sgid, "sgid is %d instead of %d", sgid, 0); 
}
END_TEST

START_TEST(test_push_from_nonroot)
{
  uid_t ruid, euid, suid;
  gid_t rgid, egid, sgid;

  fail_unless(0 == setresgid(tester_gid, tester_gid, 0),
	      "setresgid: %s", g_strerror(errno));
  fail_unless(0 == setresuid(tester_uid, tester_uid, 0),
	      "setresuid: %s", g_strerror(errno));
  push_cred(0, 0);

  fail_unless(0 == getresuid(&ruid, &euid, &suid),
	      "getresuid: %s", g_strerror(errno));
  fail_unless(0 == getresgid(&rgid, &egid, &sgid),
	      "getresgid: %s", g_strerror(errno));
  fail_unless(0 == ruid, "ruid is %d instead of %d", ruid, 0);
  fail_unless(0 == euid, "euid is %d instead of %d", euid, 0);
  fail_unless(0 == suid, "suid is %d instead of %d", suid, 0);
  fail_unless(0 == rgid, "rgid is %d instead of %d", rgid, 0);
  fail_unless(0 == egid, "egid is %d instead of %d", egid, 0);
  fail_unless(0 == sgid, "sgid is %d instead of %d", sgid, 0);
}
END_TEST

START_TEST(test_push_nonroot_then_pop_root)
{
  uid_t ruid, euid, suid;
  gid_t rgid, egid, sgid;

  push_cred(tester_uid, tester_gid);
  pop_cred();

  fail_unless(0 == getresuid(&ruid, &euid, &suid),
	      "getresuid: %s", g_strerror(errno));
  fail_unless(0 == getresgid(&rgid, &egid, &sgid),
	      "getresgid: %s", g_strerror(errno));
  fail_unless(0 == ruid, "ruid is %d instead of %d", ruid, 0);
  fail_unless(0 == euid, "euid is %d instead of %d", euid, 0);
  fail_unless(0 == suid, "suid is %d instead of %d", suid, 0);
  fail_unless(0 == rgid, "rgid is %d instead of %d", rgid, 0);
  fail_unless(0 == egid, "egid is %d instead of %d", egid, 0);
  fail_unless(0 == sgid, "sgid is %d instead of %d", sgid, 0);  
}
END_TEST

Suite *
helper_suite(void)
{
  Suite *s;
  TCase *tc;

  s= suite_create("Helper");

  tc = tcase_create ("is_root");
  tcase_add_test (tc, test_is_root_really_root);
  tcase_add_test (tc, test_is_root_notreally_root);
  suite_add_tcase (s, tc);

  tc = tcase_create("credential queue");
  tcase_add_test(tc, test_push_from_root);
  tcase_add_test(tc, test_push_from_nonroot);
  tcase_add_test(tc, test_push_nonroot_then_pop_root);
  suite_add_tcase (s, tc);
    
  return s;
}
