/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <test/check_plugin.h>
#include <test/config_file_helper.h>
#include <daemon/config_store.h>
#include <daemon/process.h>
#include <daemon/sandboxer-helper.h>
#include <glib.h>
#include <errno.h>

gchar *tempdir;
extern uid_t sandboxer_uid;
extern gid_t sandboxer_gid;
extern uid_t tester_uid;
extern gid_t tester_gid;



typedef struct _expected_group expected_group;
struct _expected_group {
  gchar *group_name; /* group name in key file */
  uid_t ruid;        /* value of ruid key */
  uid_t euid;        /* value of euid key */
  uid_t suid;        /* value of suid key */
};



static expected_group* 
expected_group_new(const gchar* group_name, 
		   const uid_t ruid, 
		   const uid_t euid, 
		   const uid_t suid);

static void
expected_group_free(expected_group *self);

static expected_group*
expected_group_new_root(const gchar *group_name);

static expected_group*
expected_group_new_sandboxer(const gchar *group_name);

static expected_group* 
expected_group_new_tester(const gchar *group_name);

static GQueue* 
setup_expectation_queue(void);







static expected_group* 
expected_group_new(const gchar* group_name, 
		   const uid_t ruid, 
		   const uid_t euid, 
		   const uid_t suid)
{
  expected_group *p;

  p= g_new(expected_group, 1); g_assert(p);
  g_assert(group_name);
  p->group_name= g_strdup(group_name);
  p->ruid= ruid;
  p->euid= euid;
  p->suid= suid;
  
  return p;
}

static void
expected_group_free(expected_group *self)
{
  if (self) {
    g_free(self->group_name);
    g_free(self);
  } 
}

static expected_group*
expected_group_new_root(const gchar *group_name)
{
  return expected_group_new(group_name, 0, 0, 0);
}

static expected_group*
expected_group_new_sandboxer(const gchar *group_name)
{
  return expected_group_new(group_name, sandboxer_uid, sandboxer_uid, 0);
}

static expected_group* 
expected_group_new_tester(const gchar *group_name)
{
  return expected_group_new(group_name, tester_uid, tester_uid, tester_uid);
}



static GQueue* 
setup_expectation_queue(void)
{
  GQueue *q;

  q= g_queue_new(); g_assert(q);
  g_queue_push_head(q, expected_group_new_sandboxer("PROCESS_INSTANCE_INIT"));
  g_queue_push_head(q, expected_group_new_sandboxer("PROCESS_PRE_FORK"));
  g_queue_push_head(q,      expected_group_new_root("PROCESS_PRE_UNSHARE"));
  g_queue_push_head(q,      expected_group_new_root("PROCESS_PRE_PIVOT_ROOT"));
  g_queue_push_head(q,      expected_group_new_root("PROCESS_PRE_UMOUNT"));
  g_queue_push_head(q,    expected_group_new_tester("PROCESS_PRE_EXEC"));
  g_queue_push_head(q, expected_group_new_sandboxer("PROCESS_INSTANCE_FINALIZE"));
  
  return q;
}



static void
check_expectation(expected_group *expectation)
{
  gchar *filename;
  GKeyFile *keyfile;
  gboolean ret;
  uid_t ruid, euid, suid;  
  GError *error= NULL;
  
  keyfile= g_key_file_new(); g_assert(keyfile);
  filename= g_strdup_printf("%s/%s", tempdir, expectation->group_name);
  ret= g_key_file_load_from_file(keyfile,
				 filename,
				 G_KEY_FILE_NONE,
				 &error);
  fail_unless(ret, "g_key_file_load_from_file(%s): %s", filename, error ? error->message : "(null)");
  
  ruid= g_key_file_get_integer(keyfile, expectation->group_name,
			       "ruid", &error);
  fail_unless(NULL == error, "g_key_file_get_integer: %s", error ? error->message : "(null)");
  fail_unless(ruid == expectation->ruid, "%d is not %d (%s)", ruid, expectation->ruid, 
	      expectation->group_name);
  
  euid= g_key_file_get_integer(keyfile, expectation->group_name,
			       "euid", &error);
  fail_unless(NULL == error, "g_key_file_get_integer: %s", error ? error->message : "(null)");
  
  fail_unless(euid == expectation->euid, "%d is not %d (%s)", euid, expectation->euid,
	      expectation->group_name);
  suid= g_key_file_get_integer(keyfile, expectation->group_name,
			       "suid", &error);
  fail_unless(NULL == error, "g_key_file_get_integer: %s", error ? error->message : "(null)");
  fail_unless(suid == expectation->suid, "%d is not %d (%s)", suid, expectation->suid,
	      expectation->group_name);
}

static void
nuke_key_files(void)
{
  const gchar *files[]= {"PROCESS_INSTANCE_INIT",
			 "PROCESS_PRE_FORK",
			 "PROCESS_PRE_UNSHARE",
			 "PROCESS_PRE_PIVOT_ROOT",
			 "PROCESS_PRE_UMOUNT",
			 "PROCESS_INSTANCE_FINALIZE",
			 NULL };
  guint i;

  for (i= 0; files[i]; i++) {
    gchar *filename;
    int ret;

    filename= g_strdup_printf("%s/%s", g_get_tmp_dir(), files[i]); g_assert(filename);
    ret= g_unlink(filename);
    if (ret && (errno != ENOENT)) {
      g_error("could not remove \"%s\": %s", filename, g_strerror(errno));
    }
    g_free(filename);
  }  
}

START_TEST(stage_test)
{
  SandboxGlobalConfig *global_config;
  SandboxConfigStore *config_store;
  SandboxerProcess *proc;
  GQueue *expectations;
  gchar *secret_cookie;
  gchar *global_config_path;
  gchar *package_file_dir;
  gchar *bintrue= "/bin/true";
  gchar *bintrue_argv[]= {bintrue, NULL};
  GError *error= NULL;

  expectations= setup_expectation_queue(); g_assert(expectations);
  write_test_package_files();
  make_sandboxerd_conf("libtest.so");
  tempdir= (gchar*)g_get_tmp_dir();
  
  global_config= sandbox_global_config_new(get_sandboxerd_conf());
  fail_if(NULL == global_config, "sandbox_global_config_new returned NULL");

  config_store= sandbox_config_store_new(get_packagedir());
  fail_if(NULL == config_store, "sandbox_config_store_new returned NULL");
  plugin_system_init();
  load_plugins();
  push_cred(sandboxer_uid, sandboxer_gid);
  proc= sandboxer_process_new(bintrue, &secret_cookie, NULL);
  mark_point();
  fail_if(NULL == proc, "sandboxer_process_new returned NULL"); 
  fail_if(NULL == secret_cookie, "secret_cookie is NULL");
  fail_unless(sandboxer_process_run(proc, secret_cookie, bintrue_argv, &error),
	      "sandboxer_process_run returned FALSE: %s", error ? error->message : "(null)");
  g_object_unref(G_OBJECT(proc));
  mark_point();
  {
    expected_group *expectation;
    
    while(expectation= g_queue_pop_head(expectations)) {
      check_expectation(expectation);
    }
  }
}
END_TEST

Suite *
plugin_suite(void)
{
  Suite *s;
  TCase *tc;

  s= suite_create("plugin");  

  tc = tcase_create ("test_plugin");
  tcase_add_checked_fixture(tc, tmpdir_setup, tmpdir_teardown);
  tcase_add_checked_fixture(tc, nuke_key_files, NULL);
  tcase_add_test(tc, stage_test);
  suite_add_tcase(s, tc);

  return s;
}
