/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <daemon/process.h>
#include <test/check_process.h>
#include <test/config_file_helper.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <glib-object.h>
#include <signal.h>
#include <string.h>
#include <pthread.h>

/*            types            */

struct process_signal_context
{
  gboolean caught;
  gint exit_status;
  gchar *dbus_path;
  GMainLoop *loop;
};

extern uid_t sandboxer_uid;
extern gid_t sandboxer_gid;

/*         prototypes         */
static void
process_run_signal_callback(gpointer instance, const gchar *dbus_path, 
			    struct process_signal_context *context);

static void
process_terminated_signal_callback(gpointer instance, const gint exit_status,
				   struct process_signal_context *context);

static void
process_run_argv_signal_callback(gpointer instance, const gint exit_status, 
				 struct process_signal_context *context);



/*       implemenations       */
static void
process_run_signal_callback(gpointer instance, const gchar *dbus_path, 
			    struct process_signal_context *context)
{
  fail_unless(g_str_equal(context->dbus_path, dbus_path), 
	      "%s is not %s", dbus_path, context->dbus_path);
  context->caught= TRUE; 
}

static void
process_terminated_signal_callback(gpointer instance, const gint exit_status,
				   struct process_signal_context *context)
{
  g_assert(context); g_assert(context->loop);
  fail_unless(context->exit_status == exit_status, 
	      "%d is not %d", exit_status, context->exit_status);
  context->caught= TRUE; 
  g_main_loop_quit(context->loop);
}

static void
process_run_argv_signal_callback(gpointer instance, const gint exit_status, 
				 struct process_signal_context *context)
{
  g_assert(context); g_assert(context->loop);
  fail_unless(WEXITSTATUS(exit_status) == context->exit_status, 
	      "%d is not %d", WEXITSTATUS(exit_status), context->exit_status);
  context->caught= TRUE; 
  g_main_loop_quit(context->loop);
}

START_TEST(process_gobject_new_no_params)
{
  gpointer gobj;

  push_cred(sandboxer_uid, sandboxer_gid);
  gobj= g_object_new(SANDBOXER_TYPE_PROCESS, NULL);
  fail_if(NULL == gobj, "g_object_new returned NULL pointer");
}
END_TEST

START_TEST(process_gobject_new_path_param)
{
  gpointer gobj;
  
  push_cred(sandboxer_uid, sandboxer_gid);
  gobj= g_object_new(SANDBOXER_TYPE_PROCESS, "process-path", "/bin/true", NULL);
  fail_if(NULL == gobj, "g_object_new returned NULL pointer");
  fail_unless(SANDBOXER_IS_PROCESS(gobj), "g_object is not a process");
}
END_TEST

START_TEST(process_new)
{
  SandboxerProcess *proc;
  gchar *secret_cookie= NULL;

  push_cred(sandboxer_uid, sandboxer_gid);
  proc= sandboxer_process_new("/bin/true", &secret_cookie, NULL);
  fail_if(NULL == proc, "got back null pointer!");
  fail_unless(SANDBOXER_IS_PROCESS(proc), "didn't get back a process!");  
  fail_if(NULL == secret_cookie, "secret cookie is unset!");
}
END_TEST


START_TEST(process_can_read_path)
{
  const gchar right_path[]= "/bin/true";
  SandboxerProcess *proc;
  gchar *secret_cookie= NULL;
  gchar *path;

  push_cred(sandboxer_uid, sandboxer_gid);
  proc= sandboxer_process_new(right_path, &secret_cookie, NULL);
  g_object_get(G_OBJECT(proc), "process-path", &path, NULL);
  fail_unless(g_str_equal(right_path, path), "set and get did not match");  
}
END_TEST

START_TEST(process_cant_set_path_after_construct)
{
  const gchar right_path[]= "/bin/true";
  const gchar wrong_path[]= "/bin/false";
  gchar *path;
  SandboxerProcess *proc;
  gchar *secret_cookie= NULL;
  GValue val= {0, };

  push_cred(sandboxer_uid, sandboxer_gid);
  proc= sandboxer_process_new(right_path, &secret_cookie, NULL);
  g_value_init (&val, G_TYPE_STRING);
  g_value_set_string (&val, wrong_path); //should have no effect
  g_object_get(G_OBJECT(proc), "process-path", &path, NULL);
  fail_if(g_str_equal(wrong_path, path), "could set path, when shouldn't be able to");  
}
END_TEST

START_TEST(process_get_dbus_path)
{
  const gchar path[]= "/bin/true";
  gchar *dbus_path;
  gchar **path_parts;
  gboolean numfail= FALSE;
  SandboxerProcess *proc;
  gchar *secret_cookie= NULL;
  gint i;

  push_cred(sandboxer_uid, sandboxer_gid);
  proc= sandboxer_process_new(path, &secret_cookie, NULL);
  g_object_get(G_OBJECT(proc), "process-dbus_path", &dbus_path, NULL);
  fail_if(NULL == dbus_path, "dbus_path was NULL");
  path_parts= g_strsplit(dbus_path, "_", 2);
  fail_unless(g_str_equal(SANDBOXER_DBUS_PROCESS_PATH, path_parts[0]),
	      "\"%s\" is not \"%s\"", path_parts[0], SANDBOXER_DBUS_PROCESS_PATH);
  for (i= 0; i < strlen(path_parts[1]); i++) {
    if (!g_ascii_isdigit(path_parts[1][i])) {
      numfail= TRUE;
    }
  }
  fail_if(numfail, "\"%s\" is not all numbers\n", path_parts[1]);
  
}
END_TEST

START_TEST(process_run_return)
{
  SandboxerProcess *proc;
  gchar *secret_cookie= NULL;
  GError *error= NULL;
  gchar *bintrue= "/bin/true";
  gchar *argv[]= {bintrue, NULL};

  push_cred(sandboxer_uid, sandboxer_gid);
  proc= sandboxer_process_new(bintrue, &secret_cookie, NULL);
  fail_unless(sandboxer_process_run(proc, secret_cookie, argv, &error), 
	      "run returned false: %s", error ? error->message : "NULL");  
  fail_if(error, "error is set but ruturned true");
}
END_TEST



START_TEST(process_run_signal)
{
  SandboxerProcess *proc;
  gchar *secret_cookie= NULL;
  GError *error= NULL;
  gchar *bintrue= "/bin/true";
  gchar *argv[]= {bintrue, NULL};
  struct process_signal_context context = {
    .caught= FALSE,
  };
  
  push_cred(sandboxer_uid, sandboxer_gid);
  proc= sandboxer_process_new(bintrue, &secret_cookie, NULL);
  g_object_get(G_OBJECT(proc), "process-dbus_path", &context.dbus_path, NULL);
  fail_if(NULL == context.dbus_path, "context.dbus_path is null");

  /* register for signal */
  g_signal_connect( G_OBJECT(proc), 
		    "run", //FIXME: use macro (DRY)
		    G_CALLBACK(process_run_signal_callback),
		    &context
		    );
  sandboxer_process_run(proc, secret_cookie, argv, &error);
  
  fail_unless(context.caught, "signal was not caught");    
}
END_TEST

START_TEST(process_run_wrong_secret)
{
  SandboxerProcess *proc;
  gchar *secret_cookie= NULL;
  const gchar wrong_cookie[]= "this is the wrong cookie";
  gchar bintrue[]= "/bin/true";
  gchar *argv[]= {bintrue, NULL};
  GError *error= NULL;
  gboolean ret;

  
  push_cred(sandboxer_uid, sandboxer_gid);
  proc= sandboxer_process_new(bintrue, &secret_cookie, NULL);
  ret= sandboxer_process_run(proc, wrong_cookie, argv, &error);
  fail_if(ret, "run returned true even though we used wrong cookie");
  fail_if(NULL == error, "error was NULL");
  fail_unless(SANDBOXER_PROCESS_ERROR == error->domain, "wrong error quark");
  fail_unless(SANDBOXER_PROCESS_ERROR_COOKIE == error->code, "wrong error code");
}
END_TEST


START_TEST(process_terminated_signal)
{
  SandboxerProcess *proc;
  gchar *secret_cookie= NULL;
  GError *error= NULL;
  gchar *binfalse= "/bin/false";
  gchar *argv[]= {binfalse, NULL};
  struct process_signal_context context = {
    .caught= FALSE,
    .exit_status= 256
  };

  push_cred(sandboxer_uid, sandboxer_gid);
  proc= sandboxer_process_new(binfalse, &secret_cookie, NULL);
  g_object_get(G_OBJECT(proc), "process-dbus_path", &context.dbus_path, NULL);
  fail_if(NULL == context.dbus_path, "context.dbus_path is null");
  /* register for signal */
  g_signal_connect( proc, 
		    SIGNAL_TERMINATED, 
		    G_CALLBACK(process_terminated_signal_callback),
		    &context  
		    );
  sandboxer_process_run(proc, secret_cookie, argv, &error);
  context.loop = g_main_loop_new (NULL, FALSE);
  g_main_loop_run (context.loop); //give chance for signal to be caught
  fail_unless(context.caught, "signal was not caught"); 
}
END_TEST

//this uses a mainloop b/c g_child_watch_add uses GMainLoop 
//to process
START_TEST(process_run_argv_return)
{
  SandboxerProcess *proc;
  gchar *secret_cookie= NULL;
  GError *error= NULL;
  gint exit_status;
  const gboolean block= TRUE;
  gboolean unchanged;
  gchar *prog;
  gchar *argv[3];
  struct process_signal_context context = {
    .caught= FALSE,
    .exit_status= 23,
  };

  push_cred(sandboxer_uid, sandboxer_gid);
  prog= g_strdup_printf("%s/return_value", get_tmpdir());
  g_assert(prog);

#if 0  //FIXME: debug
  {
    int ret;
    struct stat buf;

    g_debug("hello!!!");
    push_cred(0,0);
    fail_unless(0 == g_lstat(prog, &buf), "g_lstat returned non zero!");
    pop_cred();
    g_debug("pid writer owner is");
    g_debug("mode\town\tgrp");
    g_debug("%o  \t%d \t%d", buf.st_mode, buf.st_uid, buf.st_gid);
  }
#endif


  argv[0]= prog;
  argv[1]= "23";
  argv[2]= NULL;

  proc= sandboxer_process_new(prog, &secret_cookie, NULL);
  fail_if(NULL == proc, "proc is NULL");
  g_object_get(G_OBJECT(proc), "process-dbus_path", &context.dbus_path, NULL);
  fail_if(NULL == context.dbus_path, "context.dbus_path is null");
  /* register for signal */
  g_signal_connect( G_OBJECT(proc), 
		    SIGNAL_TERMINATED, 
		    G_CALLBACK(process_run_argv_signal_callback),
		    &context  
		    );
  {
    //FIXME: macro-ize this and turn it into a helper or something
    gboolean ret;
    
    ret= sandboxer_process_run(proc, secret_cookie, argv, &error);
    if (FALSE == ret) {
      if (NULL == error) {
	fail("run returned FALSE and error is NULL");
      }
      fail("run returned FALSE: %s", error->message);
    }
  }

  fail_if(error, "error is set but run returned TRUE");
  context.loop = g_main_loop_new (NULL, FALSE);
  mark_point();
  g_main_loop_run(context.loop); //give chance for signal to be caught
  mark_point();
  
  fail_unless(context.caught, "signal was not caught"); 
}
END_TEST

START_TEST(process_kill_valid)
{
  SandboxerProcess *proc;
  SandboxGlobalConfig *global_config;
  gint exit_status;
  gchar *prog; 
  gchar *argv[2];
  gboolean ret;
  GError *error= NULL;
  gchar *secret_cookie= NULL;
  struct process_signal_context context = {
    .caught= FALSE,
    .exit_status= 15
  };
  
  push_cred(sandboxer_uid, sandboxer_gid);
  global_config= sandbox_global_config_new(NULL);
  fail_if(NULL == global_config, "global config is NULL");

  prog= g_strdup_printf("%s/sleep_forever", get_tmpdir());
  argv[0]= prog; argv[1]= NULL;
  mark_point();
  proc= sandboxer_process_new(prog, &secret_cookie, NULL);
  ret= sandboxer_process_run(proc, secret_cookie, argv, &error);
  if (FALSE == ret) {
    if (NULL == error) fail("run returned FALSE and error is NULL");
    fail("run returned FALSE: %s", error->message);
  }
  mark_point();
  g_signal_connect(proc,
		   SIGNAL_TERMINATED,
		   G_CALLBACK(process_terminated_signal_callback),
		   &context);
  fail_unless(sandboxer_process_kill(proc, secret_cookie, SIGTERM, &error), "couldn't kill!");
  
  context.loop = g_main_loop_new (NULL, FALSE);
  g_main_loop_run (context.loop); //give chance for signal to be caught
  fail_unless(context.caught, "signal was not caught"); 
}
END_TEST

START_TEST(process_kill_wrong_secret)
{
  const gchar wrong_cookie[]= "this is the wrong cookie";
  SandboxerProcess *proc;
  SandboxGlobalConfig *global_config;
  gboolean ret;
  gchar *prog; 
  gchar *argv[2];
  gchar *secret_cookie= NULL;
  GError *error= NULL;
  struct process_signal_context context = {
    .caught= FALSE,
    .exit_status= 15
  };

  push_cred(sandboxer_uid, sandboxer_gid);
  global_config= sandbox_global_config_new(NULL);
  fail_if(NULL == global_config, "global config is NULL");

  prog= g_strdup_printf("%s/sleep_forever", get_tmpdir());
  argv[0]= prog; argv[1]= NULL;
  proc= sandboxer_process_new(prog, &secret_cookie, NULL);
  ret= sandboxer_process_run(proc, secret_cookie, argv, &error);
  if (FALSE == ret) {
    if (NULL == error) fail("run returned FALSE and error is NULL");
    fail("run returned FALSE: %s", error->message);
  }
  ret= sandboxer_process_kill(proc, wrong_cookie, SIGTERM, &error);
  fail_if(ret, "kill returned true even though we used wrong cookie");
  fail_if(NULL == error, "error was NULL");
  fail_unless(SANDBOXER_PROCESS_ERROR == error->domain, "wrong error quark");
  fail_unless(SANDBOXER_PROCESS_ERROR_COOKIE == error->code, "wrong error code");
  //kill with right cookie
  fail_unless(sandboxer_process_kill(proc, secret_cookie, SIGTERM, &error), "couldn't kill!");

  g_signal_connect(proc,
		   SIGNAL_TERMINATED,
		   G_CALLBACK(process_terminated_signal_callback),
		   &context);
  context.loop = g_main_loop_new (NULL, FALSE);

  g_main_loop_run (context.loop); //give chance for signal to be caught
  fail_unless(context.caught, "signal was not caught"); 
}
END_TEST


Suite *
process_suite(void)
{
  Suite *s;
  TCase *tc;

  s= suite_create("Process");

  tc = tcase_create ("Process New");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, tmpdir_teardown);  
  tcase_add_test (tc, process_gobject_new_no_params);
  tcase_add_test (tc, process_gobject_new_path_param);  
  tcase_add_test (tc, process_new);
  suite_add_tcase (s, tc);

  tc= tcase_create ("Process Properties");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, tmpdir_teardown);  
  tcase_add_test (tc, process_cant_set_path_after_construct);
  tcase_add_test (tc, process_can_read_path);
  suite_add_tcase (s, tc);

  tc= tcase_create ("Process Run");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, tmpdir_teardown);  
  tcase_add_test (tc, process_run_return);
  tcase_add_test (tc, process_run_signal);
  tcase_add_test (tc, process_terminated_signal);
  tcase_add_test (tc, process_run_argv_return);
  tcase_add_test (tc, process_run_wrong_secret);
  suite_add_tcase (s, tc);

  tc= tcase_create ("Process kill");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, tmpdir_teardown);  
  tcase_add_test (tc, process_kill_valid);
  tcase_add_test (tc, process_kill_wrong_secret);
  suite_add_tcase (s, tc);
  
  return s;
}

