/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <check_sandbox.h>
#include <daemon/sandbox.h>
#include <config_file_helper.h>
#include <client/sandboxer-process-glue.h>
#include <daemon/sandboxer-sandbox-client-glue.h>
#include <daemon/sandboxer-service.h>
#include <daemon/process.h>
#include <glib.h>

extern uid_t sandboxer_uid;
extern gid_t sandboxer_gid;

START_TEST(test_new)
{
  SandboxerSandbox *self;
  
  push_cred(sandboxer_uid, sandboxer_gid);
  self= sandboxer_sandbox_new("coreutils_sandbox");
  fail_if(NULL == self, "sandboxer_sandbox_new returned NULL");
  fail_unless(SANDBOXER_IS_SANDBOX(self), "self is not a SandboxerSandbox");	
}
END_TEST

START_TEST(test_run_true)
{
  SandboxerSandbox *self;
  gchar *process_dbus_path= NULL;
  gchar *secret_cookie= NULL;
  gchar *unique_name= NULL;
  GError *error= NULL;

  push_cred(sandboxer_uid, sandboxer_gid);
  self= sandboxer_sandbox_new("coreutils_sandbox");
  fail_unless(sandboxer_sandbox_get_process(self, "/bin/true",
					    &process_dbus_path,
					    &secret_cookie,
					    &unique_name,
					    &error),
	      "get_process failed: %s", error ? error->message : "(NULL)");
  fail_if(NULL == process_dbus_path, "process_dbus_path is NULL");
  fail_if(NULL == secret_cookie, "secret_cookie is NULL");
  fail_if(NULL == unique_name, "unique name is NULL");
}
END_TEST

START_TEST(test_sandbox_service_get_date)
{
  SandboxerSandboxService *sandserv;
  gboolean ret;
  gchar  *process= NULL;
  gchar *secret_cookie= NULL;
  gchar *unique_name= NULL;
  GError *err= NULL;

  push_cred(sandboxer_uid, sandboxer_gid);
  sandserv= sandboxer_sandbox_service_new();
  fail_if(NULL == sandserv, "sandserv is NULL");
  ret= sandboxer_service_get_process(sandserv, "/bin/date", 
				     &process, &secret_cookie, &unique_name, 
				     &err);
  fail_unless(ret, "returned error: %s", err ? err->message : "(NULL)");
  fail_unless(NULL == err, "err is not NULL");
  fail_unless(g_str_has_prefix(process, SANDBOXER_DBUS_PROCESS_PATH), 
	      "%s does not start with %s", process, SANDBOXER_DBUS_PROCESS_PATH);
  fail_if(NULL == secret_cookie, "got empty secret cookie");
}
END_TEST

static void
unlink_if_there(const gchar *path)
{
  g_assert(path);

  push_cred(0, 0);
  if (g_file_test(path, G_FILE_TEST_EXISTS)) {
    fail_unless(0 == g_unlink(path, "couldn't unlink %s", path));
  }
  pop_cred();
}

//FIXME: this one and the next
//       1) share too much code
START_TEST(test_sandbox_do_it_by_hand)
{
  gchar *touch_argv[]= {"/bin/touch", NULL, NULL };
  gchar *touch_file;
  SandboxerSandboxService *sandserv;
  DBusGConnection *connection;
  gboolean ret;
  DBusGProxy *process_proxy;
  uid_t saved;
  gchar  *process= NULL;
  gchar *secret_cookie= NULL;
  gchar *unique_name= NULL;
  GError *error= NULL;

  touch_file= g_strdup_printf("%s/right_now-%d", g_get_tmp_dir(), getppid());
  touch_argv[1]= touch_file;
  unlink_if_there(touch_file);
 
  push_cred(sandboxer_uid, sandboxer_gid);
  sandserv= sandboxer_sandbox_service_new();
  sandboxer_service_get_process(sandserv, touch_argv[0], 
				&process, &secret_cookie, &unique_name,
				NULL);
  connection= dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
  fail_if(NULL == connection, "coulnd't get a connection: %s", 
	  error ? error->message : "(NULL)");
  process_proxy = dbus_g_proxy_new_for_name (connection,
					     unique_name,
					     process,
					     "com.intel.sandboxer.Process");
  fail_if(NULL == process, "couldn't get proxy for \"%s\"", process);
  fail_unless(com_intel_sandboxer_Process_run (process_proxy, 
					       secret_cookie, 
					       (const gchar**)touch_argv,
					       &error),
	      "Could not run Process.Run method: %s", error ? error->message : "(NULL)");
  sleep(5); //wait out race
  fail_unless(g_file_test(touch_file, G_FILE_TEST_EXISTS), "%s doesn't exist", touch_file);
  pop_cred();
}
END_TEST

START_TEST(test_sandbox_do_it_by_hand_multiple)
{
  gchar *touch_argv[]= {"/bin/touch", NULL, NULL };
  const guint TIMES= 10;
  gchar *touch_file;
  SandboxerSandboxService *sandserv;
  DBusGConnection *connection;
  gboolean ret;
  uid_t saved;
  guint i;
  DBusGProxy *process_proxy;
  gchar  *process= NULL;
  gchar *secret_cookie= NULL;
  gchar *unique_name= NULL;
  GError *error= NULL;

  touch_file= g_strdup_printf("%s/right_now-%d", g_get_tmp_dir(), getppid());
  touch_argv[1]= touch_file;
  unlink_if_there(touch_file);

  push_cred(sandboxer_uid, sandboxer_gid);
  sandserv= sandboxer_sandbox_service_new();
  for (i= 0; i < TIMES; i++) {
    g_debug("iteration: %d", i);
    sandboxer_service_get_process(sandserv, touch_argv[0], 
				  &process, &secret_cookie, &unique_name,
				  NULL);
    connection= dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
    fail_if(NULL == connection, "coulnd't get a connection: %s", 
	    error ? error->message : "(NULL)");
    process_proxy = dbus_g_proxy_new_for_name (connection,
					       unique_name,
					       process,
					       "com.intel.sandboxer.Process");
  fail_if(NULL == process, "couldn't get proxy for \"%s\"", process);
  fail_unless(com_intel_sandboxer_Process_run (process_proxy, 
					       secret_cookie, 
					       (const gchar**)touch_argv,
					       &error),
	      "Could not run Process.Run method: %s", error ? error->message : "(NULL)");

  g_usleep(G_USEC_PER_SEC / 100);
  fail_unless(g_file_test(touch_file, G_FILE_TEST_EXISTS), "%s doesn't exist", touch_file);
  }
  pop_cred();
}
END_TEST

/*
 * we use unchecked fixture because Sandbox actually does unshare in this process, rather
 * than across a fork.  So, we teardown the tmpdir in the parent to avoid EBUSY
 */
Suite *
sandbox_suite(void)
{
  Suite *s;
  TCase *tc;

  s= suite_create("Sandbox");

  tc = tcase_create ("Sandbox New");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, NULL);
  tcase_add_unchecked_fixture(tc, NULL, tmpdir_teardown);
  tcase_add_test (tc, test_new);
  suite_add_tcase (s, tc);

  tc = tcase_create ("Sandbox Run");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, NULL);
  tcase_add_unchecked_fixture(tc, NULL, tmpdir_teardown);
  tcase_add_test (tc, test_run_true);
  suite_add_tcase (s, tc);

  tc = tcase_create ("Sandboxer Service Get (through Sandbox)");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, NULL);
  tcase_add_unchecked_fixture(tc, NULL, tmpdir_teardown);
  tcase_set_timeout(tc, 105);
  tcase_add_test (tc, test_sandbox_service_get_date);
  suite_add_tcase (s, tc);

  tc = tcase_create ("Process Run (through Sandbox)");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, NULL);
  tcase_add_unchecked_fixture(tc, NULL, tmpdir_teardown);
  tcase_set_timeout(tc, 105);
  tcase_add_test (tc, test_sandbox_do_it_by_hand);
  suite_add_tcase (s, tc);

  tc = tcase_create ("Process Run (through Sandbox) Multiple");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, NULL);
  tcase_add_unchecked_fixture(tc, NULL, tmpdir_teardown);
  tcase_add_test (tc, test_sandbox_do_it_by_hand_multiple);
  suite_add_tcase (s, tc);

  return s;
}
