/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <test/check_sandboxer_service.h>
#include <test/config_file_helper.h>
#include <daemon/process.h>
#include <daemon/sandboxer-service.h>

#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <dbus/dbus-glib.h>

extern uid_t sandboxer_uid;
extern gid_t sandboxer_gid;

START_TEST (test_new)
{
  gpointer gobj;
  
  push_cred(sandboxer_uid, sandboxer_gid);
  gobj= g_object_new(SANDBOXER_TYPE_SANDBOX_SERVICE, NULL);
  fail_if (NULL == gobj, "g_object_new returned NULL pointer");
  fail_unless (SANDBOXER_IS_SANDBOX_SERVICE(gobj), "g_object_new returned something that is not a sandbox service");
  pop_cred();
}
END_TEST


START_TEST(test_get_process)
{
  gpointer gobj;
  gboolean ret;
  gchar  *process;
  GError *err= NULL;
  gchar *secret_cookie= NULL;
  gchar *unique_name= NULL;

  push_cred(sandboxer_uid, sandboxer_gid);
  gobj= g_object_new(SANDBOXER_TYPE_SANDBOX_SERVICE, NULL);
  ret= sandboxer_service_get_process(SANDBOXER_SANDBOX_SERVICE(gobj), "/bin/true", 
				     &process, &secret_cookie, &unique_name, &err);
  fail_unless(ret, "returned error!");
  fail_unless(NULL == err, "err is not NULL");
  fail_unless(g_str_has_prefix(process, SANDBOXER_DBUS_PROCESS_PATH), 
	      "%s does not start with %s", process, SANDBOXER_DBUS_PROCESS_PATH);
  fail_if(NULL == secret_cookie, "got empty secret cookie");
  pop_cred();
}
END_TEST

START_TEST(test_find_process_by_pid_invalid_pid)
{
  SandboxerSandboxService* sandserv;
  GError *error= NULL;
  gchar *bintrue= "/bin/true";
  gchar *process_path= NULL;
  gchar *secret_cookie= NULL;
  gchar *unique_name= NULL;
  gboolean ret;

  push_cred(sandboxer_uid, sandboxer_gid);
  sandserv= sandboxer_sandbox_service_new();
  fail_if(NULL == sandserv, "failed to get sandbox service");
  ret= sandboxer_service_find_process_by_pid(sandserv, 12345, 
					     &process_path, &secret_cookie, &unique_name,
					     &error);
  fail_if(ret, "finding in invalid pid did not return error");
  fail_if(NULL == error, "returned FALSE but error was still NULL");
  pop_cred();
}
END_TEST

START_TEST(test_list_all_processes_empty)
{
  SandboxerSandboxService* sandserv;
  GError *error= NULL;
  gboolean ret;
  char **paths= NULL;
  gchar **secret_cookies= NULL;
  gchar **unique_names= NULL;

  push_cred(sandboxer_uid, sandboxer_gid);
  sandserv= sandboxer_sandbox_service_new();
  fail_if(NULL == sandserv, "failed to get sandbox service");
  ret= sandboxer_service_list_all_processes(sandserv, 
					    &paths, &secret_cookies, &unique_names, 
					    &error);
  fail_unless(ret, "list_all_processes returned false");
  fail_if(NULL == paths, "paths is NULL");
  fail_if(NULL == secret_cookies, "secret cookies is NULL");
  fail_if(NULL == unique_names, "unique names is NULL");
  fail_unless(NULL == *paths, "path_list is not empty");
  fail_unless(NULL == *secret_cookies, "secret_cookies list is not empty");
  fail_unless(NULL == *unique_names, "unique_names list is not empty");
  pop_cred();
}
END_TEST

START_TEST(test_list_all_processes_notempty)
{ 
  const guint size= 5;
  SandboxerSandboxService* sandserv;
  gchar *secret_cookie= NULL;
  gchar *unique_name;
  GError *error= NULL;
  gchar *paths[size];
  gchar **paths_returned= NULL;
  gchar **secret_cookies_returned= NULL;
  gchar **unique_names_returned= NULL;
  guint matches[size];
  gchar *bintrue= "/bin/true";
  gchar *argv[]= {bintrue, NULL};
  gboolean ret;
  guint i;

  push_cred(sandboxer_uid, sandboxer_gid);
  sandserv= sandboxer_sandbox_service_new();
  fail_if(NULL == sandserv, "failed to get sandbox service");
  memset(paths, 0, sizeof(gchar*)*size);
  for (i= 0; i < size; i++) {
    ret= sandboxer_service_get_process(sandserv, 
				       bintrue, &(paths[i]), 
				       &secret_cookie, &unique_name, 
				       &error);
    fail_unless(TRUE == ret, "returned error!");
    fail_unless(NULL == error, "error is not NULL");
    fail_if(NULL == paths[i], "paths[%d] is NULL", i);
  }

  /* test to make sure dbus_path appears once and only once in paths */
  memset(matches, 0, sizeof(guint)*size);
  for (i= 0 ; i < size ; i++){
    guint j;
    
    for (j= 0; j < size; j++) {
      if (g_str_equal(paths[i], paths[j])) {
	matches[i]++;
      }
    }
  }
  for (i= 0; i < size; i++) {
    fail_unless(1 == matches[i], "%s was listed in paths %d times", paths[i], matches[i]);
  }

  ret= sandboxer_service_list_all_processes(sandserv, 
					    &paths_returned, &secret_cookies_returned,
					    &unique_names_returned,
					    &error);
  fail_unless(TRUE == ret, "list_all_processes returned error");  
  fail_unless(NULL == error, "error is not NULL");
  fail_if(NULL == paths_returned, "paths_returned is NULL");
 
  /* test to see that there is one and only one match to the original paths */
  memset(matches, 0, sizeof(guint)*size);
  for (i= 0 ; i < size ; i++){
    guint j;
    
    for (j= 0; j < size; j++) {
      if (g_str_equal(paths[i], paths_returned[j])) {
	matches[i]++;
      }
    }
  }
  for (i= 0; i < size; i++) {
    fail_unless(1 == matches[i], "%s was returned %d times", paths[i], matches[i]);
  }  
  pop_cred();
}
END_TEST

//FIXME: abuse of secret_cookie -- also really hard to follow
START_TEST(test_find_process_by_pid)
{
  const gchar *fname= "/tmp/check_sandboxer_service.pid";
  pid_t pid;
  FILE* pidfile;
  SandboxerSandboxService* sandserv;
  gchar *pid_writer;
  gchar *argv[2];
  gboolean ret;
  GSList *proclist;
  SandboxerProcess *proc;
  gchar *secret_cookie= NULL;
  gchar *unique_name= NULL;
  GError *error= NULL;
  gchar *dbus_path= NULL;
  gchar *dbus_path_found= NULL;

  push_cred(sandboxer_uid, sandboxer_gid);
  pid_writer= g_strdup_printf("%s/pid_writer", get_tmpdir());
  g_assert(pid_writer);
  argv[0]= pid_writer;
  argv[1]= NULL;

  mark_point();

  sandserv= sandboxer_sandbox_service_new();
  fail_if(NULL == sandserv, "failed to get sandbox service");
  ret= sandboxer_service_get_process(sandserv, pid_writer, 
				     &dbus_path, &secret_cookie, &unique_name, 
				     &error);
  fail_unless(TRUE == ret, "returned error!");
  fail_unless(NULL == error, "error is not NULL");
  fail_if(NULL == dbus_path, "dbus_path is NULL");
  
  /* 
   * here, we cheat and use a unit-test helper to avoid
   * getting the process object through dbus 
   */
  proclist= sandboxer_service_give_process_list(sandserv);
  fail_if(NULL == proclist, "process list is NULL");
  proc= g_slist_nth_data(proclist, 0);
  mark_point();
  fail_if(NULL == proc, "process is NULL");

  push_cred(0,0);
  if (-1 == unlink(fname)) {
    if (ENOENT != errno) {
      fail("unlink(%s): (%d) %s", fname, errno, g_strerror(errno));
    }
  }
  pop_cred();

  ret= sandboxer_process_run(proc, secret_cookie, argv, &error);
  mark_point();
  fail_unless(NULL == error, "run returned error: %s", 
	      error ? error->message : "NULL");
  mark_point();
  fail_unless(ret, "returned error!");
 
  g_usleep(G_USEC_PER_SEC / 100);
  mark_point();

  pidfile= fopen(fname, "r");
  fail_if(NULL == pidfile, "could not open pidfile");
  fscanf(pidfile, "%d", &pid);
  fclose(pidfile);
  g_debug("pid is %d", pid);

  mark_point();

  g_usleep(G_USEC_PER_SEC / 100);

  ret= sandboxer_service_find_process_by_pid(sandserv, pid, 
					     &dbus_path_found, &secret_cookie, &unique_name,
					     &error);
  mark_point();

  if (!ret) {
    if (error) {
      fail("FindProcessByPid returned FALSE: %s", error->message);
    } else {
      fail("FindProcessByPid returned FALSE and error was NULL");
    }
  }
  fail_if(NULL == dbus_path_found, "dbus_path_found is NULL");
  fail_unless(g_str_equal(dbus_path, dbus_path_found), "%s is not %s", dbus_path_found, dbus_path);
  kill(pid, SIGTERM);
  pop_cred();
}
END_TEST

START_TEST(test_sandboxer_service_add_package_file_valid)
{
  SandboxerSandboxService* sandserv;
  GError *error= NULL;
  gboolean ret;

  push_cred(sandboxer_uid, sandboxer_gid);
  sandserv= sandboxer_sandbox_service_new();
  fail_if(NULL == sandserv, "failed to get sandbox service");
  ret= sandboxer_service_add_package_file(sandserv, 
					  "/tmp/a_test_thingy.package", 
					  &error);
  fail_unless(ret, "sandboxer_service_add_package_file returned FALSE");
  fail_unless(NULL == error, "error is not NULL");
}
END_TEST

START_TEST(test_sandboxer_service_add_package_file_invalid)
{
  SandboxerSandboxService* sandserv;
  GError *error= NULL;
  gboolean ret;

  push_cred(sandboxer_uid, sandboxer_gid);
  sandserv= sandboxer_sandbox_service_new();
  fail_if(NULL == sandserv, "failed to get sandbox service");
  ret= sandboxer_service_add_package_file(sandserv, 
					  "/tmp/invalid.package", 
					  &error);
  fail_if(ret, "sandboxer_sandbox_service_new returned TRUE");
  fail_if(NULL == error, "error is NULL");
  pop_cred();
}
END_TEST

START_TEST(test_sandboxer_service_remove_package_valid)
{
  SandboxerSandboxService* sandserv;
  GError *error= NULL;
  gboolean ret;

  push_cred(sandboxer_uid, sandboxer_gid);
  sandserv= sandboxer_sandbox_service_new();
  fail_if(NULL == sandserv, "failed to get sandbox service");
  ret= sandboxer_service_remove_package(sandserv,
					"coreutils",
					&error);
  fail_unless(ret, "sandboxer_service_remove_package returned FALSE");
  fail_unless(NULL == error, "error is not NULL");
  pop_cred();
}
END_TEST


Suite *
sandboxer_suite(void)
{
  Suite *s;
  TCase *tc;

  s= suite_create("Sandboxer");

  tc = tcase_create ("Sandboxer New");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, tmpdir_teardown);
  tcase_add_test (tc, test_new);
  suite_add_tcase (s, tc);

  tc= tcase_create ("Sandboxer Get Process");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, tmpdir_teardown);
  tcase_add_test (tc, test_get_process);
  suite_add_tcase(s, tc);

  tc= tcase_create ("Sandboxer List All Processes");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, tmpdir_teardown);
  tcase_add_test (tc, test_list_all_processes_empty);
  tcase_add_test (tc, test_list_all_processes_notempty);
  suite_add_tcase(s, tc);

  tc= tcase_create ("Sandboxer Find Process By Pid (Valid)");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, NULL);
  tcase_add_unchecked_fixture (tc, NULL, tmpdir_teardown);
  tcase_add_test (tc, test_find_process_by_pid);
  suite_add_tcase(s, tc);

  tc= tcase_create ("Sandboxer Find Process By Pid (Invalid)");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, NULL);
  tcase_add_unchecked_fixture (tc, NULL, tmpdir_teardown);
  tcase_add_test (tc, test_find_process_by_pid_invalid_pid);
  suite_add_tcase(s, tc);
  
  tc= tcase_create ("Sandboxer Add Package File");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, tmpdir_teardown);
  tcase_add_test (tc, test_sandboxer_service_add_package_file_valid);
  tcase_add_test (tc, test_sandboxer_service_add_package_file_invalid);
  suite_add_tcase(s, tc);
  
  tc= tcase_create ("Sandboxer Remove Package");
  tcase_add_checked_fixture (tc, setup_tmpdir_and_config_store, tmpdir_teardown);
  tcase_add_test (tc, test_sandboxer_service_remove_package_valid);
  suite_add_tcase(s, tc);

  return s;
}
