/*
 * application-sandboxer: an application isolation mechanism
 * Copyright (C) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St 
 * - Fifth Floor, Boston, MA 02110-1301 USA
 *
 */


#include <config.h>
#include <check.h>
#include <test/config_file_helper.h>
#include <daemon/config_store.h>
#include <daemon/sandboxer-helper.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#define TEST1_PACKAGE_NAME "test1.package"
#define TEST1_PACKAGE_CONTENTS			\
  "[Sandbox]\n"					\
  "SandboxName=test1_sandbox\n"			\
  "PackageName=test1_package_name\n"		\
  "Users=tester;nobody;daemon\n"		\
  "ExecPaths=/foo/bar;/baz/qux\n"

#define TEST2_PACKAGE_NAME "test2.package"
#define TEST2_PACKAGE_CONTENTS			\
  "[Sandbox]\n"					\
  "SandboxName=test1_sandbox\n"			\
  "PackageName=test2_package_name\n"		\
  "Users=nobody\n"				\
  "ExecPaths=/random1;/random2\n"

#define TEST3_PACKAGE_NAME "test3.package"
#define TEST3_PACKAGE_CONTENTS			\
  "[Sandbox]\n"					\
  "SandboxName=test2_sandbox\n"			\
  "PackageName=test3_package_name\n"		\
  "Users=nobody\n"				\
  "ExecPaths=/this;/that\n"

#define COREUTILS_PACKAGE_NAME "coreutils.package"
#define COREUTILS_PACKAGE_CONTENTS		\
  "[Sandbox]\n"					\
  "SandboxName=coreutils_sandbox\n"		\
  "PackageName=coreutils\n"			\
  "Users=tester;\n"				\
  "ExecPaths=/bin/true;/bin/false;\n"		

#define COREUTILS2_PACKAGE_NAME "coreutils2.package"
#define COREUTILS2_PACKAGE_CONTENTS		\
  "[Sandbox]\n"					\
  "SandboxName=coreutils_shared_sandbox\n"	\
  "PackageName=coreutils2\n"			\
  "Users=tester;\n"				\
  "ExecPaths=/bin/date;/bin/uname;\n"

#define COREUTILS3_PACKAGE_NAME "coreutils3.package"
#define COREUTILS3_PACKAGE_CONTENTS		\
  "[Sandbox]\n"					\
  "SandboxName=coreutils_shared_sandbox\n"	\
  "PackageName=coreutils3\n"			\
  "Users=tester;\n"				\
  "ExecPaths=/bin/touch;/bin/sleep;\n"

#define TESTUTILS_PACKAGE_NAME "testutils.package"
#define TESTUTILS_PACKAGE_TEMPLATE		\
  "[Sandbox]\n"					\
  "SandboxName=testutils_sandbox\n"		\
  "PackageName=test_utils\n"			\
  "Users=tester;\n"							\
  "ExecPaths=@tmpdir@/pid_writer;@tmpdir@/return_value;@tmpdir@/sleep_forever;\n"

#define SANDBOXERD_CONF_NAME "sandboxerd.conf"
#define SANDBOXERD_CONF_TEMPLATE				    \
  "[Sandboxerd]\n"						    \
  "EnabledPlugins=@enabled_plugins@\n"				    \
  "PluginsPath=@tmpdir@/plugins\n"				    \
  "SandboxerUser=sandboxer\n"					    \
  "PackageFilePath=@tmpdir@/packages\n"				    \
  "JailDirPath=@tmpdir@/jails\n"				    \
  "PidFilePath=@tmpdir@/sandboxerd.pid\n"			    \
  "LibExecPath=@tmpdir@/bin\n"					    \
  "DirsInJail=/bin;/usr;/lib;/opt;/tmp;/var;/etc;/dev;/proc;/sys\n" \
  "LogLevel=DEBUG\n" 


extern uid_t sandboxer_uid;
extern gid_t sandboxer_gid;
extern uid_t tester_uid;
extern gid_t tester_gid;




static void
nuke_test_config(gchar *prefix);

static gchar*
make_testdir(const gchar* base, const gchar *name, const int mode);

static void
copy_test_programs();

static void
copy_binaries();

static void
copy_plugins();

static gboolean 
regex_callback(const GMatchInfo *match_info,
	       GString *result,
	       GHashTable *hash);








void
write_test_package_files(void)
{
  gchar *test_utils_contents;
  gchar *cwd;

  cwd= g_get_current_dir();
  if (-1 == g_chdir(get_packagedir()))
    g_error("couldn't change to %s", get_packagedir());
  push_cred(sandboxer_uid, sandboxer_gid);
  if (!g_file_set_contents(TEST1_PACKAGE_NAME,
			   TEST1_PACKAGE_CONTENTS,
			   -1,
			   NULL))
    g_error("couldn't write %s", TEST1_PACKAGE_NAME);
  if (!g_file_set_contents(TEST2_PACKAGE_NAME,
			   TEST2_PACKAGE_CONTENTS,
			   -1,
			   NULL))
    g_error("couldn't write %s", TEST2_PACKAGE_NAME);
  if (!g_file_set_contents(TEST3_PACKAGE_NAME,
			   TEST3_PACKAGE_CONTENTS,
			   -1,
			   NULL))
    g_error("couldn't write %s", TEST3_PACKAGE_NAME);

  if (!g_file_set_contents(COREUTILS_PACKAGE_NAME,
			   COREUTILS_PACKAGE_CONTENTS,
			   -1,
			   NULL))
    g_error("couldn't write %s", COREUTILS_PACKAGE_NAME);
  if (!g_file_set_contents(COREUTILS2_PACKAGE_NAME,
			   COREUTILS2_PACKAGE_CONTENTS,
			   -1,
			   NULL))
    g_error("couldn't write %s", COREUTILS2_PACKAGE_NAME);
  if (!g_file_set_contents(COREUTILS3_PACKAGE_NAME,
			   COREUTILS3_PACKAGE_CONTENTS,
			   -1,
			   NULL))
    g_error("couldn't write %s", COREUTILS3_PACKAGE_NAME);
  
  test_utils_contents= template(TESTUTILS_PACKAGE_TEMPLATE, "tmpdir", get_tmpdir(), NULL);
  g_assert(test_utils_contents);
  if (!g_file_set_contents(TESTUTILS_PACKAGE_NAME,
			   test_utils_contents,
			   -1,
			   NULL))
    g_error("couldn't write %s", TESTUTILS_PACKAGE_NAME);
  g_free(test_utils_contents);  
  pop_cred();
  if (-1 == g_chdir(cwd))
    g_error("couldn't chdir to %s", cwd);
}

/*
 * gets the uid and gid of supplied name
 * 
 */
void 
get_uidgid(const char *name, uid_t *puid, gid_t *pgid)
{
  struct passwd *ppasswd;
  
  g_assert(name);
  g_assert(puid);

  ppasswd= getpwnam(name);
  if (!ppasswd)
    g_error("no user named \"%s\"", name);
  *puid= ppasswd->pw_uid;
  if (pgid)
    *pgid= ppasswd->pw_gid;
}

static void
nuke_test_config(gchar *prefix)
{
  gchar *filename;
  
  g_assert(prefix);
  filename= g_strdup_printf("%s/a_test_package.package", prefix);;

  if (g_file_test(filename, G_FILE_TEST_EXISTS)) {
    fail_unless(0 == g_unlink(filename), "couldn't unlink %s", filename);
  }
}



void
setup_tmpdir_and_config_store(void)
{
  SandboxGlobalConfig *global_config;
  SandboxConfigStore *config_store;
  gchar *tmpdir;

  mark_point();
  tmpdir_setup();
  tmpdir= (gchar*)get_tmpdir();
  nuke_test_config(tmpdir);
  global_config= sandbox_global_config_new(get_sandboxerd_conf());
  fail_if(NULL == global_config, "global_config is NULL");
  config_store= sandbox_config_store_new(get_packagedir());
  fail_if(NULL == config_store, "config_store is NULL!"); 
  mark_point();
}

static gchar*
make_testdir(const gchar* base, const gchar *name, const int mode)
{
  gchar *combined_name;
  mode_t old_umask;
  
  old_umask= umask(0);
  if (name) {
    combined_name= g_strdup_printf("%s/%s", base, name);
  } else {
    combined_name= g_strdup(base);
  }
  fail_unless(0 == g_mkdir_with_parents(combined_name, mode), 
	      "failed to make %s: %s", combined_name, g_strerror(errno));
  umask(old_umask);
  return combined_name;
}

const gchar* 
get_tmpdir(void)
{
  static gchar *tmpdir= NULL;

  if (!tmpdir) {
    tmpdir= g_strdup_printf("%s/sandboxer_unit_test-%d", 
			    g_get_tmp_dir(),
			    getppid());
  }
  return tmpdir;
}

const gchar*
get_packagedir(void)
{
  static gchar *package_dir= NULL;
  
  if (!package_dir) {
    package_dir= g_strdup_printf("%s/packages", get_tmpdir());
  }
  return package_dir;
}

const gchar* 
get_sandboxerd_conf(void)
{
  static gchar *sandboxerd_conf= NULL;

    if (!sandboxerd_conf) {
      sandboxerd_conf= g_strdup_printf("%s/sandboxerd.conf", get_tmpdir());
    }
  return sandboxerd_conf;
}

const gchar*
get_bindir(void)
{
  static gchar *bindir= NULL;

  if (!bindir) {
    bindir= g_strdup_printf("%s/bin", get_tmpdir());
  }
  return bindir;
}


void 
make_sandboxerd_conf(gchar *enabled_plugins)
{ 
  gchar *sandboxerd_conf_contents;
  gchar *cwd;
  gchar *tmpdir;
  
  if (!enabled_plugins) {
    enabled_plugins= NULL;
  }
  tmpdir= (gchar*)get_tmpdir();
  cwd= g_get_current_dir();
  sandboxerd_conf_contents= template(SANDBOXERD_CONF_TEMPLATE, 
				     "tmpdir", tmpdir,
				     "enabled_plugins", enabled_plugins,
				     NULL);
  g_assert(sandboxerd_conf_contents);
  
  push_cred(0,0);//begin root
  if (-1 == g_chdir(tmpdir)) g_error("can't chdir to %s", tmpdir);
  if (!g_file_set_contents(SANDBOXERD_CONF_NAME, sandboxerd_conf_contents, -1, NULL))
    g_error("couldn't write %s", SANDBOXERD_CONF_NAME);
  if (-1 == g_chdir(cwd)) g_error("can't chdir to %s", cwd);
  pop_cred(); //end root
}

static void
copy_test_programs()
{
  const gchar *helper_names[]= { "pid_writer", "return_value", "sleep_forever", NULL};
  guint i;
  
  for (i= 0; helper_names[i]; i++) {
    gchar *command_line;
    
    //install -m mode -g group -o owner source dest
    command_line= g_strdup_printf("install ./%s %s", helper_names[i], get_tmpdir());
    if (WEXITSTATUS(0) != system(command_line)) g_error("failed to run %s", command_line);	  
    g_free(command_line);
  }
}

static void
copy_binaries()
{
  const gchar *binary_names[]= { "../daemon/sandboxerd", "../daemon/stubd",
				 "../client/sandboxer", "../util/sandboxer_add_remove_package",
				 NULL}; 
  const gchar *symlink_names[]= { "sandboxer_add_package", "sandboxer_remove_package", NULL};
  gchar *cwd;
  guint i;

  push_cred(0,0);
  for (i= 0; binary_names[i]; i++) {
    gchar *command_line;
    
    //install -m mode -g group -o owner source dest
    command_line= g_strdup_printf("install %s %s", binary_names[i], get_bindir());
    if (WEXITSTATUS(0) != system(command_line)) g_error("failed to run %s", command_line);	  
    g_free(command_line);
  }
  cwd= g_get_current_dir();
  if (g_chdir(get_bindir())) g_error("couldn't chdir to %s", get_bindir());
  for (i= 0; symlink_names[i]; i++) {
    gchar *old_path;

    old_path= g_strdup_printf("%s/sandboxer_add_remove_package", get_bindir());
    if (symlink("sandboxer_add_remove_package", symlink_names[i])) {
      g_error("symlink returned error: %s", g_strerror(errno));
    }
    g_free(old_path);
  }
  if (g_chdir(cwd)) g_error("couldn't chdir to %s", cwd); //TODO: functionify
  pop_cred();
}

static void
copy_plugins()
{
  const gchar *plugin_names[]= {"libdbus.so", "libtest.so", "libenv.so", "libxauth.so", NULL};
  const gchar *plugin_prefix= "../plugins/.libs";
  guint i;

  push_cred(0,0);
  for (i= 0; plugin_names[i]; i++) {
    gchar *command_line;
    
    //install -m mode -g group -o owner source dest
    command_line= g_strdup_printf("install %s/%s %s/plugins", 
				  plugin_prefix, plugin_names[i], get_tmpdir());
    if (WEXITSTATUS(0) != system(command_line)) g_error("failed to run %s", command_line);	  
    g_free(command_line);
  }
  pop_cred();
}


void
tmpdir_setup(void)
{
  gchar *tmpdir;

  mark_point();
  tmpdir= (gchar*)get_tmpdir(); g_assert(tmpdir);
  if (g_file_test(tmpdir, G_FILE_TEST_EXISTS)) {
    tmpdir_teardown();
  }

  push_cred(0,0);
  make_testdir(tmpdir, NULL, (S_IRWXU | 
			      S_IRGRP | S_IXGRP |
			      S_IROTH | S_IXOTH));
  make_testdir(tmpdir, "bin", (S_IRWXU | 
			       S_IRGRP | S_IXGRP |
			       S_IROTH | S_IXOTH));
  pop_cred(); 

  push_cred(0, sandboxer_gid);
  make_testdir(tmpdir, "jails", S_IRWXU | S_IRWXG);
  make_testdir(tmpdir, "packages", S_IRWXU | S_IRWXG);
  make_testdir(tmpdir, "plugins", (S_IRWXU | 
				   S_IRGRP | S_IXGRP |
				   S_IROTH | S_IXOTH));
  pop_cred();

  make_sandboxerd_conf(NULL);
  write_test_package_files();
  copy_test_programs();
  copy_binaries();
  copy_plugins();

#if 0  //for debug
  {
    gchar *command_line;
    
    command_line= g_strdup_printf("ls -lR %s", get_tmpdir());
    g_debug("LS OUTPUT:");
    push_cred(0,0);
    system(command_line);
    pop_cred();
    g_free(command_line);
  }
#endif
  mark_point();
}


void
tmpdir_teardown(void)
{
  gchar *cmd_line;

  mark_point();
  nukem();
  mark_point();
  cmd_line= g_strdup_printf("rm -rf %s", get_tmpdir());
  push_cred(0,0);
  mark_point();
  system(cmd_line);
  mark_point();
  pop_cred();
  mark_point();
}

static gboolean 
regex_callback(const GMatchInfo *match_info,
	       GString *result,
	       GHashTable *hash)
{
  gchar *value;
  gchar *key;
  
  key= g_match_info_fetch(match_info, 1);
  value= g_hash_table_lookup(hash, key);
  if (value)
    g_string_append(result, value);
  return FALSE;
}


/**
 * template: replace keys in input stream with values from key list
 *
 * @input: a string in the form of <literal>"foo @bar@foo"</literal>
 * where the value of <literal>"bar"</literal> is looked for in the
 * list of args
 *
 * @...: a NULL terminated list of keys and values.  Note that an even
 * number of arguments (one value per key) is required
 *
 * Returns: a newly allocated string with the keys replaced by values
 */
gchar*
template(gchar *input, ...)
{
  va_list ap;
  gchar *key;
  GRegex *regex;
  gchar *result;
  GHashTable *hash;
  GError *error= NULL;

  g_assert(input);
  regex= g_regex_new("@([\\w]+)@", 0, 0, &error);
  g_assert(regex);

  hash= g_hash_table_new(g_str_hash, g_str_equal); g_assert(hash);
  va_start(ap, input);
  while (key= va_arg(ap, gchar*)) {
    gchar *value;

    value= va_arg(ap, gchar*);
    if (!value) break;
    g_hash_table_insert(hash, (gpointer)key, (gpointer)value);
  }
  
  result= g_regex_replace_eval(regex, input, -1, 0, 0, 
			       (GRegexEvalCallback)regex_callback, 
			       (gpointer)hash, &error);
  if (!result)
    g_error("error: %s", error->message);

  g_hash_table_unref(hash);
  va_end(ap);
  return result;
}


GNode*
g_node_new_from_pid(const GPid pid)
{ 
  GNode* node;
  GPid* child_pid;
  
  child_pid= g_new(GPid, 1);
  *child_pid= pid;
  node= g_node_new((gpointer)child_pid);  
}

void
insert_child_pids(const GPid pid, GNode *node)
{
  gchar *stdout;
  gchar *stderr;
  gchar *cmd;
  gint exit_status;
  GError *error= NULL;

  g_assert(node);
  cmd= g_strdup_printf("ps --ppid %d -o pid", pid);
  if (!g_spawn_command_line_sync(cmd, &stdout, &stderr, &exit_status, &error)) {
    g_error("couldn't run \"%s\": %s", cmd, error->message);
  }
  g_free(cmd);
  g_free(stderr);  
  if (0 == WEXITSTATUS(exit_status)) {
    gchar **pids;
    guint i;

    pids= g_strsplit(stdout, "\n", 0);
    g_assert(pids);
    for (i= 1; pids[i]; i++) {
      GNode *child_node;
      GPid child_pid;
      
      if (g_str_equal("", pids[i])) break;
      
      child_pid= strtol(pids[i], NULL, 10);
      child_node= g_node_new_from_pid(child_pid);
      g_node_insert(node, -1, child_node);
      insert_child_pids(child_pid, child_node);
    }     
    g_strfreev(pids);
  }
  g_free(stdout);
}

static gboolean
kill_traversal_function(GNode *node, const int *sig)
{
  GPid pid;

  g_assert(sig);
  g_assert(node); 
  g_assert(node->data);


  pid= *((GPid*)(node->data));
  if (pid) {
    if (kill(pid, *sig)) {
      if (ESRCH != errno) {
	g_error("couldn't kill %d: %s", pid, g_strerror(errno));
      }
    }
  }
}

void
kill_all_nodes(GNode *root_node,
	       int sig)
{
  g_assert(root_node);
  mark_point();
  g_node_traverse(root_node,
		  G_POST_ORDER,
		  G_TRAVERSE_ALL,
		  -1,
		  (GNodeTraverseFunc)kill_traversal_function,
		  (gpointer)(&sig));
  mark_point();
}

//nuke any lollygaggers
void 
nukem(void)
{
  GNode* node;

  mark_point();
  node= g_node_new_from_pid(0);
  push_cred(0,0);
  insert_child_pids(getpid(), node);

  kill_all_nodes(node, SIGKILL);
  g_usleep(200*MS);

  system("killall -q -9 sandboxerd stubd");
  g_usleep(100*MS);
  pop_cred();
  mark_point();
}
