import subprocess
import thread
from threading import Thread
import unittest
import signal
import time
import os


class Command(Thread):
    """A wrapper around Popen"""
    def __init__(self, cmd, output, passthrough=False):
        Thread.__init__(self)
        self.cmd= cmd
        self.output= output
        self.output['stdout']= ''
        self.output['stderror']= ''
        self.output['returncode']= None
        self.passthrough= passthrough

    def run(self):
        self.proc = subprocess.Popen(self.cmd,
                                     shell=True,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE)
        stdout, stderr= self.proc.communicate()
        self.output['stdout']= stdout
        self.output['stderr']= stderr

    def kill(self, signal):
        self.output['returncode']= self.proc.returncode
        if self.output['returncode'] == None:
            os.kill(self.proc.pid, signal)
            time.sleep(0.1)
            self.output['returncode']= self.proc.returncode

class SandboxerTestCase(unittest.TestCase):
    """Base class of sandboxer test cases.  Modifies run so that
    tearDown can have the test result in it.  The idea is that
    sandboxerd is run in setUp and the debug output is given in
    tearDown if the test fails.  It is currently a little awkward as
    one test failing will cause all tests to output their debug output.

    """

    def startSandboxerd(self):
        """Start the sandboxer daemon."""
        self.sandboxerd_output= {}
        self.sandboxerd= Command('/usr/sbin/sandboxerd', self.sandboxerd_output)
        self.sandboxerd.start()
        time.sleep(1) # to give things a chance to settle
    
    def setUp(self):
        self.startSandboxerd()

    def run(self, result=None):
        """Slightly hacked version from unittest.TestCase to pass
        result to tearDown
        
        """
        if result is None: result = self.defaultTestResult()
        result.startTest(self)
        testMethod = getattr(self, self._testMethodName)
        try:
            try:
                self.setUp()
            except KeyboardInterrupt:
                raise
            except:
                result.addError(self, self._exc_info())
                return

            ok = False
            try:
                testMethod()
                ok = True
            except self.failureException:
                result.addFailure(self, self._exc_info())
            except KeyboardInterrupt:
                raise
            except:
                result.addError(self, self._exc_info())

            try:
                self.tearDown(result)
            except KeyboardInterrupt:
                raise
            except:
                result.addError(self, self._exc_info())
                ok = False
            if ok: result.addSuccess(self)
        finally:
            result.stopTest(self)

    def stopSandboxerd(self, result=None):
        """Kill sandboxerd.  Works, but ought to be more throrough in
        its checks of failures as well as if sandboxerd crashed.
        
        """
        self.sandboxerd.kill(signal.SIGTERM)
        this_test= "%s (%s)" % (self._testMethodName, unittest._strclass(self.__class__))
        if this_test in [str(foo[0]) for foo in result.failures]:
            print "sandboxerd returned: ", self.sandboxerd_output['returncode']
            print "sandboxerd   stdout:"
            print self.sandboxerd_output['stdout']
            print "sandboxerd   stderr:" 
            self.sandboxerd_output['stderr']        
            print ""

    def tearDown(self, result=None):
        self.stopSandboxerd(result)
